;;; xref.el - (X)Emacs interface to Xrefactory source browser (body)

;; Copyright (C) 1997-2002 Marian Vittek, Xref-Tech
 
;; This file  is released  under the terms  of the GNU  General Public
;; License  version 2 as  published by  the Free  Software Foundation.
;; You should have  received a copy of the  GNU General Public License
;; along  with  this program;  if  not,  write  to the  Free  Software
;; Foundation,  Inc.,  59  Temple   Place  -  Suite  330,  Boston,  MA
;; 02111-1307,  USA.   

;; The  GNU GPL  license  applies only  on  xref.el and  xrefactory.el
;; files.   Distribution  of  other  parts of  Xrefactory,  especially
;; distribution of  the xref task and  its source code  is governed by
;; Xrefactory License Agreement.

;; This  file  is  part  of  Xrefactory  software,  it  implements  an
;; interface between  xref task and  (X)Emacs editors.  It  may invoke
;; functions  which  were implemented  by  independent developers  and
;; released  until GNU general  public license.   All the  credits for
;; those functions belong to  authors, contributors and maintainers of
;; those packages.  This concerns mainly 'compile.el', 'comint.el' and
;; 'browse-url.el'  packages.  Xrefactory is  not an  IDE but  we have
;; included "Emacs  IDE" submenu into  Xrefactory's (X)Emacs interface
;; in  order  to  allow  users  of  those  packages  to  profite  from
;; Xrefactory's  project management.   Of course,  our acknowledgement
;; belongs  also to  developers and  maintainers of  Emacs  and XEmacs
;; editors.

(provide 'xref)


(defvar xref-debug-mode nil)

(if (or (string-match "-nt" system-configuration) 
		(string-match "-win" system-configuration))
	(progn
	  (defvar xref-platform 'windows)
	  (defvar xref-slash ?\\)
	  (defvar xref-communication-file (concat (getenv "TEMP") "/xrefcomm.tmp"))
	  (defvar xref-out-communication-file (concat (getenv "TEMP") "/xrefcout.tmp"))
	  (defvar xref-run-batch-file (concat (getenv "TEMP") "/xrefrun.bat"))
	  (if (getenv "HOME")
		  (defvar xref-options-file (concat (getenv "HOME") "/_xrefrc"))
		(defvar xref-options-file "c:/_xrefrc")
		)
	  (defvar xref-user-identification "user")
	  (defvar xref-find-file-on-mouse-delimit "[^A-Za-z0-9_\\.~-]")
	  (defvar xref-path-separator ?\;)
	  )

  ;; a linux/unix platform
  (defvar xref-platform 'unix)
  (defvar xref-slash ?/)
  (defvar xref-communication-file (concat "/tmp/" (getenv "LOGNAME") "x.xrf"))
  (defvar xref-out-communication-file (concat "/tmp/" (getenv "LOGNAME") "xo.xrf"))
  (defvar xref-run-batch-file (concat "/tmp/" (getenv "LOGNAME") "xref.sh"))
  (defvar xref-options-file (concat (getenv "HOME") "/.xrefrc"))
  (defvar xref-user-identification (concat (getenv "LOGNAME") "u"))
  (defvar xref-find-file-on-mouse-delimit "[^A-Za-z0-9_/.~-]")
  (defvar xref-path-separator ?\:)
)

(if (string-match "XEmacs" emacs-version)
	(defvar xref-running-under 'xemacs)
  (defvar xref-running-under 'emacs)
)

(defun xref-add-bindings-for-chars (keymap chars fun)
  (let ((i) (len))
	(setq i 0)
	(setq len (length chars))
	(while (< i len)
	  (define-key keymap (substring chars i (+ i 1)) fun)
	  (setq i (+ i 1))
	  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                 Here are Xrefactory local keymaps                     ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                         Completion local keymap

(defvar xref-completion-mode-map
  (let ((map (make-keymap)))
    (define-key map "\C-c\C-c" 'interactive-xref-completion-escape)
    (define-key map "\e\e\e" 'interactive-xref-completion-escape)
    (define-key map "\C-q" 'interactive-xref-completion-escape)
    (define-key map "\eq" 'interactive-xref-completion-escape)
    (define-key map "\C-m" 'interactive-xref-completion-select)
    (define-key map "\C-d" 'interactive-xref-completion-goto)
    (define-key map " " 'interactive-xref-completion-goto)
    (define-key map [(backspace)] 'xref-completion-auto-search-back)
    (define-key map [(delete)] 'xref-completion-auto-search-back)
    (define-key map "\C-w" 'xref-completion-auto-search-w)
    (define-key map [tab] 'xref-completion-auto-search-s)
    (define-key map "\C-b" 'interactive-xref-completion-browse)
    (define-key map "\C-B" 'interactive-xref-completion-browse)
    (define-key map "\C-j" 'interactive-xref-completion-javadoc)
    (define-key map "\C-J" 'interactive-xref-completion-javadoc)
	(define-key map [(shift left)] 'xref-scroll-right)
	(define-key map [(shift right)] 'xref-scroll-left)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-completion-mouse-select)
		  (define-key map 'button2 'undefined)
		  (define-key map 'button3up 'undefined)
		  (define-key map 'button3 'xref-popup-xemacs-completion-menu)
		  (define-key map '(shift button2up) 'interactive-xref-completion-mouse-browse)
		  (define-key map '(shift button2) 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-completion-mouse-select)
	  (define-key map [mouse-3] 'xref-compl-3bmenu)
	  (define-key map [S-mouse-2] 'interactive-xref-completion-mouse-browse)
	)
    map)
  "Keymap for `xref-completion-list-mode'."
)
(xref-add-bindings-for-chars 
 xref-completion-mode-map
 "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_$1234567890"
 'xref-completion-auto-search)
(xref-add-bindings-for-chars xref-completion-mode-map
							"`[];',./-=\\~!@#%^&*()+|{}:\"<>?"
							'xref-completion-auto-switch)
(xref-add-bindings-to-keymap xref-completion-mode-map)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                     List of references local keymap


(defvar xref-list-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\C-c\C-c" 'interactive-xref-pop)
    (define-key map "\e\e\e" 'interactive-xref-pop)
    (define-key map "\C-m" 'interactive-xref-list-goto)
    (define-key map " " 'interactive-xref-list-goto)
    (define-key map "+" 'interactive-xref-list-filter-plus)
    (define-key map "-" 'interactive-xref-list-filter-minus)
    (define-key map [kp-add] 'interactive-xref-list-filter-plus)
    (define-key map [kp-subtract] 'interactive-xref-list-filter-minus)
    ;;(define-key map [(left)] 'xref-scroll-right)
    ;;(define-key map [(right)] 'xref-scroll-left)
    (define-key map [(shift left)] 'xref-scroll-right)
    (define-key map [(shift right)] 'xref-scroll-left)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-list-mouse-goto)
		  (define-key map 'button2 'undefined)
		  (define-key map 'button3 'interactive-xref-popup-xemacs-list-menu)
		  (define-key map 'button3up 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-list-mouse-goto)
	  (define-key map [mouse-3] 'interactive-xref-list-3bmenu)
	)
    map)
  "Keymap for `xref-list-mode'."
)
(xref-add-bindings-to-keymap xref-list-mode-map)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                 Selection of symbol from symbol/class list
;; This local keymaps applies on selection of symbol after a reference pushing
;; action. 

(defvar xref-menu-selection-mode-map
  (let ((map (make-sparse-keymap)))
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 
			'interactive-xref-menu-selection-mouse-select)
		  (define-key map 'button2 'undefined)
		  (define-key map 'button3 'interactive-xref-popup-xemacs-menu-menu)
		  (define-key map 'button3up 'undefined)
		  (define-key map '(shift button2) 'interactive-xref-menu-selection-mouse-inspect-symbol)
		  (define-key map '(shift button2up) 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-menu-selection-mouse-select)
	  (define-key map [mouse-3] 'interactive-xref-menu-3bmenu)
	  (define-key map [(shift mouse-2)] 'interactive-xref-menu-selection-mouse-inspect-symbol)
;;	  (define-key map [menu-bar] (make-sparse-keymap))
;;	  (define-key map [menu-bar xref-view]
;;		(cons "Xref-View" (make-sparse-keymap "Xref-View")))
;; 	  (define-key map [menu-bar xref-view name]
;;		'("Symbols of the same name" . xref-menu-view-name))
;; 	  (define-key map [menu-bar xref-view profil]
;;		'("Symbols with the same profil" . xref-menu-view-profil))
;; 	  (define-key map [menu-bar xref-view profil]
;;		'("Symbols virtualy adequate" . xref-menu-view-virtual))
	  )
	;; commons
    ;;(define-key map [(left)] 'xref-scroll-right)
    ;;(define-key map [(right)] 'xref-scroll-left)
    (define-key map [(shift left)] 'xref-scroll-right)
    (define-key map [(shift right)] 'xref-scroll-left)
    (define-key map "\e\e\e" 'interactive-xref-pop)
    (define-key map "q" 'interactive-xref-pop)
    (define-key map "\C-m" 'interactive-xref-menu-selection-go)
    (define-key map "g" 'interactive-xref-menu-selection-go)
    (define-key map "G" 'interactive-xref-menu-selection-go)
    (define-key map [insert] 'interactive-xref-menu-selection-select)
    (define-key map "t" 'interactive-xref-menu-selection-select)
    (define-key map "T" 'interactive-xref-menu-selection-select)
    (define-key map "d" 'interactive-xref-menu-selection-inspect-symbol)
    (define-key map "D" 'interactive-xref-menu-selection-inspect-symbol)
    (define-key map "s" 'interactive-xref-menu-selection-inspect-symbol)
    (define-key map "S" 'interactive-xref-menu-selection-inspect-symbol)
    (define-key map "c" 'interactive-xref-menu-selection-inspect-class)
    (define-key map "c" 'interactive-xref-menu-selection-inspect-class)
    (define-key map "a" 'interactive-xref-menu-selection-all)
    (define-key map "A" 'interactive-xref-menu-selection-all)
    (define-key map "n" 'interactive-xref-menu-selection-none)
    (define-key map "N" 'interactive-xref-menu-selection-none)
    (define-key map " " 'interactive-xref-menu-select-and-inspect-one)
    (define-key map "p" 'interactive-xref-menu-filter-plus)
    (define-key map "P" 'interactive-xref-menu-filter-plus)
    (define-key map "+" 'interactive-xref-menu-filter-plus)
    (define-key map [kp-add] 'interactive-xref-menu-filter-plus)
    (define-key map "m" 'interactive-xref-menu-filter-minus)
    (define-key map "M" 'interactive-xref-menu-filter-minus)
    (define-key map "-" 'interactive-xref-menu-filter-minus)
    (define-key map [kp-subtract] 'interactive-xref-menu-filter-minus)
    map)
  "Keymap for xref-menu for symbol selection."
)
(xref-add-bindings-to-keymap xref-menu-selection-mode-map)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                   Class (sub)tree result local keymap
;; This is keymap for window resulting from 'View class tree' function, not
;; for the class tree resulting from pushing actions.

(defvar xref-class-tree-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\e\e\e" 'xref-delete-windows)
    (define-key map [left] 'xref-scroll-right)
    (define-key map [right] 'xref-scroll-left)
    (define-key map [(shift left)] 'xref-scroll-right)
    (define-key map [(shift right)] 'xref-scroll-left)
    (define-key map "\C-m" 'xref-ct-inspect-def)
    (define-key map " " 'xref-ct-inspect-def)
    (define-key map "d" 'xref-ct-inspect-def)
    (define-key map "D" 'xref-ct-inspect-def)
    (define-key map "f" 'interactive-xref-class-tree-frame)
    (define-key map "F" 'interactive-xref-class-tree-frame)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-ct-inspect-def)
		  (define-key map 'button2 'undefined)
		  (define-key map '(shift button2up) 'interactive-xref-ct-inspect-def)
		  (define-key map '(shift button2) 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-ct-inspect-def)
	  (define-key map [S-mouse-2] 'interactive-xref-ct-inspect-def)
	)
    map)
  "Keymap for xref class tree."
)
(xref-add-bindings-to-keymap xref-class-tree-mode-map)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                     Search in Tag file local keymap

(defvar xref-tag-search-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\C-c\C-c" 'interactive-xref-completion-escape)
    (define-key map "\e\e\e" 'interactive-xref-completion-escape)
    (define-key map "q" 'interactive-xref-completion-escape)
    (define-key map "\C-m" 'interactive-xref-completion-select)
    ;;(define-key map [(left)] 'xref-scroll-right)
    ;;(define-key map [(right)] 'xref-scroll-left)
    (define-key map [(shift left)] 'xref-scroll-right)
    (define-key map [(shift right)] 'xref-scroll-left)
    (define-key map " " 'interactive-xref-completion-goto)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-completion-mouse-browse)
		  (define-key map 'button2 'undefined)
		  (define-key map '(shift button2up) 'interactive-xref-completion-mouse-goto)
		  (define-key map '(shift button2) 'undefined)
		  (define-key map 'button3 'xref-popup-xemacs-tag-selection-menu)
		  (define-key map 'button3up 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-completion-mouse-browse)
	  (define-key map [S-mouse-2] 'interactive-xref-completion-mouse-goto)
	  (define-key map [mouse-3] 'xref-tags-3bmenu)
	)
    map)
  "Keymap for `xref-search-in-tag-file-list-mode'."
)
(xref-add-bindings-to-keymap xref-tag-search-mode-map)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;               Xrefactory non customizable variables                   ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; following variable controls how lists of references are passed to (X)Emacs
;; passing through file is faster on most systems.
;;(defvar xref-reflist-synchro 'sync)       ;; pipe
(defvar xref-reflist-synchro 'syncfile)     ;; file

(defvar xref-qualified-name-max-context 500)
(defvar xref-expression-max-context 1000)
(defvar xref-dynamic-to-static-new-arg-name "new_arg")
(defvar xref-move-target-class "")
(defvar xref-move-method-end-marker (make-marker))
(defvar xref-move-target-marker (make-marker))
(defvar xref-tmp-mark (make-marker))
(defvar xref-move-tmp-mark (make-marker))
(defvar xref-undo-marker (make-marker))
(defvar xref-refactoring-init-marker (make-marker))
(defvar xref-encapsulate-setter-marker (make-marker))
(defvar xref-encapsulate-getter-marker (make-marker))
(defvar xref-reduce-this-marker (make-marker))
(defvar xref-completion-marker (make-marker))

(defvar xrefsrv-process nil)
(defvar xrefxref-process nil)
(defvar xref-frame-links nil)
(defvar xref-lastref-file " ")
(defvar xref-lastref-line 0)
(defvar xref-lastref-coll 0)
(defvar xref-filter-initialized nil)
(defvar xref-pilot-char ?\ )
(defvar xref-last-message-filtered 0)
(defvar xref-sync-last-message "Xref2end3of1synchronised7output0")
(defvar xref-filter-whole-output "")
(defvar xref-list-buffer-content nil)
(defvar xrefsrv-buffer) 
(defvar xref-move-refid) 
(defvar xref-move-usage)
(defvar xref-list-invocation-window) 
(defvar xref-backward-pass-identifier-regexp "\\(\\`\\|[^A-Za-z0-9_$]\\)")
(defvar xref-forward-pass-identifier-regexp "\\(\\'\\|[^A-Za-z0-9_$]\\)")
(defvar xref-pass-identifier-regexp "\\(\\'\\|\\`\\|[^A-Za-z0-9_$]\\)")
(defvar newline-char (string-to-char "
"))
(defvar xref-menu-selection-prefix-lines 2)
(defvar xref-fontification nil)
(defvar xref-pushrefs-action nil)
(defvar xref-project-list)
(defvar xref-list-filter-level 0)
(defvar xref-sym-menu-filter-level 2)
(defvar xref-next-push-manual nil)
(defvar xref-last-message nil)
(defvar xref-last-referencing-action nil)
(defvar xref-last-change-attr-old-attribute "")
(defvar xref-last-foreign-project nil)
(defvar xref-run-this-option "runthis")
(defvar xref-insert-arg-renaming-new-name nil)   ;; this is tricky

(defvar xref-directory-dep-prj-name "++ Automatic (directory dependent) ++")
(defvar xref-abandon-deletion "++ Cancel (no deletion) ++")
(defvar xref-list-buffer-message "button2(' ') - inspect;  button3 - menu;")

(defvar xref-query-replace-map (make-sparse-keymap "Xref-query-replace"))
(define-key xref-query-replace-map "y" 'answer-yes)
(define-key xref-query-replace-map "Y" 'answer-yes)
(define-key xref-query-replace-map "n" 'answer-no)
(define-key xref-query-replace-map "N" 'answer-no)
(define-key xref-query-replace-map "a" 'answer-all)
(define-key xref-query-replace-map "A" 'answer-all)
(define-key xref-query-replace-map "c" 'answer-confirmed)
(define-key xref-query-replace-map "C" 'answer-confirmed)

(defvar xref-string-passing-regexp "\\([^\\]\\(\\\\\\\\\\)*\"\\|\n\\)")  
;; there can be a pair number of \ before ", take it into account
;; old value was "\\([^\\]\"\\|\n\\)")

(defvar xref-keywords-regexp
"\\b\\(auto\\|break\\|c\\(ase\\|har\\|on\\(st\\|tinue\\)\\)\\|d\\(efault\\|o\\(uble\\)?\\)\\|e\\(lse\\|num\\|xtern\\)\\|f\\(loat\\|or\\)\\|goto\\|i\\(f\\|nt\\)\\|long\\|re\\(gister\\|turn\\)\\|s\\(hort\\|i\\(gned\\|zeof\\)\\|t\\(atic\\|ruct\\)\\|witch\\)\\|typedef\\|un\\(ion\\|signed\\)\\|vo\\(id\\|latile\\)\\|while\\)\\b"
)

;; somewhere they don't have the 'regexp-opt' function ...
;;
;;(defvar xref-keywords-regexp
;;  (eval-when-compile
;;	(concat "\\b"
;;			(regexp-opt
;;			 '("auto" "extern" "register" "static" "typedef" "struct"
;;			   "union" "enum" "signed" "unsigned" "short" "long"
;;			   "int" "char" "float" "double" "void" "volatile" "const"
;;			   "break" "continue" "do" "else" "for" "if" "return"
;;			   "switch" "while" "goto" "case" "default" "sizeof"
;;			   ) t )
;;			"\\b")
;;	))

(defvar xref-font-lock-compl-keywords
    (cons (cons xref-keywords-regexp '(xref-keyword-face))
	  '(("^\\([a-zA-Z0-9_ ]*\\):" xref-list-pilot-face))
	  )
    "Default expressions to highlight in Xref completion list modes.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                            The interface                              ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-list-buffer-name ()
  (let ((res))
	(setq res (format " *xref-list* (%s)" (xref-get-frame-id)))
	res
))

(defun xref-symbol-resolution-buffer-name ()
  (let ((res))
	(setq res (format " *xref-symbol-resolution* (%s)" (xref-get-frame-id)))
	res
))

(defun xref-class-tree-buffer-name ()
  (let ((res))
	(setq res (format "*xref-class-tree* (%s)" (xref-get-frame-id)))
	res
))


(defun xref-move-target-file ()
  (let ((res))
	(setq res "")
	(if (marker-buffer xref-move-target-marker)
		(setq res (buffer-file-name (marker-buffer xref-move-target-marker)))
	  )
	res
))

(defun xref-move-target-line ()
  (let ((res) (cb) (cp))
	(save-excursion
	  (xref-set-to-marker xref-move-target-marker)
	  (setq res (count-lines 1 (+ (point) 1)))
	  )
	res
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; secure buffer erase

(defun xref-erase-buffer ()
  (let ((bname) (suf))
	(setq bname (buffer-file-name nil))
	(if bname
		(setq suf (xref-file-name-extension bname))
	  (setq suf "")
	  )
	(if (or (equal suf "java") (equal suf "c") (equal suf "y"))
		(progn
		  (message "[xref] internal error: Attempt to erase '%s'!" bname)
		  (if (yes-or-no-p (format "![warning] Attempt to erase '%s'! Really erase? " (xref-file-last-name bname)))
			  (erase-buffer)
			)
		  )
	  (erase-buffer)
	  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;  highlighting ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-font-lock-list-keywords (sym)
    (cons (cons xref-keywords-regexp '(xref-keyword-face))
	(cons (cons (concat "\\b" sym "\\b") '(xref-list-symbol-face))
	  '(("^.\\([^:]*:[0-9]*\\):" xref-list-pilot-face))
	  ))
)


(defun completion-symbol-highlight (begpos endpos)
  (let ((line) (sym) (spos) (mbeg) (mend))
	(setq line (buffer-substring begpos endpos))
	(setq spos (string-match "[^a-zA-Z0-9_]" line 0))
	(setq sym (substring line 0 spos))
	(setq mbeg (string-match (concat "\\b[a-zA-Z0-9_\\.]*\\b" sym "\\b") line spos))
	(if mbeg
		(progn
		  (setq mend (match-end 0))
		  (put-text-property (+ mbeg begpos) 
							 (+ mend begpos) 
							 'face 'xref-list-symbol-face)
		  ))
))

(defun hightlight-keywords (begpos endpos keywords-regexp keyword-face)
  (let ((buff) (bpos) (loop-flag) (mbeg) (mend)
		)
	(if (< begpos endpos)
		(progn
		  (setq buff (buffer-substring begpos endpos))
		  (setq bpos 0)
		  (setq loop-flag t)
		  (while loop-flag
			(setq mbeg (string-match keywords-regexp buff bpos))
			(if (eq mbeg nil)
				(setq loop-flag nil)
			  (setq mend (match-end 0))
			  (put-text-property (+ mbeg begpos) (+ mend begpos) 'face keyword-face)
			  (setq bpos mend)
	  ))))
))

(defun xref-fontify-region (begpos endpos kw-font-list)
  (let ((kwl) (kw))
	(setq kwl kw-font-list)
	(while kwl
	  (setq kw (car kwl))
	  (hightlight-keywords begpos endpos (car kw) (car (cdr kw)))
	  (setq kwl (cdr kwl))
	  )
))

(defun line-hightlight (begpos endpos multiple-lines-flag hbeg-offset) 
  (let ((bp) (loop-flag) (cpos) (prev-line) (dp-pos)
		)
;;(insert (format "[ %d-%d ]" begpos endpos))
	(setq cpos (char-search begpos -1 newline-char))
	(if (< cpos 0) (setq cpos 0))
	(if (and xref-coloring (< (+ cpos 1) endpos))
		(xref-fontify-region (+ cpos 1) endpos xref-fontification)
	  )
	(if (or xref-mouse-highlight xref-coloring)
		(progn
		  (while (< (+ cpos 1) endpos)
			(setq bp cpos)
			(setq loop-flag t)
			(while (and loop-flag 
						multiple-lines-flag
						(eq (char-after (+ bp 1)) ?\ )
						(eq (char-after (+ bp 2)) ?\ )
						)
			  (setq prev-line (char-search (- bp 1) -1 newline-char))
			  (if (< prev-line 1)
				  (setq loop-flag nil)
				(setq bp prev-line)
				)
			  )
			(setq cpos (char-search (+ cpos 1) 1 newline-char))
			(if (and xref-mouse-highlight (< (+ bp hbeg-offset) cpos))
				(put-text-property (+ bp hbeg-offset) cpos 'mouse-face 'highlight)
			  )
			(if (and xref-coloring (< (+ bp 1) cpos))
				(completion-symbol-highlight (+ bp 1) cpos)
			  )
			)))
))

;;;;;;;;;;;;;;;;;;;;;;;; various functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-buffer-has-one-of-suffixes (bname suffixes)
  (let ((suf) (sl) (res))
	(setq suf (concat "." (xref-file-name-extension bname)))
	(setq sl suffixes)
	(while (and sl (not (equal suf (car sl))))
	  (setq sl (cdr sl))
	  )
	(if (not sl)
		(setq res nil)
	  (setq res t)
	  )
	res
))

(defun xref-initial-file-suffix-check (allowed-suf err-message)
  (if (not (xref-buffer-has-one-of-suffixes (buffer-file-name) allowed-suf))
	  (error "%s" err-message)
	)
)

(defun xref-file-directory-name (fname)
  (let ((spos) (res) (len))
	(setq res nil)
	(setq len (length fname))
	(if (eq len 0)
		(setq res nil)
	  (setq spos (- len 1))
	  (while (and (> spos 0) 
				  (not (eq (elt fname spos) ?/))
				  (not (eq (elt fname spos) ?\\)))
		(setq spos (- spos 1))
		)
	  (if (> spos 0)
		  (setq res (substring fname 0 spos))
		(setq res nil)
		)
	  (if (and (eq xref-platform 'windows)
			   (eq spos 3)
			   (eq (elt fname (- spos 1)) ?\:))
		  (setq res nil)
		))
	res
))

(defun xref-file-name-extension (fname)
  (let ((spos) (res) (len))
	(setq res nil)
	(setq len (length fname))
	(if (eq len 0)
		(setq res nil)
	  (setq spos (- len 1))
	  (while (and (> spos 0) 
				  (not (eq (elt fname spos) ?.)))
		(setq spos (- spos 1))
		)
	  (if (> spos 0)
		  (setq res (substring fname (+ spos 1) len))
		(setq res fname)
		))
	res
))

(defun xref-file-last-name (fname)
  (let ((spos) (res) (len))
	(setq res nil)
	(setq len (length fname))
	(if (eq len 0)
		(setq res nil)
	  (setq spos (- len 1))
	  (while (and (> spos 0) 
				  (not (eq (elt fname spos) ?/))
				  (not (eq (elt fname spos) ?\\)))
		(setq spos (- spos 1))
		)
	  (if (> spos 0)
		  (setq res (substring fname (+ spos 1) len))
		(setq res fname)
		))
	res
))

(defun xref-file-name-without-suffix (fname)
  (let ((spos) (res) (len))
	(setq res nil)
	(setq len (length fname))
	(if (eq len 0)
		(setq res nil)
	  (setq spos (- len 1))
	  (while (and (> spos 0) 
				  (not (eq (elt fname spos) ?.)))
		(setq spos (- spos 1))
		)
	  (if (> spos 0)
		  (setq res (substring fname 0 spos))
		(setq res fname)
		))
	res
))

(defun xref-pass-to-fun-lpar ()
  (search-forward-regexp xref-forward-pass-identifier-regexp)
  (backward-char)
  (search-forward-regexp "[^
	 ]")
)

(defun xref-identifier-after-point ()
  (let ((cpoin (point)) (cid))
	(if (or (eq (char-after cpoin) ?#) (eq (char-after cpoin) ?$))
		(forward-char)
	  )
	(search-forward-regexp xref-forward-pass-identifier-regexp)
	(if (not (eq (point) (point-min))) (backward-char))
	;; I would like to delete it here, but you have to reworked
	;; renaming, removing of static prefix, ...
	(push-mark)
	(setq cid (buffer-substring cpoin (point)))
	(goto-char cpoin)
	cid
))

(defun xref-get-identifier-on (poin)
  (let ((opoin) (res))
	(setq opoin (point))
	(goto-char poin)
	(search-forward-regexp xref-forward-pass-identifier-regexp)
	(if (not (eq (point) (point-min))) (backward-char))
	(setq res (buffer-substring poin (point)))
	(goto-char opoin)
	res
))

(defun char-search (pos step char-code)
  (let ((res pos))
	(while (and (>= res 1) (< res (point-max)) 
				(not (eq (char-after res) char-code)))
	  (setq res (+ res step))
	  )
	res
))

(defun xref-delete-pending-ident ()
  (let ((p (point)))
	(search-forward-regexp xref-forward-pass-identifier-regexp)
	(if (not (eq (point) (+ (buffer-size) 1)))
		(backward-char)
	  )
	(if (eq p (point))
		()
	  (delete-char (- p (point)) t)
	  (message "** Pending identifier killed (C-y can restore it) **")
	  )
))

(defun xref-display-buffer (buffer)
  (let ((pp) (res))
	(setq pp pop-up-frames)
	(setq pop-up-frames nil)
	(setq res (display-buffer buffer))
	(setq pop-up-frames pp)
	res
))

(defun xref-string-has-prefix (str prefix case-unsens)
  (let ((res) (len) (pf))
	(setq res nil)
	(setq len (length prefix))
	(if (>= (length str) len)
		(progn
		  (setq pf (substring str 0 len))
		  (if (or (equal pf prefix)
				  (and case-unsens  (equal (downcase pf)
										   (downcase prefix))))
		(setq res t)
	  )))
	res
))

(defun xref-cut-string-prefix (ss prefix case-unsens)
  (let ((res) (len))
	(setq res ss)
	(setq len (length prefix))
	(if (>= (length ss) len)
		(if (or (equal (substring ss 0 len) prefix)
				(and case-unsens (equal (downcase (substring ss 0 len)) 
										(downcase prefix))))
			(setq res (substring ss len))
		  )
	  )
	res
))


(defun xref-cut-string-suffix (ss suffix case-unsens)
  (let ((res) (len) (sslen))
	(setq res ss)
	(setq len (length suffix))
	(setq sslen (length ss))
	(if (>= sslen len)
		(if (or (equal (substring ss (- sslen len)) suffix)
				(and case-unsens (equal (downcase (substring ss (- sslen len))) 
										(downcase suffix))))
			(setq res (substring ss 0 (- sslen len)))
		  )
	  )
	res
))


(defun xrefsrv-filter (process output)
  (let ((lenn) (tmp-pos) (tmp-pos2) (rof) (lnnum) (wsize))
;;(message "filter: %s" output)
	(if (eq xref-last-message-filtered nil)
		(progn
		  ;; it is synchronized dialog, wait until all input readed
		  (setq xref-filter-whole-output (concat xref-filter-whole-output output))
		  (setq xref-last-message-filtered 
				(string-match xref-sync-last-message xref-filter-whole-output))
		  (if (eq xref-last-message-filtered nil)
			  (setq output "")
			(setq output (substring xref-filter-whole-output 0 xref-last-message-filtered))
			)))
;;(message "filter2: %s" output)
	(setq lenn (length output))
	(if (eq lenn 0) 
		(setq xref-pilot-char ?\ )       ;; set it to blank
	  (if (not xref-filter-initialized)
		  (progn 
			(setq xref-filter-initialized t)
			  (setq xref-pilot-char (string-to-char (substring output 0 1)))
			  (setq output (substring output 1))
;;(message "pilot char %c" xref-pilot-char)
			(cond
			 (
			  (eq xref-pilot-char ?\;)
				  (get-buffer-create (xref-list-buffer-name))
				  (set-buffer (xref-list-buffer-name))
				  (setq buffer-read-only nil)
				  (xref-erase-buffer)
				  (setq truncate-lines t) ;;xref-buffers-truncate-lines)
				  (setq xref-list-buffer-content nil)
			 )(
			  (eq xref-pilot-char ?\')
				  (get-buffer-create " *xref-execute*")
				  (set-buffer " *xref-execute*")
				  (setq buffer-read-only nil)
				  (xref-erase-buffer)
			 )(
			  (eq xref-pilot-char ?\<)
				  (get-buffer-create (xref-class-tree-buffer-name))
				  (set-buffer (xref-class-tree-buffer-name))
				  (setq buffer-read-only nil)
				  (xref-erase-buffer)
				  (setq truncate-lines t) ;; xref-buffers-truncate-lines)
				  (insert "Class-Tree\n\n")
				  (setq buffer-read-only t)
			 )(
			  (eq xref-pilot-char ?\>)
				  (get-buffer-create (xref-symbol-resolution-buffer-name))
				  (set-buffer (xref-symbol-resolution-buffer-name))
				  (setq buffer-read-only nil)
				  (xref-erase-buffer)
				  (setq truncate-lines nil) ;; warning messages
				  (insert 
				   (concat
					"Select references:   "
					"INS)toggle, SPACE)one only, a)ll, n)one, q)uit g)o (<CR>);\n"
					"Filter: p)lus('+'), m)inus('-')\n"
				   ))
				  (setq buffer-read-only t)
			 )(
				(or (eq xref-pilot-char ?\.) (eq xref-pilot-char ?\,) 
					(eq xref-pilot-char ?\-) (eq xref-pilot-char ?\*)
					(eq xref-pilot-char ?\_) (eq xref-pilot-char ?\=)
					(eq xref-pilot-char ?^) 
					)
					nil
			 )(t
			   (get-buffer-create " *xrefsrv*")
			   (set-buffer " *xrefsrv*")
			   (setq buffer-read-only nil)
			   (xref-erase-buffer)
			   (setq truncate-lines nil) ;; error mesages, ...
			   )
			  )
			)
		)
   (if (eq lenn 0) ()
   (cond
	(
	(eq xref-pilot-char ?\,) 
	(set-buffer xrefsrv-buffer)
	(insert output)
	(if xref-single-completion-closes-window
		(delete-completion-window)
	  )
	(message " ")
	(xref-delete-pending-ident)
   )(
	(eq xref-pilot-char ?\.) 
	(set-buffer xrefsrv-buffer)
	(insert output)
	(if xref-single-completion-closes-window
		(delete-completion-window)
	  )
	(xref-delete-pending-ident)
	(message "** Multiple completions **")
   )(
	(eq xref-pilot-char ?\:)
	;; this is now used only to display trivial pre-checks messages
	(set-buffer " *xrefsrv*")
	(xref-display-buffer " *xrefsrv*")
	(goto-char (point-max))
	(setq tmp-pos (point))
	(insert output)
	;; (line-hightlight tmp-pos (point) t 1)
	(goto-char (point-min))
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\!)
	(set-buffer " *xrefsrv*")
	(xref-display-buffer " *xrefsrv*")
	(goto-char (point-max))
	(insert output)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\%)
	(set-buffer " *xrefsrv*")
	(goto-char (point-max))
	(setq tmp-pos (point))
	(insert output)
	(goto-char (point-min))
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\~)
	(set-buffer " *xrefsrv*")
	(goto-char (point-max))
	(insert output)
	(goto-char (point-min))
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\>)
	(set-buffer (xref-symbol-resolution-buffer-name))
	(xref-display-buffer (xref-symbol-resolution-buffer-name))
	(goto-char (point-max))
	(setq tmp-pos (point))
	(setq rof buffer-read-only)
	(setq buffer-read-only nil)
	(insert output)
	(line-hightlight tmp-pos (point) nil 2)
	(goto-char (point-min))
	(setq buffer-read-only rof)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\<)
	(set-buffer (xref-class-tree-buffer-name))
	(setq tmp-pos2 (point))
	(goto-char (point-max))
	(setq tmp-pos (point))
	(setq rof buffer-read-only)
	(setq buffer-read-only nil)
	(insert output)
	(line-hightlight tmp-pos (- (point) 1) nil 3)
	(goto-char tmp-pos2)
	(setq buffer-read-only rof)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\; )
	(set-buffer (xref-list-buffer-name))
	(xref-display-buffer (xref-list-buffer-name))
	(goto-char (point-max))
	(setq tmp-pos (point))
	(setq buffer-read-only nil)
	(insert output)
	(line-hightlight tmp-pos (point) nil 1)
	(setq lnnum (count-lines 1 (point-max)))
	(setq wsize (window-height (get-buffer-window (xref-list-buffer-name))))
	(if (and (> xref-window-minimal-size 0) 
			 (> wsize (+ lnnum xref-window-minimal-size)))
		(progn
		  (enlarge-window (- wsize lnnum xref-window-minimal-size))
		  ))
	(goto-char (point-min))
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\' )
	(set-buffer " *xref-execute*")
	(goto-char (point-max))
	(insert output)
	(delete-completion-window)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\#)
	(set-buffer " *xrefsrv*")
	(insert output)
	;;(delete-completion-window)
	(set-buffer xrefsrv-buffer)
   )(
	(eq xref-pilot-char ?\*) 
	(set-buffer xrefsrv-buffer)
	(setq xref-last-message output)
	(message "%s" output)
   )(
	(eq xref-pilot-char ?\-) 
	(set-buffer xrefsrv-buffer)
	;;(delete-completion-window)
	(message "** No completion possible **")
   )(
	(eq xref-pilot-char ?\_) 
	(set-buffer xrefsrv-buffer)
	;;(delete-completion-window)
	(message "** No reference **")
   )(
	(eq xref-pilot-char ?\=) 
	(set-buffer xrefsrv-buffer)
	;;(delete-completion-window)
	(message "** Empty stack, no pushed references **")
   )(
	 (eq xref-pilot-char ?^) 
	 ;; just synchronisation answer
   )(
	t
	(set-buffer " *xrefsrv*")
	(xref-display-buffer " *xrefsrv*")
	(message "**************** error in input format *****************")
	(insert output)
	(set-buffer xrefsrv-buffer)
   )
  )))
))

(defun start-xrefsrv-proc ()
  "sklsdkzxkzxk"
  (let ((oldpct process-connection-type))
	(setq process-connection-type nil)
	(setq xrefsrv-process (apply 'start-process 
                                 (concat xref-home-directory "xref") 
                                 nil 
                                 (concat xref-home-directory "xref")
								 "-task_regime_server"
                                 nil))
	(set-process-filter xrefsrv-process 'xrefsrv-filter)
	(setq process-connection-type oldpct)
  )
  (process-kill-without-query xrefsrv-process)
)

(defun xref-kill-xref-process (coredump)
"Kill currently running  xref process. 

If there is a xref process creating/updating tag file it is killed. If
there is no  such process this function kills  the xref server process
answering completion,  browsing and  refactoring requests.  It  can be
used when  xref task goes in  an inconsistent state.   Xref process is
then  restarted  automatically  at  next  invocation  of  any  of  its
functions.  The optional argument  COREDUMP if present causes that the
core of the process is dumped for debugging purposes.
"
;;  (interactive (list (if current-prefix-arg 'full)))
;; if previous do not work, use the next line rather
  (interactive "P")
  (if (and (not (eq xrefxref-process nil))
		   (eq (process-status xrefxref-process) 'run))
	  (progn
        (delete-process xrefxref-process)
        (setq xrefxref-process nil)
		(message "Extern Xref process killed.")
        )
	(setq xrefxref-process nil)
	(if (not (eq xrefsrv-process nil))
		(progn
		  (if current-prefix-arg
			  (progn
				(shell-command (format "kill -3 %d && echo Core dumped into this buffer directory." (process-id xrefsrv-process)))
				)
			(delete-process xrefsrv-process)
			(setq xrefsrv-process nil)
			(message "Emacs Xref server process killed.")
		  ))
	  (message "** No process to kill. **")
	  ))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-get-windows-status ()
  (let ((srw) (lw) (tw) (res))
	(setq res nil)
	(setq srw (get-buffer-window (xref-symbol-resolution-buffer-name) nil))
	(setq lw (get-buffer-window (xref-list-buffer-name) nil))
	(setq tw (get-buffer-window (xref-class-tree-buffer-name) nil))
	(if srw
		(setq res (cons 'xref-symbol-resolution res))
	  )
	(if lw
		(setq res (cons 'xref-list res))
	  )
	(if tw
		(setq res (cons 'xref-class-tree res))
	  )
	(if (and srw lw)
		(progn
		  (setq xref-win3-sym-res-win-size (window-height srw))
		  (setq xref-win3-list-win-size (window-height lw))
		  (xref-goto-source-window)
		  (setq xref-win3-src-win-size (window-height))		  
		  ))
	res
))

(defun xref-set-this-window-size (size)
  (let ((dff))
	(setq dff (- size (window-height)))
	(if (not (equal dff 0))
		(enlarge-window dff)
	  )
))

(defun xref-resize3-windows ()
  (let ((tsize) (srw) (lw) (wn))
	(setq wn 0)
	(setq srw (get-buffer-window (xref-symbol-resolution-buffer-name) nil))
	(setq lw (get-buffer-window (xref-list-buffer-name) nil))
	(if (and srw lw 
			 (<= (+ xref-win3-list-win-size 
					xref-win3-sym-res-win-size
					xref-win3-src-win-size)
				 (frame-height)
				 ))
		(progn
		  (xref-goto-source-window)
		  (xref-set-this-window-size xref-win3-src-win-size)
		  (select-window lw)
		  (xref-set-this-window-size xref-win3-list-win-size)
		  (select-window srw)
		  (xref-set-this-window-size xref-win3-sym-res-win-size)
		  ))
))

(defun xref-reconstitute-windows (wins)
  (let ((www))
	(setq www wins)
	(while www
	  (if (eq (car www) 'xref-symbol-resolution)
		  (xref-top-refs-resolution (point))
		)
	  (if (eq (car www) 'xref-list)
		  (xref-list-top-references)
		)
	  (if (eq (car www) 'xref-class-tree)
		  (xref-show-class-tree)
		)
	  (setq www (cdr www))
	  )
	(xref-resize3-windows)
))

(defun xref-show-stack-symbol-and-references ()
"Show in a separate  window all symbols on  the top of the reference stack. 

It also shows all references pushed on the top of the stack in another
window."
  (interactive "")
  (xref-delete-windows)
  (xref-reconstitute-windows (list 'xref-symbol-resolution 'xref-list))
)

(defun xref-delete-window (buffername)
  (let ((displayed (get-buffer-window buffername nil)))
    (if displayed
        (progn
          (delete-window displayed)
		  (kill-buffer buffername)
        )
    )
  )
)

(defun xref-soft-delete-window (buffername)
  (let ((displayed (get-buffer-window buffername nil)))
    (if displayed
		(delete-window displayed)
    )
  )
)



(defun delete-completion-window ()
  "Delete some Emacs/XEmacs windows created by Xrefactory."
  (interactive "")
  (xref-delete-window  (xref-list-buffer-name))
  (xref-delete-window  " *xrefsrv*")
  (xref-delete-window  " *xref-list*")
  (xref-delete-window  " *xref-execute*")
  (xref-soft-delete-window  "*background-1*")
)


(defun xref-delete-windows ()
  "Delete Emacs/XEmacs windows created by Xrefactory."
  (interactive "")
  (xref-get-windows-status)
  (delete-completion-window)
  (xref-delete-window  (xref-list-buffer-name))
  (xref-delete-window  (xref-symbol-resolution-buffer-name))
  (xref-soft-delete-window  (xref-class-tree-buffer-name))
  (xref-soft-delete-window  "*xref-run*")
  (xref-soft-delete-window  "*compilation*")
  (delete-other-windows)
  (message "Done.")
)

(defun write-xref-tmp-buff (tmpFile from-p to-p)
  (let ((inhib-o t))
	(if (boundp 'inhibit-eol-conversion)
		(progn
		  (setq inhib-o inhibit-eol-conversion)
		  (setq inhibit-eol-conversion t)
		  ))
	(write-region from-p to-p tmpFile nil 'no-message)
	(if (boundp 'inhibit-eol-conversion)
		(setq inhibit-eol-conversion inhib-o)
	  )
	)
)

(defun xref-execute-command-is-complete ()
  (let ((res) (marker))
	(save-excursion
	  (set-buffer " *xref-execute*")
	  (beginning-of-buffer)
	  (end-of-line)
	  (setq marker (buffer-substring 1 (point)))
	  (setq res (search-forward marker nil t))
	  )
	res
))

(defun xref-execute-elisp-buffer ()
  (let ((mpos))
	(set-buffer " *xref-execute*")
	(setq mpos (xref-execute-command-is-complete))
	(goto-char mpos)
	(beginning-of-line)
	(kill-line)
	(beginning-of-buffer)
	(kill-line)
	(eval-buffer " *xref-execute*" nil)
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar xref-xemacs-frame-counter 0)

(defun xref-get-frame-id ()
  (let ((res nil))
	(if (eq xref-running-under 'xemacs)
		(progn
		  (setq res (frame-property (selected-frame) 'window-id nil))
		  (if (not res)
			  (progn
				;; O.K. windows xemacs still does not implement frame id, 
				;; do it via custom user properties
				(setq res (frame-property (selected-frame) 'xref-frame-number nil))
				(if (not res)
					(progn
					  (setq xref-xemacs-frame-counter (+ xref-xemacs-frame-counter 1))
					  (set-frame-property (selected-frame) 
										  'xref-frame-number 
										  (format "%d" xref-xemacs-frame-counter)
										  )
					  (setq res (format "%d" xref-xemacs-frame-counter))
					  )))))
	  ;; emacs, take frame id
	  (setq res (cdr (assq 'window-id (frame-parameters (selected-frame)))))
	  )
	(if (not res) (setq res ""))
	res
))

(defun xref-user-frame-id ()
  (let ((res))
	(setq res (format "%s%s" xref-user-identification (xref-get-frame-id)))
	res
))

(defun call-completion (sync save-flag poin mark callString arg1 arg2 arg3)
  ;; backward compatibility function, do not use it!
  (xref-call-process sync save-flag nil poin mark callString arg1 arg2 arg3)
)

(defun xref-call-process (sync save-flag file-name poin mark callString arg1 arg2 arg3)
  (let ((originBuff (buffer-name))
	    (originPos (point))
		(olcursor (- poin 1)) (olmark (- mark 1))
		(original-size (+ (buffer-size) 1))
		(tmpFile xref-communication-file)
		(multibyte (and xref-allow-multibyte
						(boundp 'enable-multibyte-characters) 
						(not (eq nil enable-multibyte-characters))))
		(whole-output-string)
	   )
  (if (eq file-name nil)
	  (setq file-name (buffer-file-name))
	)
  (if (eq file-name nil)
	  (setq file-name originBuff)
	)
  (if (eq xrefsrv-process nil)
	  (start-xrefsrv-proc)
	(if (eq (process-status xrefsrv-process) 'run)
		()
	  (delete-process xrefsrv-process)
	  (start-xrefsrv-proc)
	)
  )
  (setq xrefsrv-buffer originBuff)
  (if (eq save-flag 0) 
	  ()
	(if multibyte
		(progn
		  (if (> poin 1)
			  (progn
				(write-xref-tmp-buff tmpFile 1 poin)
				(setq olcursor (nth 7 (file-attributes tmpFile)))
				))
		  (if (> mark 1)
			  (progn
				(write-xref-tmp-buff tmpFile 1 mark)
				(setq olmark (nth 7 (file-attributes tmpFile)))
				))
		  )
	  )
	(write-xref-tmp-buff tmpFile 1 (+ (buffer-size) 1))
	)
  (get-buffer-create " CompletionBuffer")
  (set-buffer " CompletionBuffer")
  (xref-erase-buffer)
  (if (eq save-flag 0) 
	  ()
	(insert tmpFile)
	(newline)
	)
  (if xref-debug-mode
	  (progn
		(insert "-debug")
		(newline)	  
		))
  (if (eq file-name nil) 
	  ()
	(insert "-file")
	(newline)	  
	(insert-char ?" 1) (insert file-name) (insert-char ?" 1) 
	(newline)
	(insert (format "-olcursor=%d" olcursor))
	(newline)
	(insert (format "-olmark=%d" olmark))
	(newline)
;;	(insert "-olnodialog")
;;	(newline)
	(if xref-next-push-manual
		(progn
		  (insert "-olmanualresolve")
		  (newline)
		  (setq xref-next-push-manual nil)
		  ))
	)
  (insert "-user ")
  (insert (format "\"%s\"" (xref-user-frame-id)))
  (newline)
  (insert "-emacs ")
  (newline)
  (if xref-browse-url-manual-redirect
	  (insert "-urlmanualredirect \n")
	)
  (if xref-java-fqt-name-completion
	  (insert "-olfqtnametocompletions \n")
	)
;;  (insert (format "-set file \"%s\"" (backslashify-name file-name)))
;;  (newline)
;;  (insert (format "-set dir \"%s\"" (backslashify-name (xref-file-directory-name file-name))))
;;  (newline)
  (if xref-buffers-truncate-lines
	  (insert "-olinelen=50000")
	(insert (format "-olinelen=%d" (window-width)))
	)
  (newline)
  (if xref-current-project
	  (progn
		(insert (format "-p \"%s\"" xref-current-project))
		(newline)
		))
  (insert arg1)
  (newline)	  
  (insert arg2)
  (newline)	  
  (insert arg3)
  (newline)	  
  (if (eq sync 'sync)
	  (progn
		(insert (format "-last_message=%s" xref-sync-last-message))
		(newline)	  
		(setq xref-last-message-filtered nil)
		(setq xref-filter-whole-output "")
		)
	(if (eq sync 'syncfile)
		(progn
		  (insert (format "-last_message=*%s" xref-sync-last-message))
		  (newline)
		  (insert (format "-o %s" xref-out-communication-file))
		  (newline)
		  (setq xref-last-message-filtered nil)
		  (setq xref-filter-whole-output "")
		  )
	  (setq xref-last-message-filtered 0)
	))
  (newline)	  
  (insert "end-of-options")
  (newline)	  
  (newline)	  
  (setq xref-filter-initialized nil)
  (setq xref-last-referencing-action 'xref-process-called)
;;(message "%s" (buffer-substring 1 (point)))
  (process-send-region xrefsrv-process 1 (point))
;;  (xref-erase-buffer)
  (kill-buffer nil)
  (get-buffer-create originBuff)
  (set-buffer originBuff)
  (message "")
  (accept-process-output xrefsrv-process)
  (if (or (eq sync 'sync) (eq sync 'syncfile))
	  (progn
		;; synchronous dialog, wait until the end of answer :( or coredump)
		;; this may block, if xref does not send synchronisation string from
		;; some strange reason.
		(while (and (eq xref-last-message-filtered nil) 
					(eq (process-status xrefsrv-process) 'run))
		  (accept-process-output xrefsrv-process)
		  )
		(if (eq sync 'syncfile)
			(progn
			  ;; if synchro via file, read a file into a string, and apply filter
			  (get-buffer-create " CompletionBuffer")
			  (set-buffer " CompletionBuffer")
			  (xref-erase-buffer)
			  (insert-file-contents xref-out-communication-file)
			  (setq whole-output-string (buffer-substring (point-min) (point-max)))
			  (kill-buffer " CompletionBuffer")
			  (delete-file xref-out-communication-file)
			  (setq xref-filter-initialized nil)
			  (setq xref-last-message-filtered 0)
			  (setq xref-last-referencing-action 'xref-process-called)
			  (xrefsrv-filter xrefsrv-process whole-output-string)
			  )
		  ;; normal synchro operation, check for pathological exit
		  (if (not (eq (process-status xrefsrv-process) 'run))
			  (progn
				;; try to process at least what you have
				(xrefsrv-filter xrefsrv-process (concat "![xref.el] error, server process exited\n"
														xref-sync-last-message))
				(setq xref-pilot-char ?\!)
				))
		  )
		)
	;; assync, but even here you have to wait at least until pilot-char is set
	;; berk! I hope there is no more any assync communication
	(accept-process-output xrefsrv-process .1)
	)
  (if (eq xref-pilot-char ?\!)
	  (progn
		;; unexpected error, wait before delete the communication file
		(sit-for 0.1)
		(while (accept-process-output xrefsrv-process 1)
		  (sit-for 0.1)
		  )
		(error "Execution aborted.")
		))
  (if (eq xref-pilot-char ?\')
	  (while (not (xref-execute-command-is-complete))
		(accept-process-output xrefsrv-process)
		)
	)
;;  the following can cause problem, because 'xrefsrv' in debug
;;  can output some message before processing input file, so it passes
;;  the 'accept-process-output' and deletes input, before opening it.
  (if (eq save-flag 0) 
	  ()
	(xref-delete-communication-file)
	)
  (if (eq xref-pilot-char ?\')
	  (xref-execute-elisp-buffer)
	)
  (set-buffer originBuff)
))

(defun xref-delete-communication-file ()
  (if (not xref-debug-mode)
	  (delete-file xref-communication-file)
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-add-to-imports (new-import)
  (interactive "s")
  (let ((iline) (iline-val))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
		"You should edit a Java file when using this function.")
	(xref-trivial-pre-checks "-getlastimportline")
	(setq iline xref-last-message)
	(message "")
	(setq iline-val (string-to-int iline))
	(save-excursion
	  (goto-line (+ iline-val 1))
	  (beginning-of-line)
	  (insert (format "import %s;\n" new-import))
	  (sit-for 1)
	  )
))

(defun xref-general-import-from-type-name (ss)
  (let ((pp) (res) (loop))
	(setq pp (length ss))
	(setq loop 0)
	(while loop
	  (setq loop (string-match "\\." ss (+ loop 1)))
	  (if loop (setq pp loop))
	  )
	(setq res (format "%s.*" (substring ss 0 pp)))
	res
))

(defun xref-completion-add-import (import-on-demand)
  (let ((lineno 1) (replace-flag) (isfqt) (eol) (istr))
	(if (eq (point) (point-max))
		(setq replace-flag nil)
	  (setq replace-flag t)
	  (setq lineno (count-lines 1 (+ (point) 1)))
	  (save-excursion
		(end-of-line)
		(setq eol (point))
		(beginning-of-line)
		(setq isfqt (search-forward "fully qualified" eol t))
		(if (not isfqt)
			(progn
			  (error "This function is available only for fully qualified type name completions. ")
			  ))
		))
	(other-window -1)
	(if replace-flag
		(progn
		  (xref-call-process 'sync 0 nil 0 0 "" (format "-olcomplselect%d" lineno)
							 "-olcxrefsuffix=0:#" "")
		  (save-excursion (xref-moveref t))      ;; burk, I need move to set refid
		  (if import-on-demand 
			  (setq istr (xref-general-import-from-type-name xref-move-refid))
			(setq istr xref-move-refid)
			)
		  (xref-add-to-imports (substring istr 1))
		  (xref-delete-windows)
		  ))
	(message " ")
	t
))

(defun interactive-xref-completion-add-import (&optional argp)
"Add fully qualified type name to imports. 

This  function is available  only for  completions of  fully qualified
type names in Java."
  (interactive "P")
  (xref-completion-add-import nil)
)

(defun interactive-xref-completion-mouse-add-import (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-add-import)
  t
)

(defun interactive-xref-completion-xemacs3b-add-import (event)
  (interactive "e")
  (interactive-xref-completion-add-import)
  (other-window -1)
  t
)


(defun interactive-xref-completion-add-general-import (&optional argp)
"Add general 'import-on-demand' import. 

This function takes  the fully qualified type name,  replaces its last
part by * and puts the result to imports.

This  function is available  only for  completions of  fully qualified
type names in Java."
  (interactive "P")
  (xref-completion-add-import t)
)

(defun interactive-xref-completion-mouse-add-general-import (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-add-general-import)
  t
)

(defun interactive-xref-completion-xemacs3b-add-general-import (event)
  (interactive "e")
  (interactive-xref-completion-add-general-import)
  (other-window -1)
  t
)


;;;;;;;;;;;;;;;;;;;;

(defun interactive-xref-completion-select (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((lineno 1) (replace-flag))
	(if (eq (point) (point-max))
		(setq replace-flag nil)
	  (setq replace-flag t)
	  (setq lineno (count-lines 1 (+ (point) 1)))
	  )
	(other-window -1)
	(xref-call-process 'sync 0 nil 0 0 "" (format "-olcomplselect%d" lineno)
					 "-olcxrefsuffix=0:#" "")
	(xref-moveref t)
	(if replace-flag
		(progn
		  (insert (substring xref-move-refid 1))
		  (xref-delete-pending-ident)
		  ))
	(xref-delete-windows)
	(message " ")
	t
))

(defun interactive-xref-completion-mouse-select (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-select)
  t
)

(defun interactive-xref-completion-escape (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((lineno (count-lines 1 (+ (point) 1))))
	(other-window -1)
	(xref-call-process 'sync 0 nil 0 0 "" (format "-olcomplselect%d" lineno)
					 "-olcxrefsuffix=0:#" "")
	(xref-moveref t)
	(xref-delete-windows)
	(message " ")
	t
))

(defun interactive-xref-completion-mouse-escape (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-escape)
  t
)

(defun interactive-xref-completion-goto (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((lineno (count-lines 1 (+ (point) 1))))
	(other-window -1)
	(xref-call-process 'sync 0 nil 0 0 "" (format "-olcxcgoto%d" lineno) 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\~)
		(xref-browse-tmp-file-url (get-url-to-browse))
	  (xref-moveref t)
	  (sit-for 1)
	  )
;;	(put-text-property (point) (mark t) 'face 'highlight)
	(other-window 1)
  )
  t
)

(defun interactive-xref-completion-mouse-goto (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-completion-goto)
  (other-window -1)
  t
)

(defun interactive-xref-completion-xemacs3b-goto (event)
  (interactive "e")
  (interactive-xref-completion-goto)
  (other-window -1)
  t
)


(defun get-url-to-browse ()
  (let ((url "index.html"))
  (save-excursion
	(set-buffer " *xrefsrv*")
	(goto-char (point-min))
	(next-line 1)
	(setq url (buffer-substring 1 (- (point) 1)))
	)
;; following is because Netscape (as well as IE?) can't parse
;; urls with ')' and ',' while using remote control (IE also #?)
  (if (string-match "[,)#]" url)
	  (setq url (substring url 0 (string-match "#" url)))
	)
  url
))

(defun xref-sit-for-no-redisplay (delay)
  (if (eq xref-running-under 'xemacs)
	  (sit-for delay t)
	(sit-for delay nil t)
	)
)

(defun xref-browse-tmp-file-url (url)
  (let ((ff))
	(setq ff (selected-frame))
	(browse-url url)
	;; try to get focus as soon as possible
	(if (> xref-browse-url-focus-delay 0) 
		(progn
		  ;; first sleep and discard windows generated events
		  (sleep-for .05)
		  (discard-input)
		  (xref-sit-for-no-redisplay xref-browse-url-focus-delay)
		  (raise-frame ff)
		  ))
;;  (delete-file url)
))

(defun xref-completion-browse (javadoc-only)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((url "index.html") (lineno (count-lines 1 (+ (point) 1))))
	(other-window -1)
	(xref-call-process 'sync 0 nil 0 0 "" (format "-olcxcbrowse%d" lineno) 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\~)
		  (xref-browse-tmp-file-url (get-url-to-browse))
	  (if (not javadoc-only)
		  (progn
			;; (sit-for 0.5)
			(other-window 1)
			(interactive-xref-completion-goto)
			(other-window -1)
			)
	  ))
	(other-window 1)
  )
  t
)

(defun interactive-xref-completion-browse (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (xref-completion-browse nil)
)

(defun interactive-xref-completion-mouse-browse (event)
  (interactive "e")
  (mouse-set-point event)
  (xref-completion-browse nil)
  (other-window -1)
  t
)

(defun interactive-xref-completion-javadoc (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (xref-completion-browse t)
)

(defun interactive-xref-completion-menu3b-javadoc (event)
  (interactive "e")
  (mouse-set-point event)
  (xref-completion-browse t)
  (other-window -1)
  t
)

(defun interactive-xref-completion-xemacs3b-javadoc (event)
  (interactive "e")
  (xref-completion-browse t)
  (other-window -1)
  t
)

(defvar xref-compl-3bmenu (make-sparse-keymap "Completions Menu"))
(fset 'xref-compl-3bmenu (symbol-value 'xref-compl-3bmenu))
(define-key xref-compl-3bmenu [xref-compl-3bclose] '("Close" . xref-delete-windows))
(define-key xref-compl-3bmenu [xref-compl-3bexit] '("Close and Return" . interactive-xref-completion-escape))
(define-key xref-compl-3bmenu [xref-compl-3baddgimp] '("Add General Import" . interactive-xref-completion-mouse-add-general-import))
(define-key xref-compl-3bmenu [xref-compl-3baddimp] '("Add to Imports" . interactive-xref-completion-mouse-add-import))
(define-key xref-compl-3bmenu [xref-compl-3bselect] '("Complete" . interactive-xref-completion-mouse-select))
(define-key xref-compl-3bmenu [xref-compl-3binspect] '("Inspect Definition" . interactive-xref-completion-mouse-goto))
(define-key xref-compl-3bmenu [xref-compl-3bjdoc] '("JavaDoc" . interactive-xref-completion-menu3b-javadoc))

(defvar xref-xemacs-compl-3bmenu
  '("Completions Menu"
   ["JavaDoc"      interactive-xref-completion-xemacs3b-javadoc t]
   ["Inspect Definition"      interactive-xref-completion-xemacs3b-goto t]
   ["Complete"       interactive-xref-completion-select t]
   ["Add to Imports"      interactive-xref-completion-xemacs3b-add-import t]
   ["Add General Import"      interactive-xref-completion-xemacs3b-add-general-import t]
   ["Close and Return"       interactive-xref-completion-escape t]
   ["Close"       xref-delete-windows t]
	) "Xref button3 popup menu" 
)

(defun xref-popup-xemacs-completion-menu (event)
  (interactive "e")
  (select-window (get-buffer-window (xref-list-buffer-name)))
  (goto-char (event-closest-point event))
  (beginning-of-line)
  (popup-menu xref-xemacs-compl-3bmenu)
)

(defvar xref-completion-auto-search-list nil)

(defun xref-auto-search-search (str)
  (let ((sss) (ns))
	(setq sss (format "\\(^\\|^[^ ]*\\.\\)%s" str))
	(search-forward-regexp sss)
;;	(setq ns str)
	(setq ns (buffer-substring (match-beginning 0) (match-end 0)))
	(set-text-properties 0 (length ns) nil ns)
	(setq xref-completion-auto-search-list 
		  (cons
		   ns
		   (cons (point) xref-completion-auto-search-list)))
))

(defun xref-completion-auto-source-mod (action)
  (let ((cb) (sb))
	(setq cb (get-buffer-window (current-buffer)))
	(setq sb (get-buffer-window (marker-buffer xref-completion-marker)))
	(if sb 
		(select-window sb)
	  (other-window -1)
	  (xref-switch-to-marker xref-completion-marker)
	  )
	(xref-set-to-marker xref-completion-marker)
	(eval action)
	(set-marker xref-completion-marker (point))
	(select-window cb)
	)
)

(defun xref-completion-auto-search-w ()
  (interactive "")
  (let ((pp) (lend) (ns) (addstr))
	(setq pp (point))
	(end-of-line)
	(setq lend (point))
	(goto-char (+ pp 1))
	(search-forward-regexp xref-forward-pass-identifier-regexp lend t)
	(setq addstr (buffer-substring pp (- (point) 1)))
	(setq ns (format "%s%s" (car xref-completion-auto-search-list) addstr))
	(beginning-of-line)
	(xref-auto-search-search ns)
	(xref-completion-auto-source-mod (list 'insert addstr))
))

(defun xref-completion-auto-search-s ()
  (interactive "")
  (xref-auto-search-search (car xref-completion-auto-search-list))
)

(defun xref-completion-auto-search-back ()
  (interactive "")
  (let ((op) (cstr) (pstr))
  (if (> (length xref-completion-auto-search-list) 2)
	  (progn
		(setq pstr (car xref-completion-auto-search-list))
		(setq xref-completion-auto-search-list 
			  (cdr (cdr xref-completion-auto-search-list )))
		(setq cstr (car xref-completion-auto-search-list))
		(setq op (car (cdr xref-completion-auto-search-list)))
		(goto-char op)
		(xref-completion-auto-source-mod (list 'backward-delete-char (- (length pstr) (length cstr))))
		))
))

(defun xref-completion-auto-search ()
  (interactive "")
  (let ((ns) (os) (nns))
	(setq os (car xref-completion-auto-search-list))
	(setq ns (format "%s%c" os last-command-char))
	(beginning-of-line)
	(xref-auto-search-search ns)
	(setq nns (car xref-completion-auto-search-list))	
	(xref-completion-auto-source-mod (list 'insert (substring nns -1)))
))

(defun xref-completion-auto-switch ()
  (interactive "")
  (other-window -1)
  (self-insert-command 1)
)

(defvar xref-tags-3bmenu (make-sparse-keymap "Tag Search Menu"))
(fset 'xref-tags-3bmenu (symbol-value 'xref-tags-3bmenu))
(define-key xref-tags-3bmenu [xref-tags-3bclose] '("Close" . xref-delete-windows))
(define-key xref-tags-3bmenu [xref-tags-3bexit] '("Close and Return" . interactive-xref-completion-escape))
(define-key xref-tags-3bmenu [xref-tags-3bselect] '("Complete" . interactive-xref-completion-mouse-select))
(define-key xref-tags-3bmenu [xref-tags-3binspect] '("Inspect" . interactive-xref-completion-mouse-goto))

(defvar xref-xemacs-tags-3bmenu
  '("Tag Search Menu"
   ["Inspect"      interactive-xref-completion-xemacs3b-goto t]
   ["Complete"       interactive-xref-completion-select t]
   ["Close and Return"       interactive-xref-completion-escape t]
   ["Close"       xref-delete-windows t]
	) "Xref button3 popup menu" 
)

(defun xref-popup-xemacs-tag-selection-menu (event)
  (interactive "e")
  (select-window (get-buffer-window (xref-list-buffer-name)))
  (goto-char (event-closest-point event))
  (beginning-of-line)
  (popup-menu xref-xemacs-tags-3bmenu)
)

(defun xref-fontify-list (compl symbol)
  (let ((list-keywords))
;;	(font-lock-mode 1)
;;	(make-local-variable 'font-lock-defaults)
;;	(if compl
;;		(setq font-lock-defaults '(xref-font-lock-compl-keywords nil nil))
;;	  (setq list-keywords (xref-font-lock-list-keywords symbol))
;;	  (setq font-lock-defaults '(list-keywords nil nil))
;;	  )
;;	(font-lock-fontify-buffer)
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-should-auto-switch-to-completions ()
  (let ((res))
	(setq res xref-auto-switch-to-completion-window)
	(if res
		(progn
		  (beginning-of-buffer)
		  (setq res (search-forward-regexp "^[a-zA-Z_]" nil t 2))
		  ))
	res
))


(defun xref-completion ()
"Complete  identifier.   

This  function parses  the  currently edited  buffer  until the  point
position.  Then if the identifier on point have a unique completion it
completes  it. If there  is a  list of  possible completions  they are
displayed in a separate window.
"
  (interactive "")
  (let ((copt "") (cwin) (cid) (poin))
	(if (equal (buffer-name) (xref-list-buffer-name))
		(xref-delete-windows)
	  )
	  ;; repeated invocation? if yes, completion will scroll down
	(if (and (eq last-command 'xref-completion) 
			 (eq xref-pilot-char ?\;)
			 (get-buffer-window (xref-list-buffer-name)))
		(if (equal (get-buffer-window (current-buffer)) 
				   (get-buffer-window (xref-list-buffer-name)))
			(scroll-up nil)
		  (scroll-other-window nil)
		  )
	  (if (eq (buffer-file-name) nil)
		  (message "You should edit a C or Java source file when using Xref")
		(setq poin (point))
		(if xref-java-completion-linkage-check
			(setq copt "-olchecklinkage")
		  )
		(if (not xref-java-completion-access-check)
			(setq copt (concat copt " -olnocheckaccess"))
		  )
		(setq xref-fontification xref-font-lock-compl-keywords)
		(setq cid (xref-get-identifier-before-point))
		(setq xref-completion-auto-search-list (list cid (+ (length cid) 1)))
		(set-marker xref-completion-marker (point))
		(xref-call-process xref-reflist-synchro 1 nil poin 0 "" 
						   (format "-olcxcomplet -maxCompls=%d " xref-max-completions) 
						   copt "-olcxrefsuffix=0:#")
		(if (eq xref-pilot-char ?\;)
			(progn
			  (set-buffer (xref-list-buffer-name))
			  (if (xref-should-auto-switch-to-completions)
				  (progn
					(select-window (get-buffer-window (xref-list-buffer-name)))
					(xref-auto-search-search cid)
					))
			  (xref-display-buffer (xref-list-buffer-name))
			  (use-local-map xref-completion-mode-map)
			  (message "button2 (<RET>) - select;  S-button2 ('C-b') - Browse;")
			  )
		  )
		(if (eq xref-pilot-char ?\:)
			(progn
			  (set-buffer " *xrefsrv*")
			  (if (xref-should-auto-switch-to-completions)
				  (progn
					(select-window (get-buffer-window (xref-list-buffer-name)))
					(xref-auto-search-search cid)
					))
			  (xref-display-buffer " *xrefsrv*")
			  (use-local-map xref-completion-mode-map)
			  (message "button2 (<RET>) - select;  S-button2 ('C-b') - Browse;")
			  )
		  )
	  ))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-search-in-completions (poin)
"Search  string(s)  in  completions.   

This function first  asks for string(s) to search,  then it parses the
currently edited buffer until the POIN position.  Then it inspects all
contextually adequate  completions and displays  only those containing
the given string(s).
"
  (interactive "d")
  (let ((sstr))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (setq sstr (xref-get-search-string xref-allow-completion-in-tag-search-prompt nil))
	  (setq xref-fontification xref-font-lock-compl-keywords)
	  (setq xref-completion-auto-search-list '("" 0))
	  (xref-call-process 'sync 1 nil poin 0 "C@$./@mVeDitznACc" 
						 "-olcxsearch" "-olcxrefsuffix=0:#"
						 (concat "\"-olcxcplsearch=" sstr "\"")
						 )
	  (if (eq xref-pilot-char ?\;)
		  (progn
			(set-buffer (xref-list-buffer-name))
			(xref-display-buffer (xref-list-buffer-name))
			(use-local-map xref-completion-mode-map)
			(message "button2 (<RET>) - select;  S-button2 ('C-b') - Browse;")
			;;	(xref-delete-communication-file)
			))
	  )
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun create-escaped-string (sstr)
  (let ((ind) (maxi) (ress) (cc))
	(setq ress "")
	(setq ind 0)
	(setq maxi (length sstr))
	(while (< ind maxi)
	  (setq cc (substring sstr ind (+ ind 1)))
	  (if (string-match "[A-Za-z_]" cc)
		  (setq ress (concat ress cc))
		(setq ress (concat ress "\\" cc))
		)
	  (setq ind (+ ind 1))
	  )
	ress
))

(defun create-xref-tag-search-fontif (sstr)
  (let ((res) (ress) (loop) (ind) (ind0) (prefix))
	(setq prefix "\\(")
	(setq loop t)
	(setq ind 0)
	(while loop
	  (setq ind0 (string-match "[^ \n\t]" sstr ind)) 
	  (if (not ind0)
		  (setq loop nil)
		(setq ind (string-match "[ \t\n]" sstr ind0))
		(if (not ind)
			(setq loop nil)
		  (setq ress (concat ress prefix (create-escaped-string (substring sstr ind0 ind))))
		  (setq prefix "\\|")
		  ))
	)
	(if ind0
		(setq ress (concat ress prefix (create-escaped-string (substring sstr ind0))))
	  )
	(setq ress (concat ress "\\)"))
	(setq res 
		  (cons (cons ress '(xref-list-symbol-face))
		  (cons (cons ":.*$" '(xref-list-default-face))
				nil))
		  )
	res
))

(defun xref-get-identifier-before-point ()
  (let ((sym) (poin))
	(setq poin (point))
	(search-backward-regexp xref-backward-pass-identifier-regexp)
	(forward-char 1)
	(setq sym (buffer-substring (point) poin))
	(goto-char poin)
	sym
))

(defun xref-get-identifier-on-point ()
  (let ((sym) (poin))
	(setq poin (point))
	(search-backward-regexp xref-backward-pass-identifier-regexp)
	(forward-char 1)
	(setq sym (xref-identifier-after-point))
	(goto-char poin)
	sym
))

(defun xref-get-all-tags (searchopt)
  (let ((res) (loop) (b))
	(save-excursion
	  (xref-call-general-search-in-tag-file "" searchopt)
	  (set-buffer (xref-list-buffer-name))
	  (setq res nil)
	  (beginning-of-buffer)
	  (setq loop (not (eq (point) (point-max))))
	  (while loop
		(setq b (point))
		(end-of-line)
		(setq res (cons (buffer-substring b (point)) res))
		(setq loop (not (eq (point) (point-max))))
		(if loop
			(progn
			  (beginning-of-line)
			  (next-line 1)
			  (setq loop (not (eq (point) (point-max))))
			  )))
	  (xref-delete-window (xref-list-buffer-name))
	  )
	res
))

(defun xref-get-search-string (allowc fullscan) 
  (let ((sstr) (sym) (poin) (table))
	(setq sym (xref-get-identifier-on-point))
	(if allowc
		(progn
		  (if fullscan
			  (setq table (xref-get-list-for-completions (xref-get-all-tags "-searchshortlist")))
			(setq table (xref-get-list-for-completions (xref-get-all-tags "-searchdefshortlist")))
			)
		  (setq sstr (completing-read (format "String(s) to search [%s]:" sym) table nil nil sym))
		  )
	  (setq sstr (read-from-minibuffer (format "String(s) to search [%s]:" sym)))
	  )
;;(error "'%s'" sstr)
	(if (equal sstr "")
		(setq sstr sym)
	  )
	sstr
))

(defun xref-current-column()
  (let ((poin) (res))
	(setq poin (point))
	(beginning-of-line)
	(setq res (- poin (point)))
	(goto-char poin)
	res
))

(defun xref-switch-to-marker (marker)
  (switch-to-buffer (marker-buffer marker))
  (goto-char (marker-position marker))
)

(defun xref-set-to-marker (marker)
  (set-buffer (marker-buffer marker))
  (goto-char (marker-position marker))
)

(defun xref-call-general-search-in-tag-file (sstr sup-opt)
  (let ((line) (col))
	(setq xref-fontification (create-xref-tag-search-fontif sstr))
	(search-backward-regexp xref-backward-pass-identifier-regexp)
	(forward-char 1)
	(setq line (count-lines (point-min) (+ (point) 1)))
	(setq col (xref-current-column))
	(xref-call-process xref-reflist-synchro 1 nil (point) 0 ""
					   (format "%s -olcxlccursor=%d:%d" sup-opt line col)
					   (format "\"-olcxtagsearch=%s\"" sstr)
					   "-olcxrefsuffix=0:#")
))

(defun xref-general-search-in-tag-file (sstr sup-opt)
  (xref-call-general-search-in-tag-file sstr sup-opt)
  (set-buffer (xref-list-buffer-name))
  (xref-display-buffer (xref-list-buffer-name))
  (use-local-map xref-tag-search-mode-map)
  (message xref-list-buffer-message)
  (select-window (get-buffer-window (xref-list-buffer-name)))
  ;;	  (xref-delete-communication-file)
)

(defun xref-search-in-tag-file (poin)
"Search string(s) in the tag  file.  

This function  asks for strings to  search, then it  inspects the Xref
tag  file looking for  symbols containing  given string(s).   For each
such symbol it  displays the symbol name together  with its definition
place. 

When looking  for a symbol the full  symbol name as it  appears in Tag
file is checked to match searched string. This means that also strings
appearing in  methods profiles and  in fully qualified type  names can
successfully match.

See also: `xref-search-definition-in-tag-file',
          `xref-search-in-foreign-tag-file'.
"
  (interactive "d")
  (let ((sstr))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (setq sstr (xref-get-search-string xref-allow-completion-in-tag-search-prompt t))
	  (xref-general-search-in-tag-file sstr "")
)))

(defun xref-search-definition-in-tag-file (poin)
"Search the tag  file for definition of a symbol.

This function  asks for strings to  search, then it  inspects the Xref
tag  file looking for  symbols containing  given string(s).   For each
such symbol beeing  defined (having source code) in  active project it
displays the symbol name together with its definition place. 

There  are two main  differences to  `xref-search-in-tag-file'. First,
only symbols  having definition in  the project are  reported. Second,
only symbol names are matched  against the string to search, excluding
profiles for methods and package names for classes.

See also: `xref-search-in-tag-file',
          `xref-search-in-foreign-tag-file'.
"
  (interactive "d")
  (let ((sstr))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (setq sstr (xref-get-search-string xref-allow-completion-in-tag-search-prompt nil))
	  (xref-general-search-in-tag-file sstr "-searchdef")
)))


(defun xref-get-list-for-completions (ppp)
  (let ((res) (prs))
	(setq res nil)
	(while ppp
	  (progn
		(setq res (cons (cons (car ppp) nil) res))
		(setq ppp (cdr ppp))
		))
	res
))

(defun xref-get-project-list-completions ()
  (let ((res))
	(setq res (xref-get-list-for-completions (xref-get-project-list)))
	res
))


(defun xref-search-in-foreign-tag-file (poin)
"Search string(s) in a tag file of another project.  

This function asks for strings to  search and for the project where to
look for Tag  file, then it inspects the tag  file looking for symbols
containing  given string(s).   For each  such symbol  it  displays the
symbol  name together  with  its definition  place.   

The main  usage of this function  is when you are  using unknown large
libraries  (usually as  .jar files)  and you  do want  to  index those
libraries into your  current project as it would  make updating of tag
file slow.   In this  case you can  create one project  containing all
those  libraries and  another (working)  project containing  only your
sources.   Then  when  needing  to  retrieve a  symbol  (method)  from
large libraries you can use this function.

See also: `xref-search-definition-in-tag-file',
          `xref-search-in-tag-file'.
"
  (interactive "d")
  (let ((foreign-project) (fp) (prs) (sstr) (cpl) (ppp))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (setq sstr (xref-get-search-string xref-allow-completion-in-tag-search-prompt t))
	  (xref-delete-windows)
	  (setq fp nil)
;;	  (setq fp (xref-get-env "foreign-project"))
;;	  (if (not fp) (xref-delete-windows))   ;; clear error message?
	  (if xref-last-foreign-project
		  (setq fp xref-last-foreign-project)
		)
;;	  (setq prs (xref-get-project-list))
;;	  (setq foreign-project 
;;			(read-from-minibuffer "Scan Tag file of project: " 
;;								  fp nil nil 'prs))
	  (setq cpl (xref-get-project-list-completions))
	  (setq foreign-project
			(completing-read "Scan Tag file of project: " cpl nil nil fp))
	  (setq xref-last-foreign-project foreign-project)
	  (xref-general-search-in-tag-file sstr (format "-p %s" foreign-project))
)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-push-name ()
"Push references of symbol of given name on the reference stack. 

This function asks for symbol name, then it inspects the Xref tag file
looking for symbols of given name.   For each such symbol it loads all
its references.

If there  are several symbols of the  same name the user  is asked for
manual  selection.   The selection  menu  is  the  same as  for  other
browsing functions, including class hierarchies for inherited symbols.

"
  (interactive "")
  (let ((sstr) (line) (col))
	(setq line (count-lines (point-min) (+ (point) 1)))
	(setq col (xref-current-column))
	(setq sstr (xref-get-search-string t t))
	(xref-call-process 'sync 1 nil (point) (mark) "" 
					   (format "\"-olcxpushname=%s\"" sstr)
					   (format "\"-olcxrefsuffix=%d:%s\"" 0 sstr)
					   (format "-olcxlccursor=%d:%d" line col)
					   )
	(xref-after-pushing-continuation nil '(pushing-post-action))
;;	(xref-after-pushing-continuation xref-three-windows '(listrefs-continuation))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun extract-process-address-arguments (headBeginPos headFinPos)
  (let ((addrlist nil) (sym) (poin)
		)
	(goto-char headBeginPos)
	(while (search-forward "@@" headFinPos t)
	  (delete-backward-char 2)
	  (setq addrlist (cons (point) addrlist))
	  )
	(while addrlist
	  (goto-char (car addrlist))
	  (setq sym (xref-identifier-after-point))
	  (setq poin (point))
	  (xref-map-on-all-references (buffer-file-name) poin t nil "-olcxrename"
							 nil nil
							 ''none-pre-fun ''renaming 
							 ''none-post-fun 
							 (concat "(*" sym ")") nil
							 nil
							 (list 'xref-popCxrefs poin)
							 )
	  (setq addrlist (cdr addrlist))
	  )
	)
)

(defun add-macro-line-continuations (reg-beg-pos reg-end-pos)
  (let ( (unmoved) (loopfl))
	(goto-char reg-end-pos)
	(setq loopfl t)
	(while loopfl
	  (end-of-line 1)
	  (insert "\\")
	  (setq unmoved (forward-line -1))
	  (setq loopfl (and (eq unmoved 0) (>= (point) reg-beg-pos)))
	  )
))

(defun extract-region (poin mark nname style)
  (let ((fcall) (fhead) (ftail) (fbody) (line) (cbuf) (pp) (addarg)
		(headBeginPos) (headFinPos) (reg-beg-pos) (reg-end-pos))
  (if (eq (buffer-file-name) nil)
	  (message "You should edit a C or Java source file when using Xref")
	(if (eq style 'macro)
		(setq addarg "-olexmacro")
	  (if (eq style 'address)
		  (setq addarg "-olexaddress")
		(setq addarg "")
		)
	  )
	(xref-call-process 'sync 1 nil poin mark "" (format "-olcxextract=%s" nname) addarg "")
	(if (not (eq xref-pilot-char ?\%))
		()
	  (setq cbuf (buffer-name))
	  (setq fbody (buffer-substring poin mark))
	  (if (not (eq (elt fbody (- (length fbody) 1)) ?\n))
		  (setq fbody (format "%s\n" fbody))
		)
	  (set-buffer " *xrefsrv*")
	  (goto-char (point-min))
	  (search-forward "!")
	  (search-forward "!")
	  (setq pp (point))
	  (search-forward "!")
	  (setq fcall (buffer-substring (+ pp 1) (- (point) 1)))
	  (search-forward "!")
	  (setq pp (point))
	  (search-forward "!")
	  (setq fhead (buffer-substring pp (- (point) 1)))
	  (search-forward "!")
	  (setq pp (point))
	  (search-forward "!")
	  (setq ftail (buffer-substring (+ pp 1) (- (point) 1)))
	  (setq pp (point))
	  (search-forward "!")
	  (setq line (string-to-int (buffer-substring pp (- (point) 1))))
	  (get-buffer-create " *xref-extraction*")
	  (set-buffer " *xref-extraction*")
	  (erase-buffer)
	  (insert "")
	  (insert "---------------   Xrefactory proposes following new definition:\n")
	  (insert fhead)
	  (insert "\t// Original code begin\n")
	  (insert fbody)
	  (insert "\t// Original code end\n")
	  (insert ftail)
	  (insert "\n")
	  (insert "---------------   which will be invoked by command:\n\n")
	  (insert fcall)
	  (beginning-of-buffer)
	  (display-buffer " *xref-extraction*")
	  (if (yes-or-no-p "Can I process the extraction ? ")
		  (progn 
			(set-buffer cbuf)
			(kill-region mark poin)
			(set-marker xref-tmp-mark (point))
			(insert fcall)
			;; (newline)
			(goto-line line)
			(setq headBeginPos (point))
			(insert fhead)
			(setq headFinPos (point))
			;;(if (not (eq style 'macro))
			;;	(newline)
			;;  )
			(setq reg-beg-pos (point))
			(insert fbody)
			(setq reg-end-pos (point))
			(insert ftail)
			(xref-delete-windows)
			(if (eq style 'macro)
				(add-macro-line-continuations reg-beg-pos reg-end-pos)
;;			  (extract-process-address-arguments headBeginPos headFinPos)
			  (goto-char headFinPos)
			  (if (fboundp 'c-indent-defun) 
				  (if (yes-or-no-p "Extracted. Re-indent now ? ")
					  (progn
						(c-indent-defun)
						(xref-switch-to-marker xref-tmp-mark)
						(c-indent-defun)
						)))
			  )
			(message "Done.")
			)
		(xref-delete-windows)
		(message "** No extraction **")
		))
;;	(xref-delete-communication-file)
)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-extract-region-to-function (poin mark nname)
"Extract region into new function/method with generated header.

The region  has to start and  end at statement boundaries,  so that it
can be moved  into body of new function/method.   This function parses
the   currently  edited   buffer,   makes  static   analysis  of   the
function/method containing  region between POIN  and MARK, classifying
all local variables to one of categories (in parameter, out parameter,
inout parameter, return value, local variable and not concerned), then
it  moves   the  region   between  the  MARK   and  POIN   before  the
function/method definition and  generates a head and the  tail for the
new  function/method.   The generated  function/method  will be  named
NNAME.  In case of Java language if the method requires several output
arguments a new  method class is created.  The  extraction is achieved
by inserting an invocation at the original position of the region.

In Java  language, Xrefactory is  not able to determine  whether there
are thrown  exceptions in the region.  Hence it is  unable to generate
throws clause in  the method header. You will need  to add this clause
manually if compiler generates compile error.
"
  (interactive "d
m
sName of the new function/method : ")
  (xref-multifile-undo-set-buffer-switch-point 
   (format "extraction into '%s'" nname))
  (extract-region poin mark nname 'function)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-extract-region-to-function-force-address (poin mark nname)
"Extract region into new procedure with generated header.

The region  has to start and  end at statement boundaries,  so that it
can be moved  into body of new function/method.   This function parses
the currently edited  buffer, makes static analysis of  the C function
containing  region between POIN  and MARK,  determines classes  of all
local variables,  then it moves the  region between the  MARK and POIN
before the function definition and generates  a head and the tail of a
new  function around  the text  of  the region,  so it  creates a  new
function definition. This function will  have no return value, the out
parameters will be passed as adress arguments.  The generated function
will be named NNAME.  The  extraction is achieved by inserting a NNAME
function invocation at the original position of the region.
"
  (interactive "d
m
sName of the new function: ")
  (xref-multifile-undo-set-buffer-switch-point 
      (format "extraction into '%s'" nname))
  (extract-region poin mark nname 'address)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-extract-region-to-macro (poin mark nname)
"Extract region into new macro with generated header.

The region  has to start and  end at statement boundaries,  so that it
can be moved  into body of new function/method.   This function parses
the currently edited  buffer, makes static analysis of  the C function
containing  region between POIN  and MARK,  determines classes  of all
local variables,  then it moves the  region between the  MARK and POIN
before the function definition and generates  a head and the tail of a
new macro  around the text  of the region,  so it creates a  new macro
definition.  The generated macro  will be named NNAME.  The extraction
is achieved by  inserting a NNAME invocation at  the original position
of the region.
"
  (interactive "d
m
sName of the new macro: ")
  (xref-initial-file-suffix-check xref-c-allowed-suffixes
	  "This refactoring is allowed only for C language.")
  (xref-multifile-undo-set-buffer-switch-point
   (format "extraction into '%s'" nname))
  (extract-region poin mark nname 'macro)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pass-fun-argument (num)
  (let ((deep 0)
		(nn num)
		(loop (> num 0))
		(ccc)
		)
	(while loop
	  (search-forward-regexp "[,\(\)\{\}/\"\']")
	  (setq ccc (char-after (- (point) 1)))
	  (cond
	   ((eq ccc ?\')
		(backward-char)
		(search-forward-regexp "[^\\]'"))
	   ((eq ccc ?\")
		(backward-char)
		(search-forward-regexp "[^\\]\""))
	   ((and (eq ccc ?/) (eq (char-after (point)) ?*))
		(search-forward "*/"))
	   ((and (eq ccc ?/) (eq (char-after (point)) ?/))
		(search-forward "
"))
	   ((and (eq ccc ?,) (eq deep 0))
		(setq nn (- nn 1))
		(setq loop (not (eq nn 0))))
	   ((and (eq ccc ?\)) (eq deep 0))
		(setq loop nil)
		(setq nn (- nn 1))
		(if (not (eq nn 0))
			(error "Out of arguments.")
		  ))
	   ((or (eq ccc ?\( ) (eq ccc ?\{))
		(setq deep (+ deep 1)))
	   ((or (eq ccc ?\) ) (eq ccc ?\}))
		(setq deep (- deep 1)))
	   )
	  )
	nn
))

(defun mark-fun-argument (num)
  (let ((rpoin) (res nil))
	(if (< num 1)
		(error "Argument number out of range")
	  (pass-fun-argument (- num 1))
	  (if (eq (char-after (- (point) 1)) ?\) )
		  (error "Argument number out of range")
		(setq rpoin (point))
		(pass-fun-argument 1)
		(backward-char)
		(push-mark)
		(goto-char rpoin)
		(setq res t)
		(message "Mark set at the end of the argument (C-w will delete it).")
		))
	res
))

(defun xref-java-this-or-super-name (file-name name)
  (let ((res))
	(if (and (or (equal name "this") (equal name "super"))
			 (xref-buffer-has-one-of-suffixes file-name '(".java")))
		(setq res t)
	  (setq res nil)
	  )
	res
))

(defun xref-moveref (check-id-and-go-arg)
  (let ((ffile) (fpoin) (fline) (fcoll) (fargn) (currentid) (refid) (rpos)
		(result nil))
	(if (not (eq xref-pilot-char ?#))
		()
	  (set-buffer " *xrefsrv*")
	  (if (< (buffer-size) 2)
		  ()
		(goto-char (point-min))
		(search-forward "#*+*#") 
		(setq ffile (buffer-substring 1 (- (point) 5)))
		(setq fpoin (point))
		(search-forward " ") 
		(setq fline (string-to-int (buffer-substring fpoin (- (point) 1))))
		(setq fpoin (point))
		(search-forward " ") 
		(setq fcoll (string-to-int (buffer-substring fpoin (- (point) 1))))
		(search-forward ":") 
		(setq xref-move-usage (char-after (point)))
		(forward-char 1)
		(setq fpoin (point))
		(search-forward ":") 
		(setq fargn (string-to-int (buffer-substring fpoin (- (point) 1))))
		(setq fpoin (point))
		(search-forward " ;;") 
		(setq refid (buffer-substring fpoin (- (point) 3)))
		(setq xref-move-refid refid)
		;;(message ffile) 
		;;(read-char)
		(kill-buffer nil)
		(xref-find-file ffile)
		(goto-line fline)
		(beginning-of-line)
		;;(message (format "%d %d" fline fcoll ))
		(setq rpos (+ (point) fcoll))
		(end-of-line)
		(if (> rpos (point))
			(setq rpos (point))
		  )
		(goto-char rpos)
;;(message "move-ref to %s %d %d" ffile fline fcoll)
		(setq xref-lastref-file ffile)
		(setq xref-lastref-line fline)
		(setq xref-lastref-coll fcoll)
		(setq currentid (xref-identifier-after-point))
;;		(message "Mark set at the end of the symbol.")
;;(setq transient-mark-mode t)
;;(setq deactivate-mark nil)
;;(setq mark-active t)
		(if check-id-and-go-arg
			(progn
			  (if (and (not (equal currentid  refid))
					   ;; maybe an #if-else-endif
					   (not (equal (substring refid 0 1) "#"))
					   ;; or an explicit constructor
					   (not (xref-java-this-or-super-name ffile currentid))
					   )
				  (error "Hmm. Not a '%s' symbol reference. Please, update tag file." refid)
				(setq result t)
				(if (> fargn 0)
					(progn
					  (xref-pass-to-fun-lpar)
					  (if (not (eq (char-after (- (point) 1)) ?\( ))
						  (progn
							(error "Hmm. Does not seem as a function symbol reference. The argument not found.")
							(goto-char rpos)
							(setq result nil)
							)
						(setq result (mark-fun-argument fargn))
					  ))))))))
	result
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun show-xref-filter-level ()
  (setq mode-name (format "Xref-List Filter:%d" xref-list-filter-level))
)

(defun show-xref-symbol-filter-level ()
  (setq mode-name (format "Xref-Menu Filter:%d" xref-sym-menu-filter-level))
)

(defun xref-frame-gc-rec (list)
  (let ((res))
	(if (not list)
		(setq res nil)
	  (if (or (not (listp (car list))) (not (frame-live-p (car (car list)))))
		  (setq res (xref-frame-gc-rec (cdr list)))
		(setq res (cons (car list) (xref-frame-gc-rec (cdr list))))
		))
	res
))

(defun xref-frame-garbage-collection (&optional argp)
  "Unlink any deleted linked frame."
  (interactive "P")
  (setq xref-frame-links (xref-frame-gc-rec xref-frame-links))
)

(defun interactive-xref-link-frame (&optional argp)
  "Open new frame and link it to this one."
  (interactive "P")
  (let ((ofr) (nfr) (wins) (ofru) (nfru))
	(if (not window-system)
		(if (not (yes-or-no-p 
				  "Really create new frame on  a character based terminal ?"))
			(error "Exiting, no frame created")
		  ))
	(setq ofr (selected-frame))
	(setq wins (xref-get-windows-status))
	(xref-delete-windows)
	(setq ofru (xref-user-frame-id))
	(setq nfr (make-frame))
	(select-frame nfr)
	;; following will also affect frame number under Xemacs/windows
	(setq nfru (xref-user-frame-id))
	(xref-call-process 'sync 0 nil 0 0 "" 
					   (format "-movefromuser \"%s\"" ofru)
					   "-olcxrefsuffix=0:" "")	
	(xref-reconstitute-windows wins)
	(xref-goto-source-window)
	(delete-window)
	(setq xref-frame-links (cons (cons nfr ofr) xref-frame-links))
	(xref-frame-garbage-collection)
	;; (select-frame ofr)
	(raise-frame)
))

(defun xref-linked-frame-moveref (check-id-and-go-arg)
  (let ((frame) (frame-assoc) (win))
	(setq frame (selected-frame))
	(setq frame-assoc (assoc frame xref-frame-links))
	(if frame-assoc 
		(progn
		  (select-frame (cdr frame-assoc))
		  (setq win (selected-window))
		  (xref-goto-source-window)
		  ))
	(xref-moveref check-id-and-go-arg)
	(if frame-assoc 
		(progn
		  (select-window win)
		  (select-frame frame)
		  ))
))

(defun interactive-xref-list-goto (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((poin (point)) (lastref) (frame) (frame-assoc)
		(lineno (count-lines 1 (+ (point) 1))))
	(beginning-of-buffer)
	(setq lastref (search-forward-regexp "^>" (point-max) t))
	(if lastref
		(progn
		  (delete-backward-char 1)
		  (insert " ")
		  )
	  )
	(goto-line lineno)
	(beginning-of-line)
	(delete-char 1)
	(insert ">")
	(goto-char poin)
	(other-window -1)
	(xref-call-process 'sync 0 nil 0 0 "" (format "-olcxgoto%d" lineno) 
					 "-olcxrefsuffix=0:" "")
	(xref-linked-frame-moveref t)
	(sit-for 1)
	(other-window 1)
  )
  (message xref-list-buffer-message)
  t
)

(defun next-reference-list-buffer ()
  (let ((cline) (nlines) (poin (point)) (lastref))
	(if (not (eq xref-list-buffer-content 'reference-list))
		(error "Not a reference list buffer")
	  )
	(beginning-of-buffer)
	(setq lastref (search-forward-regexp "^>" (point-max) t))
	(if (not lastref)
		(goto-char poin)
	  )
	(beginning-of-line)
	(setq cline (count-lines (point-min) (point)))
	(setq nlines (count-lines (point-min) (point-max)))
	(setq cline (+ cline 1))
	(if (< cline nlines)
		(progn
		  (next-line 1)
		  (interactive-xref-list-goto)
		  )
	  (message "* last reference *")
;;	  (goto-char (point-min))
	  )
))

(defun previous-reference-list-buffer ()
  (let ((cline) (nlines) (poin (point)) (lastref))
	(if (not (eq xref-list-buffer-content 'reference-list))
		(error "Not a reference list buffer")
	  )
	(beginning-of-buffer)
	(setq lastref (search-forward-regexp "^>" (point-max) t))
	(if (not lastref)
		(goto-char poin)
	  )
	(beginning-of-line)
	(setq cline (count-lines (point-min) (point)))
	(setq nlines (count-lines (point-min) (point-max)))
	(setq cline (- cline 1))
	(if (>= cline 0)
		(progn
		  (next-line -1)
		  (interactive-xref-list-goto)
		  )
	  (message "* first reference *")
;;	  (beginning-of-line nlines)
	  )
))

(defun interactive-xref-list-filter-plus (&optional argp)
  "Increase the view filter level"
  (interactive "P")
  (let ((fopt))
	(if (< xref-list-filter-level 3)
		(xref-list-set-filter (+ xref-list-filter-level 1))
	  )
	(message xref-list-buffer-message)
  t
))

(defun interactive-xref-list-filter-minus (&optional argp)
  "Decrease the view filter level"
  (interactive "P")
  (let ((fopt))
	(if (> xref-list-filter-level 0)
		(xref-list-set-filter (- xref-list-filter-level 1))
	  )
	(message xref-list-buffer-message)
	t
))

(defun xref-set-olcxfilter (level)
  (let ((fopt))
	(setq fopt (format "-olcxfilter=%d" level))
	(xref-call-process 'sync 0 nil 0 0 "" fopt "-olcxrefsuffix=0:" "")
))

(defun xref-list-set-filter (level)
  (let ((cwin-list nil))
	(setq cwin-list (string-equal (buffer-name) (xref-list-buffer-name)))
	(select-window (get-buffer-window (xref-list-buffer-name)))
	(other-window -1)
	(xref-set-olcxfilter level)
	(xref-list-top-references)
	(other-window 1)
	(setq xref-list-filter-level level)
	(show-xref-filter-level)
	(setq xref-list-buffer-content 'reference-list)
	(if (not cwin-list)
		(other-window -1)
	  )
))

(defun interactive-xref-list-mouse-goto (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-list-goto)
  (other-window -1)
  t
)

(defun interactive-xref-pop (&optional argp)
  (interactive "P")
  (xref-delete-windows)
  (xref-popCxrefs (point))
  t
)

(defun interactive-xref-mouse-pop (event)
  (interactive "e")
  (mouse-set-point event)
  (xref-popCxrefs (point))
  t
)

(defun interactive-xref-list-filter0 (event)
"Display all references."
  (interactive "e")
  (xref-list-set-filter 0)
)

(defun interactive-xref-list-filter1 (event)
"Top level usages.

Display  definitions, declarations,  l-value and  address  usages (for
variables)  or top  level usages.   For classes  this means  usages in
signatures  of   methods  and  fields  (but  not   in  local  variable
declarations).
"
  (interactive "e")
  (xref-list-set-filter 1)
)

(defun interactive-xref-list-filter2 (event)
"Extends and Implements usages.

Display definitions,  declarations and l-value  usages (for variables)
or usages at  class level. For classes it means  usages in EXTENDS and
IMPLEMENT clauses.
"
  (interactive "e")
  (xref-list-set-filter 2)
)

(defun interactive-xref-list-filter3 (event)
"Display only definitions and declarations."
  (interactive "e")
  (xref-list-set-filter 3)
)

(defvar interactive-xref-list-3bmenu (make-sparse-keymap "Reference List Menu"))
(fset 'interactive-xref-list-3bmenu (symbol-value 'interactive-xref-list-3bmenu))
(define-key interactive-xref-list-3bmenu [xref-list-3bclose] '("Close" . xref-delete-windows))
(define-key interactive-xref-list-3bmenu [xref-list-3bpop] '("Close, Pop and Return" . interactive-xref-pop))
(define-key interactive-xref-list-3bmenu [xref-list-3bsep0] '("----"))
(define-key interactive-xref-list-3bmenu [xref-list-3blinkedframe] '("Link New Frame" . interactive-xref-link-frame))
(define-key interactive-xref-list-3bmenu [xref-list-3bsep2] '("----"))
(define-key interactive-xref-list-3bmenu [xref-list-3bfilter3] '("Set View Filter 3" . interactive-xref-list-filter3))
(define-key interactive-xref-list-3bmenu [xref-list-3bfilter2] '("Set View Filter 2" . interactive-xref-list-filter2))
(define-key interactive-xref-list-3bmenu [xref-list-3bfilter1] '("Set View Filter 1" . interactive-xref-list-filter1))
(define-key interactive-xref-list-3bmenu [xref-list-3bfilter0] '("Set View Filter 0" . interactive-xref-list-filter0))
(define-key interactive-xref-list-3bmenu [xref-list-3bsep1] '("----"))
(define-key interactive-xref-list-3bmenu [xref-list-3binspect] '("This" . interactive-xref-list-mouse-goto))
(define-key interactive-xref-list-3bmenu [xref-list-3bnext] '("Next" . xref-nextref))
(define-key interactive-xref-list-3bmenu [xref-list-3bprevious] '("Previous" . xref-previousref))

(defun interactive-xref-list-xemacs3b-goto (event)
  (interactive "e")
  (interactive-xref-list-goto)
  (other-window -1)
  t
)

(defvar xref-xemacs-list-3bmenu
  '("Reference List Menu"
   ["Previous"      xref-previousref t]
   ["Next"          xref-nextref t]
   ["This"          interactive-xref-list-xemacs3b-goto t]
   "---"
   ["Set View Filter 0"      interactive-xref-list-filter0 t]
   ["Set View Filter 1"      interactive-xref-list-filter1 t]
   ["Set View Filter 2"      interactive-xref-list-filter2 t]
   ["Set View Filter 3"      interactive-xref-list-filter3 t]
   "---"
   ["Link New Frame"   interactive-xref-link-frame t]
   "---"
   ["Close, Pop and Return"       interactive-xref-pop t]
   ["Close"       xref-delete-windows t]
	) "Xref button3 popup menu" 
)

(defun interactive-xref-popup-xemacs-list-menu (event)
  (interactive "e")
  (select-window (get-buffer-window (xref-list-buffer-name)))
  (goto-char (event-closest-point event))
  (beginning-of-line)
  (popup-menu xref-xemacs-list-3bmenu)
)

(defun set-xref-listrefs-dialog-maps ()
  (set-buffer (xref-list-buffer-name))
  (xref-display-buffer (xref-list-buffer-name))
  (select-window (get-buffer-window (xref-list-buffer-name)))
  (show-xref-filter-level)
  (use-local-map xref-list-mode-map)
  (setq xref-list-buffer-content 'reference-list)
  (other-window -1) 
  (message xref-list-buffer-message)
)

(defun handle-pushrefs-action (action)
  (eval action)
;;  (if (eq action 'moveto) 
;;	  (xref-moveref t)
;;	(if (eq action 'listrefs)
;;		(if (eq xref-pilot-char ?\; )
;;			(set-xref-listrefs-dialog-maps)
;;		  )))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-update-reflist-window ()
  (let ((xlw) (swh) (xlwh) (cw) (dff))
	(setq xlw (get-buffer-window (xref-list-buffer-name)))
	(if xlw 
		(progn
		  ;; update reference list
		  (setq cw (get-buffer-window (current-buffer)))
		  (xref-goto-source-window)
		  (setq swh (window-height))
		  (setq xlwh (window-height xlw))
		  (select-window xlw)
		  ;; following should be rather '-olcxrefresh' when implem.
		  (xref-list-top-references)
		  (xref-goto-source-window)
		  (setq dff (- swh (window-height)))
		  (if (not (equal dff 0))
			  (enlarge-window dff)
			)
		  (select-window (get-buffer-window (xref-list-buffer-name)))
		  (setq dff (- xlwh (window-height)))
		  (if (not (equal dff 0))
			  (enlarge-window dff)
			)
		  (select-window cw)
		  )
	  )
))

(defun interactive-xref-menu-selection-select (&optional argp)
  "Select the symbol corresponding to this line."
  (interactive "P")
  (let ((lineno) (str) (nstr) (poin) (xlw) (mess))
	(setq poin (point))
	(setq lineno (count-lines 1 (+ (point) 1)))
	(if (> lineno xref-menu-selection-prefix-lines)
		(progn
		  (beginning-of-line)
		  (setq str (buffer-substring (point) (+ (point) 1)))
		  (if (or (string-equal str " ") (string-equal str "+"))
			  (progn
				(if (string-equal str " ")
					(setq nstr "+")
				  (setq nstr " ")
				  )
				(setq buffer-read-only nil)
				(delete-char 1)
				(insert nstr)
				(setq buffer-read-only t)
				(goto-char poin)
				(next-line 1)
				;; (other-window -1)
				(xref-call-process 'sync 0 nil 0 0 "" (format "-olcxmenuselect%d" lineno) 
								 "-olcxrefsuffix=0:" "")
				(setq mess xref-last-message)
				(xref-update-reflist-window)
				(message mess)
				;; (other-window 1)
				))
		  ))
  )
  t
)

(defun interactive-xref-menu-selection-this-only (&optional argp)
  "Select only the symbol corresponding to this line."
  (interactive "P")
  (save-excursion
	(interactive-xref-menu-selection-none argp)
	(interactive-xref-menu-selection-select argp)
	)
)

(defun apply-fun-menu-selection-lines (fromline toline fun)
  (let ((count) (str))
	(goto-line fromline)
	(setq count fromline)
	(while (<= count toline)
	  (progn
		(beginning-of-line)
		(setq str (buffer-substring (point) (+ (point) 1)))
		(if (or (string-equal str " ") (string-equal str "+"))
			  (eval fun)
		  )
		(next-line 1)
		(setq count (+ count 1))
		))
))
  

(defun interactive-xref-menu-selection-all (&optional argp)
  "Select all the symbols."
  (interactive "P")
  (select-window (get-buffer-window (xref-symbol-resolution-buffer-name)))
  (save-excursion
	(apply-fun-menu-selection-lines 
	 (+ xref-menu-selection-prefix-lines 1)
	 (count-lines 1 (point-max))
	 '(progn 
		(beginning-of-line)
		(setq buffer-read-only nil)
		(delete-char 1)
		(insert "+")
		(setq buffer-read-only t)
		)))
  (xref-call-process 'sync 0 nil 0 0 "" "-olcxmenuall" "-olcxrefsuffix=0:" "")
  (xref-update-reflist-window)
  t
)

(defun interactive-xref-menu-selection-none (&optional argp)
  "Unselect all the symbols."
  (interactive "P")
  (select-window (get-buffer-window (xref-symbol-resolution-buffer-name)))
  (save-excursion
	(apply-fun-menu-selection-lines 
	 (+ xref-menu-selection-prefix-lines 1)
	 (count-lines 1 (point-max))
	 '(progn 
		(beginning-of-line)
		(setq buffer-read-only nil)
		(delete-char 1)
		(insert " ")
		(setq buffer-read-only t)
		)))
  (xref-call-process 'sync 0 nil 0 0 "" "-olcxmenunone" "-olcxrefsuffix=0:" "")
  (xref-update-reflist-window)
  t
)

(defun interactive-xref-menu-selection-mouse-select (event)
"Select unselect the symbol for further processing of references."
  (interactive "e")
  (save-excursion
	(mouse-set-point event)
	(interactive-xref-menu-selection-select)
	)
  t
)


(defun interactive-xref-menu-selection-inspect-def (opt)
  "Inspect definition of the symbol corresponding to this line."
  (interactive "P")
  (let ((lineno) (str) (nstr) (poin) (xlw))
	(setq poin (point))
	(setq lineno (count-lines 1 (+ (point) 1)))
	(if (> lineno xref-menu-selection-prefix-lines)
		(progn
		  (beginning-of-line)
		  (xref-call-process 'sync 0 nil 0 0 "" (format "%s%d" opt lineno) 
						   "-olcxrefsuffix=0:" "")
		  (if (eq xref-pilot-char ?\~)
			  (xref-browse-tmp-file-url (get-url-to-browse))
			(setq xlw (get-buffer-window (xref-list-buffer-name)))
			(if xlw (select-window xlw))
			(other-window -1)
			(xref-linked-frame-moveref t)
			(select-window (get-buffer-window (xref-symbol-resolution-buffer-name)))
		  )))
  t
))

(defun interactive-xref-menu-selection-mouse-inspect-def (opt)
  (interactive "e")
  (let ((cw))
	(setq cw (get-buffer-window (current-buffer)))
	(beginning-of-line)
	(interactive-xref-menu-selection-inspect-def opt)
	(select-window cw)
	)
  t
)

(defun interactive-xref-menu-selection-inspect-symbol (&optional argp)
  "Inspect definition of the symbol corresponding to this line."
  (interactive "P")
  (interactive-xref-menu-selection-inspect-def "-olcxmenuinspectdef")
)

(defun interactive-xref-menu-selection-mouse-inspect-symbol (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-menu-selection-mouse-inspect-def "-olcxmenuinspectdef")
)

(defun interactive-xref-menu-selection-inspect-class (&optional argp)
  "Inspect definition of the symbol corresponding to this line."
  (interactive "P")
  (interactive-xref-menu-selection-inspect-def "-olcxmenuinspectclass")
  (message " ")     ;; to remove the message 'not a fun symbol'
)

(defun interactive-xref-menu-selection-mouse-inspect-class (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-menu-selection-mouse-inspect-def "-olcxmenuinspectclass")
  (message " ")     ;; to remove the message 'not a fun symbol'
)

(defun interactive-xref-menu-selection-mouse-this-only (event)
"Select this one only.

Unselect all  application references,  then select only  the specified
application class  and push its references onto  reference stack. Then
continue only with those references."
  (interactive "e")
  (save-excursion
	(mouse-set-point event)
	(interactive-xref-menu-selection-this-only)
	)
  t
)

(defun interactive-xref-menu-selection-go (&optional argp)
  (interactive "P")
  (xref-delete-windows)
  (xref-call-process 'sync 0 nil 0 0 "" "-olcxmenugo" "-olcxrefsuffix=0:" "")
  (handle-pushrefs-action xref-pushrefs-action)
  t
)

(defun interactive-xref-menu-selection-mouse-go (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-menu-selection-go)
  t
)

(defun xref-set-menu-filter (level)
  (let ((fopt))
	(setq xref-sym-menu-filter-level level)
	(select-window (get-buffer-window (xref-symbol-resolution-buffer-name)))
	(setq buffer-read-only nil)
	(xref-erase-buffer)
	(setq fopt (format "-olcxmenufilter=%d" xref-sym-menu-filter-level))	
	(xref-call-process 'sync 0 nil 0 0 "" fopt "-olcxrefsuffix=0:" "")
	(handle-ambiguous-browsing-dialog xref-pushrefs-action)
  t
))

(defun interactive-xref-menu-filter-plus (&optional argp)
  (interactive "P")
  (let ((fopt))
	(if (< xref-sym-menu-filter-level 2)
		(setq xref-sym-menu-filter-level (+ xref-sym-menu-filter-level 1)))
	(xref-set-menu-filter xref-sym-menu-filter-level)
	t
))

(defun interactive-xref-menu-filter-minus (&optional argp)
  (interactive "P")
  (let ((fopt))
	(if (> xref-sym-menu-filter-level 0)
		(setq xref-sym-menu-filter-level (- xref-sym-menu-filter-level 1)))
	(xref-set-menu-filter xref-sym-menu-filter-level)
	t
))

(defun interactive-xref-menu-filter0 (event)
"All of the same name.

Display all  symbols having  the same name  as the browsed  symbol, no
matter  the profile. Selection  filters are  meaningful only  for Java
language.
"
  (interactive "e")
  (xref-set-menu-filter 0)
)

(defun interactive-xref-menu-filter1 (event)
"All with the same name and profile.

Display  only symbols  with the  same  profile as  browsed symbol  and
propose  all  classes  where  an  definition  or  application  occurs.
Selection filters are meaningful only for Java language.
"
  (interactive "e")
  (xref-set-menu-filter 1)
)

(defun interactive-xref-menu-filter2 (event)
"Only related items.

Display  only symbols  with the  same profile  as browsed  symbol and
propose only applications on classes relevant by inheritance relation.
The last  means all superclasses (or  implemented interfaces) defining
(declaring)   the   browsed   symbol   and   all   their   sub-classes
(implementors)  where an definition  or application  occurs. Selection
filters are meaningful only for Java language.
"
  (interactive "e")
  (xref-set-menu-filter 2)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar interactive-xref-menu-3bmenu (make-sparse-keymap "References Selection Menu"))
(fset 'interactive-xref-menu-3bmenu (symbol-value 'interactive-xref-menu-3bmenu))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bclose] '("Close" . xref-delete-windows))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bgo] '("Continue (Go)" . interactive-xref-menu-selection-go))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bsep0] '("----"))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bfilter2] '("Set View Filter 2" . interactive-xref-menu-filter2))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bfilter1] '("Set View Filter 1" . interactive-xref-menu-filter1))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bfilter0] '("Set View Filter 0" . interactive-xref-menu-filter0))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bsep1] '("----"))
(define-key interactive-xref-menu-3bmenu [xref-menu-3binspectclass] '("Inspect Class" . interactive-xref-menu-selection-mouse-inspect-class))
(define-key interactive-xref-menu-3bmenu [xref-menu-3binspsym] '("Inspect Symbol" . interactive-xref-menu-selection-mouse-inspect-symbol))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bsep2] '("----"))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bnone] '("Unselect All" . interactive-xref-menu-selection-none))
(define-key interactive-xref-menu-3bmenu [xref-menu-3ball] '("Select All" . interactive-xref-menu-selection-all))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bselect] '("Toggle Select/Unselect" . interactive-xref-menu-selection-mouse-select))
(define-key interactive-xref-menu-3bmenu [xref-menu-3bthisonly] '("Select This One Only" . interactive-xref-menu-selection-mouse-this-only))

(defvar xref-xemacs-menu-3bmenu
  '("References Selection Menu"
   ["Select This One Only"      interactive-xref-menu-selection-this-only t]
   ["Toggle Select/Unselect"      interactive-xref-menu-selection-select t]
   ["Select All"        interactive-xref-menu-selection-all t]
   ["Unselect All"      interactive-xref-menu-selection-none t]
   "---"
   ["Inspect Symbol"     interactive-xref-menu-selection-inspect-symbol t]
   ["Inspect Class"      interactive-xref-menu-selection-inspect-class t]
   "---"
   ["Set View Filter 0"      interactive-xref-menu-filter0 t]
   ["Set View Filter 1"      interactive-xref-menu-filter1 t]
   ["Set View Filter 2"      interactive-xref-menu-filter2 t]
   "---"
   ["Continue (Go)"       interactive-xref-menu-selection-go t]
   ["Close"       xref-delete-windows t]
	) "Xref button3 popup menu" 
)

(defun interactive-xref-popup-xemacs-menu-menu (event)
  (interactive "e")
  (select-window (get-buffer-window (xref-symbol-resolution-buffer-name)))
  (goto-char (event-closest-point event))
  (beginning-of-line)
  (popup-menu xref-xemacs-menu-3bmenu)
)

(defun interactive-xref-menu-select-and-inspect-one (&optional argp)
"Select this one only and inspect its definition.

Unselect all  application references,  then select only  the specified
application class  and push its references onto  reference stack. Then
move to its definition."
  (interactive "P")
  (let ((win))
	(setq win (get-buffer-window (current-buffer)))
	(interactive-xref-menu-selection-this-only)
	(xref-definition-reference)
	(sit-for 1)
	(select-window win)
))


(defun handle-ambiguous-browsing-dialog (action)
  (set-buffer (xref-symbol-resolution-buffer-name))
  (goto-char (point-min))
;; this should be also accessible by the standard menu pop-up
  (xref-display-buffer (xref-symbol-resolution-buffer-name))
  (select-window (get-buffer-window (xref-symbol-resolution-buffer-name)))
  (goto-line (+ xref-menu-selection-prefix-lines 3))
  (setq xref-pushrefs-action action)
  (use-local-map xref-menu-selection-mode-map)
  (show-xref-symbol-filter-level)
  (message "button2 - select/unselect; S-button2 - browse; button3 - menu;")
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-after-pushing-continuation (autolist continuation)
  (if (eq xref-pilot-char ?\> )
	  (progn
		;; there may be some synchronisation problem
		(while (accept-process-output xrefsrv-process 0 50))
		(if autolist 
			(xref-list-top-references)
		  )
		(handle-ambiguous-browsing-dialog continuation)
		)
	(handle-pushrefs-action continuation)
	)
)

(defun push-references (synchro file-name poin symrequired autolist 
								pusharg bargnum continuation)
  (let ((name)
		(modified) (line) (col) (tbpars "")
		(res "")
		)
	(if (eq file-name nil)
		(progn
		  (setq res "You should edit a C or Java source file when using Xref")
		  (message res)
		  )
	  (setq modified (buffer-modified-p))
	  (search-backward-regexp xref-backward-pass-identifier-regexp)
	  (if (not (eq (char-after (point)) ?#))
		  (forward-char 1)
		)
	  (setq name (xref-identifier-after-point))
	  (if (and (not symrequired))  ;; (string-equal name ""))
		  (setq name "#")
		)
	  ;; this will be explicit constructor, no name check
	  (if (xref-java-this-or-super-name file-name name)
		  (setq name (format "#%s" name))
		)
	  (if modified
		  (setq tbpars "-olcxmodified")
		(setq tbpars "-olcxunmodified")
		)
	  (if (or (not modified) (not symrequired))
		  (progn
			(setq line (count-lines (point-min) (+ (point) 1)))
			(setq col (xref-current-column))
			(setq tbpars 
				  (format "%s -olcxlccursor=%d:%d -browsedsym=%s"
						  tbpars line col name))
			))
	  (goto-char poin)
	  (if (and (string-equal name "")
			   (not (eq (point) (point-min))))
		  (progn
			(setq res "No symbol, position the cursor on a symbol name.")
			(message res)
			)
		(xref-delete-windows)
		(setq xref-list-filter-level 0)
		(setq xref-sym-menu-filter-level 2)
		(xref-call-process synchro 1 file-name poin 0 "" pusharg 
						 (format "-olcxrefsuffix=%d:%s" bargnum name)
						 tbpars)
		(xref-after-pushing-continuation autolist continuation)
		(setq res nil)
		))
	res
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pushing-post-action ()
  (let ((url) (answer))
	(if (eq xref-pilot-char ?\~)
		(progn
		  ;; no definition; browse javadoc URL
		  (setq url (get-url-to-browse))
		  (if xref-ask-before-browse-javadoc
			  (progn
				(message 
				 "No definition reference found. Browse JavaDoc [y/n]?")
				(setq answer (get-single-touch-event))
				(message "")
				)
			(setq answer 'answer-yes)
			)
		  (if (eq answer 'answer-yes)
			  (xref-browse-tmp-file-url url)
			)
		  )
	  (xref-moveref t)
	  )
))


(defun xref-pushCxrefs (poin)
"Push all references of a symbol and goto its definition.

This function causes  that the currently edited buffer  is parsed, the
references of symbols are recorded, the symbol on the POIN position is
selected,  its  references are  pushed  on  the  reference stack,  the
references are then completed from the Xref tag (cross reference) file
and  finally the  point is  moved to  the definition  of  the recorded
symbol.
"
  (interactive "d")
  (if xref-auto-update-tag-file-before-push
	  (progn
		(xref-save-some-buffers nil)
		(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)
		))
  (push-references 'sync (buffer-file-name) poin t nil "-olcxpush" 0 '(pushing-post-action))
)


(defun xref-push-references (poin)
"Push all reference of a symbol.

This function causes  that the currently edited buffer  is parsed, the
references of symbols are recorded, the symbol on the POIN position is
selected, its  references are pushed  on the reference stack,  and the
references  are then  completed from  the Xref  tag  (cross reference)
file.
"
  (interactive "d")
  (if xref-auto-update-tag-file-before-push
	  (progn
		(xref-save-some-buffers nil)
	  (xref-non-interactive-update-references xref-fast-updates-inside-refactorings)
	  ))
  (push-references 'sync (buffer-file-name) poin t nil "-olcxpushonly" 0 nil)
)

(defun pushCxrefsForRename (poin)
  "push references and go to the definition"
  (interactive "d")
  (push-references 'sync (buffer-file-name) poin t nil "-olcxrename" 0 '(xref-moveref t))
)

(defun xref-push-all-references-within-method (poin)
"Push all references of all symbols inside the given function/method.

The point has to be positioned inside  a body or a header of a method.
This function causes  that the currently edited buffer  is parsed, the
references of symbols are recorded, and all references occuring inside
the method on point are pushed on the top of the reference stack.
"
  (interactive "d")
  (push-references 'sync (buffer-file-name) poin nil nil "-olcxpushallinmethod" 0 nil)
)

;;;;;;;;;;;;;;;;;;;;;;;;;; Class Tree ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun interactive-xref-class-tree-frame (&optional argp)
  (interactive "P")
  (interactive-xref-link-frame)
  ;;(xref-goto-source-window)
  ;;(delete-window)
)

(defun xref-ct-inspect-def (&optional argp)
  (interactive "P")
  (let ((lineno))
	(setq lineno (count-lines 1 (+ (point) 1)))
	(if (> lineno xref-menu-selection-prefix-lines)
		(progn
		  (beginning-of-line)
		  (xref-call-process 'sync 0 nil 0 0 "" (format "-olcxctinspectdef%d" lineno) 
							 "-olcxrefsuffix=0:" "")
		  (if (eq xref-pilot-char ?\~)
			  (xref-browse-tmp-file-url (get-url-to-browse))
			(xref-goto-source-window)
			(xref-linked-frame-moveref t)
			(select-window (get-buffer-window (xref-class-tree-buffer-name)))
		  )))
))

(defun interactive-xref-ct-inspect-def (event)
  (interactive "e")
  (let ((mpos))
	(setq mpos (mouse-position))
	(mouse-set-point event)
	(xref-ct-inspect-def)
	(set-mouse-position (car mpos) (car (cdr mpos)) (cdr (cdr mpos)))
	)
  t
)

(defun xref-scroll-left (&optional argp)
  (interactive "P")
  (scroll-left 1)
  (if (not (eq (char-after (point)) ?\n))
	  (forward-char 1)
	)
)

(defun xref-scroll-right (&optional argp)
  (interactive "P")
  (scroll-right 1)
  (if (and (> (point) 1) (not (eq (char-after (- (point) 1)) ?\n)))
	  (backward-char 1)
	)
)

(defun class-tree-dialog ()
  (let ((ctw))
	(if (eq xref-pilot-char ?\<)
		(progn
		  (set-buffer (xref-class-tree-buffer-name))
		  (goto-char (point-min))
		  (setq truncate-lines t)
		  (setq ctw (get-buffer-window (xref-class-tree-buffer-name) t))
		  (if (not ctw)
			  (progn
				(xref-display-buffer (xref-class-tree-buffer-name))
				(setq ctw (get-buffer-window (xref-class-tree-buffer-name) t))
				)
			)
		  (select-window ctw)
		  (use-local-map xref-class-tree-mode-map)
		  (message "'f' - new frame;  button2 - browse")
		))
))

(defun xref-show-class-tree ()
  (xref-call-process 'sync 0 nil 0 0 "" "-olcxshowctree" 
					 "" "")
  (class-tree-dialog)
)

(defun xref-class-tree-show (poin)
"Display (possibly partial) tree of class inheritance hierarchy.

This  function is menaingfull  only for  Java programmers.   It causes
that the currently edited buffer  is parsed, the references of symbols
are recorded, the symbol on the  POIN position is resolved. If it is a
class, then a class hierarchy consisting from its sub and superclasses
is showed. By decreasing filter level you can get the full class tree.
If the symbol under point is not a class but a member of a type beeing
a class, then this class (the type of this field) is taken as base and
a  class  hierarchy  relative to  it  is  showed.   In this  way,  the
'xref-class-tree-show' function can to determine type of a symbol, the
type is  shown together  with related hierarchy.   Unforfunately, this
feature does  not work for  local variables and method  arguments.  If
the cursor is  not positioned on a  class name or a member  of a class
type, then the full hierarchy starting by Object class is showed.
"
  (interactive "d")
  (push-references 'sync (buffer-file-name) poin nil nil "-olcxclasstree" 0 '(class-tree-dialog))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; List ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-listCxrefs (poin)
"Push references of a symbol and list them in new window.

This function causes  that the currently edited buffer  is parsed, the
references of symbols are recorded, the symbol on the POIN position is
selected,  its  references are  pushed  on  the  reference stack,  the
references are then completed from the Xref tag (cross reference) file
and all references are showed in a separate window.
"
  (interactive "d")
  (let ((iden) (opos))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (if xref-auto-update-tag-file-before-push
		  (progn
			(xref-save-some-buffers nil)
			(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)
			))
	  (save-excursion
		(get-buffer-create (xref-list-buffer-name))
		(set-buffer (xref-list-buffer-name))
		(xref-erase-buffer)
		)
	  (setq opos (point))
	  (search-backward-regexp xref-backward-pass-identifier-regexp)
	  (forward-char 1)
	  (setq iden (xref-identifier-after-point))
	  (goto-char opos)
	  (setq xref-fontification (xref-font-lock-list-keywords iden))
	  (setq xref-list-filter-level 0)
	  (xref-get-windows-status)
	  (push-references xref-reflist-synchro (buffer-file-name) poin t 
					   xref-three-windows 
					   "-olcxlist" 0 
					   '(listrefs-continuation))
)))

;;;;;;;;;;;;;;;;;;; useless long names ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-list-top-references-on-success ()
  (if (not (eq xref-pilot-char ?!))
	  (progn
		(xref-top-refs-resolution (point))
		(xref-list-top-references)
	))
)

(defun xref-list-useless-long-names (poin)
"Push occurences of useless fully qualified Java type names.

This function is meaningful only for Java programmers.  It causes that
the  currently edited  buffer  is parsed,  references  of symbols  are
recorded, the  symbol on the  POIN position is  resolved.  If it  is a
class, then all redundant fully  qualified references of this type are
pushed onto reference stack.  If the symbol under point is not a class
but a  class member then  redundant fully qualified references  of the
class containing this member are  pushed.  If the symbol has no class,
then all redundant fully qualified names of all classes are pushed and
selection can  be done  manually. References can  be filtered  to last
occurences in single fully  qualified name. Filter2 filters references
to those inside current method.
"
  (interactive "d")
  (push-references 'sync (buffer-file-name) poin nil xref-three-windows "-olcxuselesslongnames" 0 
				   '(xref-list-top-references-on-success))
)

;;;;;;;;;;;;;;;;;;; method on class dependence ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-list-method-class-dependencies (poin)
"Push all 'maybe this' references.

This function is meaningful only for Java programmers.  It causes that
input file is  parsed and all 'maybe this'  references from inside the
current  method  are pushed  onto  reference  stack.   A 'maybe  this'
reference is  either reference  of the 'this'  keyword or  a reference
which may  be preceded  by the 'this'  keyword. In other  words 'maybe
this' reference  is any  reference to a  field of the  current object.
The list of  references is displayed as well  as inheritance hierarchy
from where referenced fields were inherited.
"
  (interactive "d")
;;  (push-references 'sync (buffer-file-name) poin nil nil "-olcxmaybethis" 0 
;;				   nil)
  (push-references 'sync (buffer-file-name) poin nil xref-three-windows "-olcxmaybethis" 0 
				   '(xref-list-top-references-on-success))
)

;;;;;;;;;;;;;;;;;;; not FQT refs in method ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-list-method-non-fqt-names (poin)
"This  function is meaningful  only for  Java programmers.   It causes
that ....
"
  (interactive "d")
;;  (push-references 'sync (buffer-file-name) poin nil nil "-olcxnotfqt" 0 
;;				   nil)
  (push-references 'sync (buffer-file-name) poin nil xref-three-windows "-olcxnotfqt" 0 
				   '(xref-list-top-references-on-success))
)

;;;;;;;;;;;;;;;;;;; useless type names ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun remove-useless-long-names-yet-pushed ()
  (xref-set-olcxfilter 1)
;;(message "calling intersection")
  (xref-call-process 'sync 0 nil 0 0 "" "-olcxintersection" 
					   "" "-olcxrefsuffix=0:#")
)

(defun xref-nonint-remove-useless-long-names (poin interact threewin preaction)
	(xref-map-on-all-references  (buffer-file-name) poin nil threewin
						   "-olcxuselesslongnames"
						   interact 
						   nil
						   ''none-pre-fun
						   ''xref-fqt-name-modify-static-prefix
						   ''standard-post-fun 
						   ""
						   nil
						   (list 'progn 
								 preaction
								 (list 'xref-call-process ''sync 0 nil 
									   0 0 "" "-olcxgoto0" "" ""
									   ))
						   (list 'xref-popCxrefs poin)
						   )
)

(defun xref-reduce-useless-long-names (poin)
"Reduce all useless Java fully qualified type names of given type.

This function is meaningful only for Java programmers.  It causes that
the  currently edited  buffer  is parsed,  references  of symbols  are
recorded, the  symbol on the  POIN position is  resolved.  If it  is a
class, then all redundant fully  qualified references of this type are
pushed onto reference stack.  If the symbol under point is not a class
but a  class member then  redundant fully qualified references  of the
class containing this member are  pushed.  If the symbol has no class,
then all redundant fully qualified names of all classes are pushed and
selection can  be done manually.  After this  references are inspected
one by one and useless prefixes are removed.
"
  (interactive "d")
	(if (yes-or-no-p (format "Update Tag file before reducing long names [recommended]? "))
		(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)
		)
	(xref-nonint-remove-useless-long-names poin 
										   xref-detailed-refactoring-confirmations 
										   xref-three-windows 
										   (list 'xref-set-olcxfilter 1))
)

(defun xref-nonint-reduce-useless-long-names-of-method (poin interact)
  (save-excursion
	(search-backward-regexp "[ \t\n]" 0 t)
	(xref-nonint-remove-useless-long-names poin interact xref-three-windows (list 'xref-set-olcxfilter 2))
	)
)

(defun xref-reduce-useless-long-names-of-method (poin)
"Reduce all useless Java fully qualified type names within a method.

This function is meaningful only for Java programmers.  It causes that
the  currently edited  buffer  is parsed,  references  of symbols  are
recorded. All redundant fully  qualified references within the body of
the method where the point is are reduced to standard short names.
"
  (interactive "d")
  (xref-multifile-undo-set-buffer-switch-point "reduction of long names")
  (xref-nonint-reduce-useless-long-names-of-method poin xref-detailed-refactoring-confirmations)
)

;;; expand short names in method's body to fully qualified refs ;;;;;;

(defun xref-get-current-ref-appl-class()
  (let ((fqtname))
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxtopapplcl" "" "")
	(if (not (eq xref-pilot-char ?*))
		(error "[xref] can't get fully qualified name, internal error.")
	  (setq fqtname xref-last-message)
	  )
	fqtname
))

(defun expand-to-fqt (dummy1 dummy2)
  (let ((uchar) (fqtname))
	(setq uchar xref-move-usage)
	(setq fqtname (xref-get-current-ref-appl-class))
	(if (string-match "\\.[0-9]" fqtname)
		(message "Anonymous nested class, no expansion")
	  (if (eq uchar ?.)
		  (progn
			;; field reference, simply insert the fully qualified name
			(insert fqtname ".")
			)
		(kill-region (point) (mark t))
		(insert fqtname)
		)
)))

(defun xref-expand-short-method-refs (poin interact)
  (let ((postfun))
	(if interact
		(setq postfun ''standard-post-fun)
	  (setq postfun ''none-post-fun)
	  )
	(xref-map-on-all-references  (buffer-file-name) poin nil nil
						   "-olcxnotfqt"
						   interact
						   nil
						   ''none-pre-fun
						   ''expand-to-fqt
						   postfun
						   nil  ;; arg1?
						   nil  ;; arg2?
						   (list 'xref-call-process ''sync 0 nil (point) 0 "" 
											  "-olcxgoto0" "" "")
						   (list 'xref-popCxrefs poin)
						   )
))

(defun xref-expand-short-class-refs (poin interact)
  (let ((postfun))
	(if interact
		(setq postfun ''standard-post-fun)
	  (setq postfun ''none-post-fun)
	  )
	(xref-map-on-all-references  (buffer-file-name) poin nil nil
						   "-olcxnotfqtinclass"
						   interact
						   nil
						   ''none-pre-fun
						   ''expand-to-fqt
						   postfun
						   nil  ;; arg1?
						   nil  ;; arg2?
						   (list 'xref-call-process ''sync 0 nil (point) 0 "" 
											  "-olcxgoto0" "" "")
						   (list 'xref-popCxrefs poin)
						   )
))

(defun xref-expand-short-refs-of-method (poin)
"Expand all static references within a method to fully qualified names.

This function is meaningful only for Java programmers.  It causes that
the  currently edited  buffer  is parsed,  references  of symbols  are
recorded. All short references (not fully qualified) are inspected and
expanded to fully qualified references.  Hereby the method is ready to
be moved into another class.
"
  (interactive "d")
  (xref-multifile-undo-set-buffer-switch-point "expansion of short names")
  (xref-expand-short-method-refs poin xref-detailed-refactoring-confirmations)
)

(defun xref-expand-short-refs-of-class (poin)
"Expand all static references within class to fully qualified names.

This function is meaningful only for Java programmers.  It causes that
the  currently edited  buffer  is parsed,  references  of symbols  are
recorded. All short references (not fully qualified) are inspected and
expanded to fully qualified references.   Hereby the class is ready to
be moved to another place.
"
  (interactive "d")
  (xref-multifile-undo-set-buffer-switch-point "expansion of short names")
  (xref-expand-short-class-refs poin xref-detailed-refactoring-confirmations)
)

;;;;;;;;;;;;;;;;;;; one-line top symbol info ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-show-stack-top-symbol (poin)
"Display the name of the symbol on the top of reference stack. 
This is usually a fully qualified name.
"
  (interactive "d")
  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxtops" "" "")
)


;;;;;;;;;; one-line current top symbol application class ;;;;;;;;;;;;;;


(defun xref-show-current-ref-appl-class (poin)
"Display the name of the class  of the current reference on the top of
the reference  stack. This function is  used to expand  short names to
fully qualified names.
"
  (interactive "d")
  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxtopapplcl" "" "")
)

;;;;;;;;;;;;;;;;;;; top references resolution ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-top-refs-resolution (poin)
"Display the resolution dialog for references on the top of the symbol stack.
In this way you can edit top references.
"
  (interactive "d")
  (let ((xlw) (xlwt))
	(setq xlw (get-buffer-window (xref-list-buffer-name)))
	(if xlw
		(setq xlwt t)
	  (setq xlwt nil)
	  )
	(push-references 'sync (buffer-file-name) poin nil xlwt "-olcxedittop" 0 '(xref-delete-windows))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;; reference arithmetics  ;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-ra-intersection (poin)
"Compute intersection of references on the top of reference stack.

This function  computes an intersection  of references on the  two top
positions  of the reference  symbol stack.   I.e.  from  lastly pushed
symbol references  only those  references are kept  which are  also in
previously  pushed list  no matter  the symbol  name. Also  the column
position  is not  tested meaning  that two  references  are considered
identical if they refer to the same file and line.
"
  (interactive "d")
  (let ((wins)) 
	(setq wins (xref-get-windows-status))
	(xref-call-process 'sync 0 nil poin 0 "" "-olcxintersection" "" "")
	(if (not (eq xref-pilot-char ?!))
		(progn
		  (xref-delete-windows)
		  ;; this has to in this order because of some xemacs problems
		  (xref-reconstitute-windows wins)
		  ))
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; pop refs ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-popCxrefs (poin)
"Pop one symbol references from the reference stack.
Also move point at the position from where the references were pushed.
"
  (interactive "d")
  (let ((wins)) 
	(setq wins (xref-get-windows-status))
	;; delete windows before calling, to not kill accidentaly *xrefsrv* buffer
	(xref-delete-windows)
	(xref-call-process 'sync 0 nil poin 0 "" "-olcxpop" "" "")
	(if (eq xref-pilot-char ?#)
		(progn
		  ;; this has to in this order because of some xemacs problems
		  (xref-moveref nil)
		  ))
	(xref-reconstitute-windows wins)
	(xref-show-stack-top-symbol (point))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; re-push refs ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-re-push (poin)
"Re-push lastly poped references.
"
  (interactive "d")
  (let ((wins)) 
	(setq wins (xref-get-windows-status))
	(xref-call-process 'sync 0 nil poin 0 "" "-olcxrepush" "" "")
	(if (eq xref-pilot-char ?#)
		(progn
		  (xref-delete-windows)
		  (xref-moveref nil)
		  (xref-reconstitute-windows wins)
		  (xref-show-stack-top-symbol (point))
		  ))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;; LOCAL MOTION  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-this-symbol-local-motion (poin continuation)
  (let ((iden) (opos))
	(if (eq (buffer-file-name) nil)
		(message "You should edit a C or Java source file when using Xref")
	  (setq opos (point))
	  (search-backward-regexp xref-backward-pass-identifier-regexp)
	  (forward-char 1)
	  (setq iden (xref-identifier-after-point))
	  (goto-char opos)
	  (push-references 'sync (buffer-file-name) poin t nil "-olcxpushforlm -olnodialog" 0 continuation)
)))

(defun xref-poponly ()
  "Pop references from the top of the stack."
  (interactive "")
  (xref-delete-windows)
  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxpoponly" "" "")
)


(defun xref-this-symbol-previous-ref (poin)
  "Go to the previous reference of the symbol under the cursor."
  (interactive "d")
  (xref-this-symbol-local-motion poin 
		   '(if (eq xref-pilot-char ?\# ) 
				(progn
				  (xref-previousref) 
				  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxpoponly" "" "")
				  ))
		   )
)


(defun xref-this-symbol-next-ref (poin)
  "Go to the next reference of the symbol under the cursor."
  (interactive "d")
  (xref-this-symbol-local-motion poin 
		   '(if (eq xref-pilot-char ?\# ) 
				(progn
				  (xref-nextref) 
				  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxpoponly" "" "")
				  ))
		   )
)

(defun xref-this-symbol-definition-ref (poin)
  "Go to the definition of the symbol under the cursor."
  (interactive "d")
  (xref-this-symbol-local-motion poin 
		   '(if (eq xref-pilot-char ?\# ) 
				(progn
				  (xref-call-process 'sync 0 nil 0 0 "" "-olcxgotodef" 
								   "-olcxrefsuffix=0:" "")
				  (xref-moveref t)
				  (xref-call-process 'sync 0 nil 0 0 "" "-olcxpoponly" "" "")
				  ))
		   )
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(defun xref-nextref ()
"Go to the  next reference on the top of the  reference stack. 
After a compilation command this function moves to the next error message."
  (interactive "")
  (let ((list-win) (cwin) (compile-win))
	(if (string-equal (buffer-name) (xref-list-buffer-name))
		(next-reference-list-buffer)
	  (setq list-win (get-buffer-window (xref-list-buffer-name)))
	  (setq compile-win (get-buffer-window "*compilation*"))
	  (if (and compile-win (eq xref-last-referencing-action 'compilation))
		  (next-error 1)
		(if (and list-win (eq xref-list-buffer-content 'reference-list))
			(progn
			  (setq cwin (selected-window))
			  (select-window list-win)
			  (next-reference-list-buffer)
			  (select-window cwin)
			  )
		  (xref-call-process 'sync 0 nil 0 0 "" "-olcxplus" "-olcxrefsuffix=0:" "")
		  (xref-moveref t)
)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-previousref ()
"Go to the previous reference on the top of the reference stack. 
After a compilation command this function moves to the previous error.
"
  (interactive "")
  (let ((list-win) (cwin) (compile-win))
	(if (string-equal (buffer-name) (xref-list-buffer-name))
		(previous-reference-list-buffer)
	  (setq list-win (get-buffer-window (xref-list-buffer-name)))
	  (setq compile-win (get-buffer-window "*compilation*"))
	  (if (and compile-win (eq xref-last-referencing-action 'compilation))
		  (next-error -1)
		(if (and list-win (eq xref-list-buffer-content 'reference-list))
			(progn
			  (setq cwin (selected-window))
			  (select-window list-win)
			  (previous-reference-list-buffer)
			  (select-window cwin)
			  )
		  (xref-call-process 'sync 0 nil 0 0 "" "-olcxminus" "-olcxrefsuffix=0:" "")
		  (xref-moveref t)
)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-find-file-on-mouse (event)
  (interactive "e")
  (let ((b) (e) (p) (file) (line 0) (ne) (c1) (c2))
	(mouse-set-point event)
	(setq p (point))
	(search-backward-regexp xref-find-file-on-mouse-delimit (point-min) 0)
	(forward-char 1)
	(setq b (point))
	(if (and (eq xref-platform 'windows) (>= (point) 3))
		(if (xref-exact-string-match "[^A-Za-z][A-Za-z]:" (buffer-substring (- (point) 3) (point)))
			;; add windows drive letter to file
			(setq b (- (point) 2))
		  ))
	(search-forward-regexp xref-find-file-on-mouse-delimit (point-max) 0)
	(setq e (point))
	(if (<= (- e 1) b)
		(error "No file name on mouse")
	  )
	(setq file (buffer-substring b (- e 1)))
	(if (eq (char-after (- e 1)) ?:)
		(progn
		  ;; probably a line number
		  (forward-char 1)
		  (search-forward-regexp "[^0-9]" (point-max) 0)
		  (setq ne (point))
		  (setq line (string-to-int (buffer-substring e (- ne 1))))
		  )
	  )
	(goto-char p)
	(if (not (file-attributes file))
	  (error "Can't open %s" file)
	)
	(if (not (or (equal (substring file 0 1) "/") 
				 (and (> (length file) 3) (equal (substring file 1 3) ":\\"))
				 (equal (substring file 0 1) "\\")))
		(progn
		  ;; a relative path, concatenate with working dir
		  (setq file (concat default-directory file))
		  ))
	;; too simple for three windows situation
	(xref-goto-source-window)
	(find-file file)
	(if (not (equal line 0))
		(goto-line line)
	  )
	)
  t
)

(defun xref-simple-output-filter (process output)
  (let ((buffer (current-buffer))
		(wind (selected-window))
		(scroll scroll-step)
		(owin)
		)
;;	(set-buffer "*background-1*")
	(setq owin (get-buffer-window "*background-1*"))
	(if owin
		(progn
		  (select-window owin)
		  (setq scroll-step 1)
		  (goto-char (point-max))
		  (insert output)
		  (goto-char (point-max))
		  (sit-for 0)
		  (setq scroll-step scroll)
		  ;;	(set-buffer buffer)
		  (select-window wind)
		  )
	  (message "%s" output)
	  )
))

(defun xref-call-non-interactive-process (project updateoption sync wind message)
  (let ((buffer (current-buffer)) 
		(bdir default-directory)
		(ubuf) (args) (res 0) (dotcount) (barcount) (oldpct))
	(setq args nil)
 	(if project
		(setq args (append (list "-p" project) args))
;;    following discards prefix projects (except prefix before first project) 
;;    from beeing processed !!
;;    but commenting it out would cause that cross-references from extern file
;;    will make warning or (worst) change active project
	  (xref-call-process 'sync 1 nil 0 0 "" "-olcxgetprojectname" ;; -noerrors
						 "-olcxrefsuffix=0:" "")
	  (if (eq xref-pilot-char ?\*) 
		  (setq args (append (list "-p" xref-last-message)))
		)
	  )
	(setq args (append (list "-last_message=Done." updateoption) args))
	(if (not (equal wind 'nowin))
		(progn
		  (setq ubuf (get-buffer-create "*background-1*"))
		  (set-buffer ubuf)
		  (xref-erase-buffer)
		  (xref-display-buffer "*background-1*")
		  (setq default-directory bdir)
		  (local-set-key [(shift mouse-2)] 'xref-find-file-on-mouse)
		  (set-buffer buffer)
		  ))
	(setq oldpct process-connection-type)
	(setq process-connection-type nil)
	(setq xrefxref-process (apply 'start-process "xref-assync" nil
								  (concat xref-home-directory "xref") args))
	(if (not (equal wind 'nowin))
		(set-process-filter xrefxref-process 'xref-simple-output-filter)
	  )
	(setq process-connection-type oldpct)
	(setq dotcount "")
	(setq barcount "")
	(if (eq sync 'synchronized)
		(progn
		  (while (eq (process-status xrefxref-process) 'run)
			(progn
			  (if (equal dotcount "")
				  (message "%sPlease wait." message)
				(message "%sPlease wait. (%s%s.)" message barcount dotcount)
				)
			  (setq dotcount (concat dotcount "."))
			  (if (>= (length dotcount) 50)
				  (progn 
					(setq barcount (concat barcount "+"))
					(setq dotcount "")
					))
			  (accept-process-output xrefxref-process)
			  ;; sleep very small time, accept process sleeps
			  (sleep-for 0.05)
			  (discard-input)
			  (sit-for 0.05)
			  ))
		  (sleep-for 0.05)
		  (discard-input)
		  (sit-for 0.05)
		  (setq res (process-exit-status xrefxref-process))
		  )
	  (message (substitute-command-keys "Please wait, then you can delete new window by hitting \\[xref-delete-windows]."))
	  )
	res
))

(defun xref-project-exit-message-on-nonzero (stat)
  (if (not (equal stat 0))
	  (progn
		(if (equal stat 66)
			(error "Process exited with status %d: Project not found. Aborting further actions." stat)
		  (if (equal stat 67)
			  (error "Process exited with status %d: Evaluation license expired. Aborting further actions." stat)
			(error "Process exited with status %d. Aborting further actions." stat)
		  ))
		))
)

(defun xref-non-interactive-update-references (fast-flag) 
  (let ((updateoption) (stat) (winopt))
	(xref-save-some-buffers t)
	(if fast-flag
		(setq updateoption "-fastupdate")
	  (setq updateoption "-update")
	  )
	(if xref-detailed-refactoring-confirmations
		(setq winopt 'newwin)
	  (setq winopt 'nowin)
	  )
	(setq stat (xref-call-non-interactive-process xref-current-project updateoption 'synchronized 
										 winopt "Updating Tag File. "))
	(xref-project-exit-message-on-nonzero stat)
	(if xref-detailed-refactoring-confirmations
		(progn
		  (message "The Tag file is updated. (hit any key to continue)")
		  (apply xref-get-event nil)
		  ))
	(xref-delete-windows)
))

(defun xref-refactoring-though-point (mess)
  (message mess)
  (if xref-detailed-refactoring-confirmations
	  (progn
		(sit-for 0)
		(apply xref-get-event nil)
		))
)

(defun xref-refactoring-init-actions (undo-comment)
  ;; first of all check that a project exists
  (xref-nonint-get-active-project-or-error)
  (if xref-auto-update-tag-file-before-refactorings
	  (progn
		;; following will refresh the screen
		(xref-refactoring-though-point 
		 "STEP 0. Updating Tag File (Hit any key to continue).")
		(xref-save-some-buffers nil)
		(xref-non-interactive-update-references 
	         xref-fast-updates-inside-refactorings)
	  ))
  (xref-multifile-undo-set-buffer-switch-point undo-comment)
  (set-marker xref-refactoring-init-marker (point))
  (message "Starting refactoring %s" undo-comment)
)

(defun xref-refactoring-final-update-actions ()
  (xref-refactoring-though-point 
   "All Done. Updating Tag File after the refactoring.")
  (xref-save-some-buffers t)
  (xref-non-interactive-update-references 
   xref-fast-updates-inside-refactorings)
  (if xref-move-point-back-after-refactoring
	  (xref-switch-to-marker xref-refactoring-init-marker)
	)
  (message "All done.")
)

(defun xref-gen-html-documentation ()
"Generate HTML documentation for active project.  

This function executes the 'xref' command with '-html' option.  If the
current project  is set (with  the `xref-project-set-active' function)
then also the '-p project' option is set. The effect of the invocation
is  that  all project  files  are  parsed  and HTML  dosumentation  is
generate.  This function  requires that Tag file is  created and is up
to date.   It also requires that settings  controlling HTML generation
for active project is configured in the .xrefrc configuration file.
"
  (interactive "")
  (if (eq (buffer-file-name) nil)
	  (message "You should edit a C or Java source file when using Xref")
	(xref-delete-windows)
	(xref-call-non-interactive-process xref-current-project "-html" 'assynchronized 'newwin "")
))


(defun xref-update-refs ()
"Update Xref  cross reference file.  

This function executes the 'xref'  command with the -update option. If
the  current  project  is  set  (with  the  `xref-project-set-active'
function) then also the '-p project'  option is set. The effect of the
invocation  is that  all files  modified  since last  update (but  not
taking into  account 'fast updates'  (see also `xref-fast-update-refs'
function) are  processed as well as all  files (recursively) including
those modified  files.  In result  the cross reference file  is safely
updated.  The behavior of the  'xref' command is controlled by options
read from the '.xrefrc' file.
"
  (interactive "")
  (if (eq (buffer-file-name) nil)
	  (message "You should edit a C or Java source file when using Xref")
	(xref-delete-windows)
	(xref-call-non-interactive-process xref-current-project "-update" 'assynchronized 'newwin "")
))


(defun xref-fast-update-refs ()
"Update Xref  cross reference file.  

This function executes the 'xref' command with the -fastupdate option.
If  the current  project is  set (with  the `xref-project-set-active'
function) then also the '-p project' option is set.  The effect of the
invocation is  that files modified  since last (fast) updating  of the
cross  reference  file are  processed  and  the  cross references  are
updated.   No include-included  relation  is taken  into account  when
determining  which files  will be  processed.  Under  a  very specific
conditions, the 'fast  update' can lead to a  lost or misunderstanding
of few references, so it is recommended to make full update before any
important source  manipulating operation.  The behavior  of the 'xref'
command is controlled by options read from the '.xrefrc' file.
"
  (interactive "")
  (if (eq (buffer-file-name) nil)
	  (message "You should edit a C or Java source file when using Xref")
	(xref-delete-windows)
	(xref-call-non-interactive-process xref-current-project "-fastupdate" 'assynchronized 'newwin "")
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-create-refs ()
"Create Xref  cross reference file.  

This function execute  the 'xref' command with the  -create option. If
the  current  project  is  set  (with  the  `xref-project-set-active'
function) then also the '-p project'  option is set. The effect of the
invocation is that a cross  reference file is created, the behavior of
the 'xref'  command is controlled  by options read from  the '.xrefrc'
file.
"
  (interactive "")
  (xref-delete-windows)
  (xref-call-non-interactive-process xref-current-project "-create" 'assynchronized 'newwin "")
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun  xref-about ()
  "Show Xref version and current project setting informations."
  (interactive "")
  (xref-delete-windows)
  (xref-call-non-interactive-process "dummy" "-about" 'synchronized 'newwin "")
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun  xref-registration ()
  "Enter Xrefactory license string to run registered version."
  (interactive "")
  (let ((lic) (pp))
	(xref-delete-windows)
	(set-buffer  (get-buffer-create " *xrefsrv*"))
	(xref-erase-buffer)
	(xref-display-buffer " *xrefsrv*")
	(insert " 

Further evolution and maintenance  of Xrefactory is entirely dependent
on its users.  The xref task is proprietary software.  In order to run
it legally  you need to purchase  a license.  You  can obtain detailed
information   about   available   licenses   and  order   process   on
http://www.xref-tech.com/xrefactory/license.html  .   After successful
registration you will receive a  license string which will change your
evaluation copy into regular one.

If you  have received valid license  string, please, insert  it now on
the prompt.
")
	(beginning-of-buffer)
	(setq lic (read-from-minibuffer "Enter the license string: "))
	(if (not (equal lic ""))
		(progn
		  (find-file xref-options-file)
		  (beginning-of-buffer)
		  (setq pp (search-forward-regexp "^[ \t]*-license=" nil t))
		  (if pp
			  (progn
				(beginning-of-line)
		  (insert "// ")
		  (end-of-line)
		  (newline)
		  )
			(newline)
			(previous-line 1)
			)
		  (insert (format "-license=%s" lic))
		  (save-buffer)
		  (bury-buffer)
		  (xref-delete-windows)
		  (message "Done. The license string is inserted in %s file." xref-options-file)
		  )
	  (xref-delete-windows)
	  (message "No license string. Exiting.")
	  )
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun  xref-help ()
  "Show basic help informations for use of Xrefactory."
  (interactive "")
  (set-buffer  (get-buffer-create " *xrefsrv*"))
  (xref-erase-buffer)
  (xref-display-buffer " *xrefsrv*")
  (insert " Xref Help.

Xrefactory  is a refactoring  development environment  for C  and Java
languages.  Its  functions can be  accessed via the 'Xref'  submenu of
the  main  menu  bar.   Optionally  the most  frequent  functions  are
accessible  via hot-keys.  Documentation of  a particular  function is
available through  the \"C-h k\" key combination  (i.e.  the 'control'
key together  with 'h'  key and  then the 'k'  key) followed  by mouse
selection of the corresponding menu item.

PROJECTS:
Xrefactory  is project based,  you will  need to  create or  select an
'active project' before doing anything else. In particular, you should
start your work with Xrefactory  by invoking the 'Project -> New' menu
item.

TAG FILE: The  main object you will then work with  is the 'Tag File'.
This  file stores all  the necessary  informations about  your project
sources.  In particular informations  about all symbols, their linking
properties, definition  place(s) and  all usages.  The  maintenance of
the Tag File is on the  responsibility of the user.  A non-updated Tag
File will provoke mistakes in  source browsing.  On the other side the
code Completion  function as well  as Extract Region  refactorings are
independent  on  the  Tag File  as  they  depend  only on  file  local
informations.  Three  functions are  available for maintenance  of Tag
file depending  on the quality  of the updated references.   The 'Full
Update' function should guarant  100% correctness of tag file, however
it is recomended  to re-Create tag file from time to  time in order to
remove garbage.

BROWSING  AND REFACTORINGS:  In general  you have  to  position cursor
(point in Emacs terminology) on the browsed (refactored) symbol before
invoking any reference pushing  or refactoring function. For parameter
manipulations  you need  to position  the cursor  on the  name  of the
function/method,  not on  the parameter  itself.  Before  invoking the
extract region  function you  will need to  select a region  with your
mouse or by specifying begin of the region by C-<space> and the end by
the cursor. Xrefactory may propose  you some dialogs in a horizontally
split  window, in  general in  those screens  the middle  mouse button
makes the choice the right button proposes menu of available actions.

XREF TASK:
Emacs Xref  functions cooperate with  an external 'xref' task,  if you
feel that the task has run  into an inconsistent state, or if you wish
to interrupt  creation or update of  the Tag file, you  can invoke the
'Kill xref task' function.

CUSTOMIZATION: 
Xrefactory can be  customized via the 'Xref ->  Options' menu item and
via  an  ~/.xrefrc  configuration   file.   The  'Options'  menu  item
customizes  project independent  behaviour  which is  mainly the  user
interface.   In  the '.xrefrc'  file  you  can  specify your  projects
settings and  preferences.  There are  many options you  can customize
via .xrefrc  file, for more  informations read xref and  xrefrc manual
pages. Before using Xrefactory you should equally read the README file
included in distribution package.
")
  (beginning-of-buffer)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;       MAPPINGS        ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ref-equal (file1 file2 line1 line2 coll1 coll2) 
  (and (string-equal file1 file2) (equal line1 line2) (equal coll1 coll2)))

(defun xref-get-single-yes-no-event () 
  (let ((key) (res nil))
	(while (eq res nil)
	  (setq key (apply xref-get-event nil))
	  (setq key (vector key))
;;	  (setq key (read-key-sequence-vector "")
	  (setq res (lookup-key xref-query-replace-map key))
	  (if (eq res nil)
		  (message "Invalid key, please answer [y/n].")
		)
	  )

	res
))

(defun get-single-touch-event () 
  (let ((key) (res nil))
	(setq key (apply xref-get-event nil))
	(setq key (vector key))
;;	  (setq key (read-key-sequence-vector "")
	(setq res (lookup-key xref-query-replace-map key))
	res
))

(defun check-and-count-references (special-check)
  (let ((f0) (l0) (c0) (ff) (ll) (cc) (refcheck) (loop t) (num 0))
	(message "Checking references")
	(xref-previousref)
	(setq f0 xref-lastref-file)
	(setq l0 xref-lastref-line)
	(setq c0 xref-lastref-coll)
	(while loop
	  (setq refcheck (xref-previousref))
	  ;; previous ref also checks for name equality
	  (if (not refcheck)
		  (progn
			(setq loop nil)
			(setq num 0)
			)
		(message "Checking references")
		(setq ff xref-lastref-file)
		(setq ll xref-lastref-line)
		(setq cc xref-lastref-coll)
		(setq num (+ num 1))
		(eval special-check)
		(setq loop (not (ref-equal f0 ff l0 ll c0 cc)))
	  ))
  num
))

(defun none-pre-fun (arg1 arg2)
  nil
)

(defun none-post-fun (arg1 arg2)
  nil
)

(defun standard-post-fun (interactive rrr)
  (if (and interactive (eq rrr 'answer-yes))
	  (progn
		(message 
		 "The occurence is processed. Hit any key to pass to the next reference.")
		(apply xref-get-event nil)
;;		(get-single-touch-event)
		)
	)
)

;;;;;;;;;;;;;;;;;;;; MAPPING FUNCTION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; - pre-fun is usually positioning point and mark on important positions
;; - fun provides the action
;; - post fun make a post action before move to next reference
;; !!!!! the mapped 'fun' returns its error message or nil for O.K. !!!!
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-map-on-all-references2 (poin interact special-check 
									pre-fun fun post-fun arg1 arg2)
  (let ((num) (count) (app) (mess "") (pass-ref) (rrr) (problem))
	(setq count (check-and-count-references special-check))
	(setq num count)
	(if (eq num 0)
		(progn
		  (message "No pushed references to map.")
		)
	  (xref-nextref)
	  (while (> num 0)
		(xref-previousref)
		(setq mess "Done.")
		(setq app (apply pre-fun arg1 arg2 nil))
		(if (not (eq app nil))
			(setq mess app)
		  )
		(if (and interact (eq app nil))
			(progn
			  (message " Proceed this occurence [y/n/a]? ")
			  (setq rrr (xref-get-single-yes-no-event))
			  )
		  (setq rrr 'answer-yes)
		  )
		(if (eq rrr 'answer-all)
			(progn
			  (setq rrr 'answer-yes)
			  (setq interact nil)
			  ))
		(if (and (eq app nil) (eq rrr 'answer-yes))
			(progn
			  (setq app (apply fun arg1 arg2 nil))
			  (undo-boundary)
			  (if (not (eq app nil))
				  (setq mess app)
				)
			  )
		  )
		(setq problem (not (eq app nil)))
		(if problem (xref-wait-until-problem-confirmed mess))
		(apply post-fun interact rrr nil)
		(setq num (- num 1))
	  )
;;	  (xref-popCxrefs poin) ;; no longer poped beause of misc safety checks?
	  (message "Done.")
	)
	count
))

(defun xref-wait-until-problem-confirmed (mess)
  (let ((problem) (rrr))
	(setq problem t)
	(while problem
	  (message (concat mess " Type 'c' to confirm."))
	  (setq rrr (get-single-touch-event))
	  (setq problem (not (eq rrr 'answer-confirmed)))
	  )  
))

(defun xref-map-on-all-references (file-name poin 
							  symrequired autolist pusharg
							  interact special-check
							  pre-fun fun post-fun arg1 arg2 
							  preevaluation continuation)
  (push-references 'sync file-name poin symrequired autolist pusharg 0 
    (list 'progn
	   (list 'eval preevaluation)
	   (list 'xref-map-on-all-references2 poin interact special-check pre-fun fun post-fun arg1 arg2)
	   (list 'eval continuation)
	   ))
)

(defun exchange-text (p1b p1e p2b p2e) 
  (let ((s1 (buffer-substring p1b p1e))
		(s2 (buffer-substring p2b p2e)))
	(if (< p1b p2b)
		(progn
		  (goto-char p2b)
		  (delete-char (- p2e p2b))
		  (insert s1)
		  (goto-char p1b)
		  (delete-char (- p1e p1b))
		  (insert s2)
		  )
	  (goto-char p1b)
	  (delete-char (- p1e p1b))
	  (insert s2)
	  (goto-char p2b)
	  (delete-char (- p2e p2b))
	  (insert s1)
	  )
	)
)

(defun exchange-arguments (arg1 arg2)
  "exchange of two arguments of a function"
  (let ((pos1beg)(pos1end)(pos2beg)(pos2end)(rrr)
		(pos (point)) (fspos) (result nil) (rflag1) (rflag2))
	(xref-pass-to-fun-lpar)
	(if (not (eq (char-after (- (point) 1)) ?\( ))
		(progn 
		  (beep) 
		  (setq result "Hmm. Does not seem as a function symbol reference.")
		  )
	  (setq fspos (point))
	  (setq rflag1 (mark-fun-argument arg1))
	  (setq pos1beg (point))
	  (setq pos1end (mark t))
	  (goto-char fspos)
	  (setq rflag2 (mark-fun-argument arg2))
	  (setq pos2beg (point))
	  (setq pos2end (mark t))
	  (if (and rflag1 rflag2)
		  (exchange-text pos1beg pos1end pos2beg pos2end)
		(setq result "Argument out of range.")
	  )
  )
  (goto-char pos)
  result
))

(defun xref-arg-repush-and-change-to-arg-refs (arg name)
  (xref-re-push (point))
  (xref-call-process 'sync 0 nil 0 0 ""
					 (format "-olcxresetrefsuffix=%d:%s" arg name) 
					 "" "")
  (message " ** Done. References kept on stack for manual adjustements. **")
)

(defun xref-arg-safety-check-finish (keep arg name)
  (if (equal xref-pilot-char ?\;)
	  (progn
		;; safety check does not passed
		(listrefs-continuation)
		;; must be after 'listrefs-continuation' as it uses xref-pilot-char
		(xref-current-reference)
		(message "Safety check not passed.")
		)
	;; safety check passed
	(if keep
		(progn
		  (xref-re-push (point))
		  (xref-call-process 'sync 0 nil 0 0 ""
							 (format "-olcxresetrefsuffix=%d:%s" arg name) 
							 "" "")
		  (message "Done. No conflict detected. References kept on stack for manual adjustements.")
		  ))
	(xref-switch-to-marker xref-refactoring-init-marker)
	)
)

(defun xref-arg-safety-check (keep arg name)
  (let ((suff) (sc))
	(setq sc t)
	(setq suff (xref-file-name-extension (buffer-file-name)))
	(if (equal suff "java") 
		(xref-safety-check (list 'xref-arg-safety-check-finish keep arg name))
	  (xref-popCxrefs (point))
	  (if keep (xref-arg-repush-and-change-to-arg-refs arg name))
	  )
))


(defun xref-arg-exchange (poin)
"Exchange two arguments of a  function/method.

The cursor  has to be positioned  on the function/method  name (not on
the parameter itself) when  invoking this function.  The function will
interactively ask for the order of the arguments you wish to exchange,
then it inspects all references  of the function/method taken from the
tag file checking them  to be up to date. If any  of the references is
out of date, the action will  not occur and the function finishes with
an error message.   If all references are correct  the function passes
through  the  references  one  more  time exchanging  strings  of  the
selected  arguments (under the  condition that  the text  around looks
like a function definition/invocation).
"
  (interactive "d")
  (let ((name) (arg1) (arg2) (argns))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Position of parameter to exchange [ 1 - arity('%s') ] : " name) 
		 "1"
		 ))
	  (setq arg1 (string-to-int argns))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Exchange with parameter on position [ 1 - arity('%s') ] : " name) 
		 "2"
		 ))
	  (setq arg2 (string-to-int argns))
	  (xref-refactoring-init-actions 
	   (format "exchange of %s's parameters" name))
	  (xref-map-on-all-references (buffer-file-name) poin t nil 
								  "-olcxargmanip"
							 xref-detailed-refactoring-confirmations
							 nil
							 ''none-pre-fun ''exchange-arguments 
							 ''standard-post-fun arg1 arg2 
							 nil
							 (list 'xref-arg-safety-check nil nil nil)
							 )
)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun shift-arguments (arg1 arg2)
  "Shift an argument of a function/method."
  (let ((pos1beg)(pos1end)(pos2)(rrr)(abpos)(s1)
		(pos (point)) (result nil) (rflag1) (rflag2))
	(xref-pass-to-fun-lpar)
	(if (not (eq (char-after (- (point) 1)) ?\( ))
		(progn 
		  (beep) 
		  (setq result "Hmm. Does not seem as a function symbol reference.")
		  )
	  (setq abpos (point))
	  (setq rflag1 (mark-fun-argument arg1))
	  (setq pos1beg (point))
	  (setq pos1end (mark t))
	  (goto-char abpos)
	  (setq rflag2 (mark-fun-argument arg2))
	  (setq pos2 (point))
	  (if (not (and rflag1 rflag2))
		  (setq result "Argument out of range.")
		(setq s1 (buffer-substring pos1beg pos1end))
		(goto-char pos1end)
		(delete-char (- pos1beg pos1end))
		(if (eq (char-after (- (point) 1)) ?\( )
			(if (eq (char-after (point)) ?\) )
				()
			  (delete-char 1)
			  )
		  (delete-char -1)
		  )
		(goto-char abpos)
		(if (eq arg2 1)
			(if (eq (char-after (point)) ?\) )
				(insert s1)
			  (insert (concat s1 ","))
			  )
		  (setq rflag2 (mark-fun-argument (- arg2 1)))
		  (goto-char (mark t))
		  (if (eq (char-after (point)) ?\) )
				(insert (concat "," s1))
			(forward-char 1)
			(insert (concat s1 ","))
			)
		  )
		))
	(goto-char pos)
	result
))


(defun xref-arg-shift (poin)
"Shift  an argument  of a  function/method.

The cursor  has to be positioned  on the function/method  name (not on
the parameter itself) when  invoking this function.  The function will
interactively ask for the order of  the argument you wish to shift and
for  the order  where you  wish  to shift  it.  Then  it inspects  all
references  of the function/method  taken from  the tag  file checking
them to really point to  the function/method. If any of the references
is out  of date, the action  will not occur and  the function finishes
with an  error message.   If all references  are correct  the function
passes through the  references one more time moving  the string of the
selected argument  to the new  position (under the condition  that the
text around looks like a function definition/invocation).
"
  (interactive "d")
  (let ((name) (arg1) (arg2) (argns))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Position of parameter to move [ 1 - arity('%s') ] : " name) 
		 "1"
		 ))
	  (setq arg1 (string-to-int argns))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Move to position [ 1 - arity('%s') ] : " name) 
		 "2"
		 ))
	  (setq arg2 (string-to-int argns))
	  (xref-refactoring-init-actions
	   (format "move of %s's parameter" name))
	  (xref-map-on-all-references  (buffer-file-name) poin t nil 
								   "-olcxargmanip"
							 xref-detailed-refactoring-confirmations
							 nil
							 ''none-pre-fun ''shift-arguments 
							 ''standard-post-fun arg1 arg2
							 nil
							 (list 'xref-arg-safety-check nil nil nil)
							 )
)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pre-position-to-argument-delete (arg1 arg2)
  "position point and mark to the argument arg1 of a function"
  (let ((result nil) (rflag1))
	(xref-pass-to-fun-lpar)
	(if (not (eq (char-after (- (point) 1)) ?\( ))
		(progn 
		  (beep) 
		  (setq result "Hmm. Does not seem as a function symbol reference.")
		  )
	  (setq rflag1 (mark-fun-argument arg1))
	  (if (not rflag1)
		  (setq result "Argument out of range.")
		))
	result
))

(defun xref-count-newlines (string)
  (let ((ss) (res) (i) (len))
	(setq res 0)
	(setq ss string)
	(setq len (length string))
	(setq i 0)
	(while (< i len)
	  (if (eq (elt string i) ?\n) (setq res (1+ res)))
	  (setq i (1+ i))
	  )
	res
))

(defun delete-argument (arg1 arg2)
  "delete the argument arg1 of a function"
  (let ((result nil) (reg) (nlcount))
	(setq reg (buffer-substring (point) (mark t)))
	(delete-region (point) (mark t))
	(if (eq (char-after (- (point) 1)) ?\( )
		(if (eq (char-after (point)) ?\) )
			()
		  (delete-char 1)
		  )
	  (delete-char -1)
	  )
	;; preserve line number for safety check
	(setq nlcount (xref-count-newlines reg))
	(if (> nlcount 0) (newline nlcount))
	result
))


(defun xref-goto-argument-name (argn)
  (let ((ss) (maxp) (cc) (aerr))
	(setq aerr (pre-position-to-argument-delete argn nil))
	(if aerr (error "%s" aerr))
	(setq maxp (point-max))
	(if (xref-buffer-has-one-of-suffixes (buffer-file-name) 
										 xref-java-allowed-suffixes)
		(progn
		  ;; java program, check if used
		  ;; pass type name
		  (xref-skip-blank maxp)
		  (search-forward-regexp "[^A-Za-z0-9_$\\\.]")
		  ;; pass declarator prefix
		  (xref-skip-blank maxp)
		  (setq cc (char-after (point)))
		  (while (or (equal cc ?\[) (equal cc ?\]))
			(progn
			  (forward-char 1)
			  (xref-skip-blank maxp)
			  (setq cc (char-after (point)))
			  ))
		  )
	  ;; probably C program, let it be
	  )
))

(defun xref-move-to-paramater-name (argn)
  (xref-call-process 'sync 1 nil (point) 0 "" 
					 (format "-olcxgotoparname%d" argn) 
					 "" "-olcxrefsuffix=0:#")
  (xref-moveref t)
)

(defun xref-arg-delete-precheck-if-parameter-is-used (argn)
  (let ((poin) (pos) (used) (aname))
	(if (xref-last-move-was-to-definition)
		(progn
		  (setq pos (point))
		  (xref-move-to-paramater-name argn)
		  (setq used (not (xref-is-this-an-unused-symbol)))
		  (setq aname (xref-get-identifier-on-point))
		  (goto-char pos)
		  (if used
			  (progn
				(if (not (yes-or-no-p (format "Argument %s is used, delete anyway? " aname)))
					(progn
					  (xref-popCxrefs (point))
					  (error "Aborting.")
					  ))))
		  ))
	nil
))


(defun xref-non-interact-arg-delete (poin argn interact cont)
  (xref-map-on-all-references  (buffer-file-name) poin t nil 
							   "-olcxargmanip"
						 interact 
						 (list 'list ''xref-arg-delete-precheck-if-parameter-is-used argn)
						 ''pre-position-to-argument-delete
						 ''delete-argument ''standard-post-fun argn nil
						 nil
						 (list 'eval cont)
						 )
)

(defun xref-arg-delete (poin)
"Delete  an argument of  a function/method.

The cursor  has to be positioned  on the function/method  name (not on
the parameter itself) when  invoking this function.  The function will
ask you for  the order of of the argument to  delete. Then it inspects
all references of the function/method taken from the tag file checking
them to be up  to date.  If any of the references  is out of date, the
action will not occur and the function finishes with an error message.
If  all  references  are  correct  the  function  passes  through  the
references one more time deleting the string of the selected argument
(under  the condition  that  the  text around  looks  like a  function
definition/invocation).
"
  (interactive "d")
  (let ((name) (arg) (argns) (aname) (pos) (used))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Delete parameter from position [ 1 - arity('%s') ] : " name) 
		 "1"
		 ))
	  (setq arg (string-to-int argns))
	  (xref-refactoring-init-actions
	   (format "deletion of %s's parameter" name))
	  (xref-non-interact-arg-delete poin arg xref-detailed-refactoring-confirmations
		  (list 'xref-arg-safety-check nil nil nil)
		  )
)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun pre-position-to-argument-insert (arg1 arg2)
  "position point and mark to the argument arg1 of a function"
  (let ((result nil) (nn))
	(xref-pass-to-fun-lpar)
	(if (not (eq (char-after (- (point) 1)) ?\( ))
		(progn 
		  (beep) 
		  (setq result "Hmm. Does not seem as a function symbol reference.")
		  )
	  (if (eq arg1 0)
		  ()
		(setq nn (pass-fun-argument (- arg1 1)))
		(if (not (eq nn 0))
			(progn
			  (beep) 
			  (setq result "Argument out of range.")
			  )
		  )
		)
	  )
	result
))

(defun xref-last-move-was-to-definition ()
  (let ((res))
	(if (or (eq xref-move-usage ?*) (eq xref-move-usage ?!))
		(setq res t)
	  (setq res nil)
	  )
	res
))

(defun xref-last-move-was-to-definition-or-declaration ()
  (let ((res))
	(if (or (xref-last-move-was-to-definition)
			(eq xref-move-usage ?+))
		(setq res t)
	  (setq res nil)
	  )
	res
))

(defun xref-last-move-was-to-constructor-definition ()
  (let ((res))
	(if (eq xref-move-usage ?-)
		(setq res t)
	  (setq res nil)
	  )
	res
))



(defun reference-insert-text (argtext)
  (if (xref-last-move-was-to-definition-or-declaration)
	  (car argtext)
	(car (cdr argtext))
))

(defun insert-argument-with-no-check (arg1 argtext)
  "insert the argument arg1 of a function "
  (let ((pos) (abpos) (res))
	(if (eq (char-after (- (point) 1)) ?\) )
		(progn
		  (backward-char 1)
		  (insert-char ?, 1)
		  (setq abpos (point))
		  (insert (reference-insert-text argtext))
		  )
	  (setq abpos (point))
	  (insert (reference-insert-text argtext))
	  (setq pos (point))
	  (search-forward-regexp "[^ ]")
	  (if (eq (char-after (- (point) 1)) ?\) )
		  (goto-char pos)
		(goto-char pos)
		(insert-char ?, 1)
		)
	  )
))

(defun insert-argument (arg1 argtext)
  "insert the argument arg1 of a function "
  (let ((res) (used) (nname) (oname) (pilot) (moveusage))
	(setq res nil)
	(setq pilot xref-pilot-char)
	(setq moveusage xref-move-usage)
	(insert-argument-with-no-check arg1 argtext)
	;; here you should check that the new argument is not used
	;; but be carefull! preserve xref-pilot-char for example!
	(setq xref-insert-arg-renaming-new-name nil)
	;;(if (xref-last-move-was-to-definition-or-declaration)
	(if (xref-last-move-was-to-definition)
		(progn
		  (xref-current-reference)
		  (xref-move-to-paramater-name arg1)
		  (setq oname (xref-identifier-after-point))
		  (setq used (not (xref-is-this-an-unused-symbol)))
		  (if used
			  (progn
				(ding 1)
				(setq nname (read-from-minibuffer 
							 "New parameter conflicts with existing symbol, enter another name: " oname))
				(insert nname)
				(xref-delete-pending-ident)
				(setq xref-insert-arg-renaming-new-name nname)
				(setq res "New name inserted without further checks.")
				)
			)
		  ))
	(setq xref-pilot-char pilot)
	(setq xref-move-usage moveusage)
	res
))


(defun xref-arg-insert (poin)
"Add an  argument to a function/method.

The cursor  has to be positioned  on the function/method  name (not on
the parameter itself) when  invoking this function.  The function will
ask you  for the  order of argument  to insert  and for the  string to
insert as  default argument value.   Then it identifies the  symbol on
the point, pushes all its references on the reference stack, checks if
all the references point really  to this symbol, if the references are
out of date, the function provides an error message. If all references
are correct  it inspects all  references one more time,  inserting the
comma and the default string at the argument position.
"
  (interactive "d")
  (let ((name) (arg) (argns) (text) (textdef) (textval))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Insert parameter on position [ 1 - arity('%s') ] : " name)
		 "1"
		 ))
	  (setq arg (string-to-int argns))
	  (setq textdef (read-from-minibuffer 
		 "String to insert as declaration of the parameter: "
		 "int arg"
		 ))
	  (setq textval (read-from-minibuffer 
		 "String to insert as default value of the parameter: "
		 "0"
		 ))
	  (setq text (list 'quote (cons textdef (cons textval nil))))
	  (xref-refactoring-init-actions
	   (format "insertion of %s's parameter" name))
	  (xref-map-on-all-references  (buffer-file-name) poin t nil 
								  "-olcxargmanip"
							 xref-detailed-refactoring-confirmations
							 nil
							 ''pre-position-to-argument-insert
							 ''insert-argument ''standard-post-fun arg text
							 nil
							 (list 'xref-arg-safety-check t arg name))
)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-arg-push (poin)
"Push references to  an argument of a function/method.

This  function asks  you for  the order  of the  argument you  wish to
browse, then  it identifies  the symbol on  the point, pushes  all its
references on  the reference stack,  moves on the definition  place of
the symbol  and them moves to  its n-th argument  (under the condition
that the the text here  looks like a function definition/invocation) .
Then you  can inspect next/previous  reference of the symbol  with the
'xref-nextref' and 'xref-previousref'  functions. Each time the cursor
will be moved at the n-th argument position.
"
  (interactive "d")
  (let ((name) (argn) (argns))
	(if (eq (buffer-file-name) nil)
		(message "You should edit C or Java file when using Xref")
	  (xref-delete-windows)
	  (setq name (xref-get-identifier-on-point))
	  (setq argns (read-from-minibuffer 
		 (format 
		  "Order of argument to  browse [ 1 - arity('%s') ] : " name) 
		 "1"
		 ))
	  (setq argn (string-to-int argns))
	  (push-references 'sync (buffer-file-name) poin t nil "-olcxpush" argn 
					   '(progn 
						  (xref-moveref t)
						  (message "Pushed. Use 'Previous ref' and 'Next refs' to inspect the references.")
						  )))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-goto-source-window ()
  (let ((clw) (xlw) (srw) (ctw) (bfn))
	(setq clw (get-buffer-window "*compilation*" nil))
	(if clw
		(select-window clw)
	  (setq ctw (get-buffer-window (xref-class-tree-buffer-name) nil))
	  (if ctw
		  (select-window ctw)
		(setq xlw (get-buffer-window (xref-list-buffer-name) nil))
		(if xlw
			(select-window xlw)
		  (setq srw (get-buffer-window (xref-symbol-resolution-buffer-name) nil))
		  (if srw
			  (select-window srw)
			(other-window 1)
		  ))))
	(other-window -1)
))

(defun renaming (new-name dummy)
  (let ((pos (point)))
	(kill-region (point) (mark t))
	(insert new-name)
	(goto-char pos)
	nil
))

(defun xref-definition-reference ()
"Find a definition reference on the top of the stack and go there."
  (interactive "")
  (xref-goto-source-window)
  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
  (xref-moveref t)
  (xref-update-reflist-window)
)

(defun xref-caller-reference ()
"Go to the position from where references on the top of the stack were pushed."
  (interactive "")
  (xref-goto-source-window)
  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotocaller" "" "")
  (xref-moveref t)
  ;; commented out, as it do not move current reference, (for the moment)
;;  (xref-update-reflist-window)
)

(defun xref-current-reference ()
"Go to the current reference of the stack."
  (interactive "")
  (xref-goto-source-window)
  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotocurrent" "" "")
  (xref-moveref nil)
)

(defun listrefs-continuation ()
  (if (eq xref-pilot-char ?\; ) 
	  (progn
		(set-xref-listrefs-dialog-maps)
		(xref-resize3-windows)
	))
)

(defun xref-number-of-windows ()
  (let ((owin) (res))
	(setq res 1)
	(setq owin (buffer-name))
	(other-window -1)
	(while (not (equal owin (buffer-name)))
	  (progn
		(other-window -1)
		(setq res (+ res 1))
		))
	res
))

(defun xref-list-top-references ()
"List references on the top of the reference stack."
  (interactive "")
  (let ((srw) (winn) (source-win-exists t))
	;; first of all arrange windows
	(setq srw (get-buffer-window (xref-symbol-resolution-buffer-name) nil))
	(setq winn (xref-number-of-windows))
	(if (> winn 1) (xref-delete-window (xref-list-buffer-name)))
	(if srw
		(progn
		  ;; there is yet symbol resolution window, enter three windows mode
		  (select-window srw)
		  (other-window -1)
		  (split-window-vertically)
		  (if (equal (buffer-name) (xref-symbol-resolution-buffer-name))
			  (progn
				(other-window -1)
				(setq source-win-exists nil)
				))
		  ))
  (xref-call-process xref-reflist-synchro 0 nil (point) 0 "" 
					 "-olcxlisttop" "" "")
  (setq xref-list-buffer-content 'reference-list)
  (listrefs-continuation)
  (if source-win-exists (xref-current-reference))
))

(defun xref-safety-check-prompt ()
  (if (yes-or-no-p "Done. Save modified files and make safety check? ")
	  (progn
		(xref-save-some-buffers t)
		t
		)
	(message "** Changes unsaved, save buffers manually **")
	nil
	)
)

(defun xref-safety-check-finish ()
  ;; must be in this order as 'listrefs-continuation' is checking xref-pilot-char
  (if (eq xref-pilot-char ?\;)
	  (progn
		(listrefs-continuation)
		(xref-current-reference)
		))
)

(defun xref-refactoring-final-safety-check-finish ()
  ;; must be in this order as 'listrefs-continuation' is checking xref-pilot-char
  (if (eq xref-pilot-char ?\;)
	  (progn
		(listrefs-continuation)
		(xref-current-reference)
		)
	(xref-refactoring-final-update-actions)
	)
)


(defun xref-safety-check (continuation)
  (let ((res) (skip-check1))
	(setq res t)
	(if (xref-safety-check-prompt)
		(progn
		  ;; should be via markers?
		  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
		  ;; goto definition without checking the symbol name
		  (xref-moveref nil)
		  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxsafetycheckinit" 
							 "-olcxrefsuffix=0:#" "")
		  (if (xref-buffer-has-one-of-suffixes (buffer-file-name) '(".java"))
			  (setq skip-check1 t)
			(setq skip-check1 nil)
			)
		  (if (not skip-check1)
			  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxsafetycheck1" 
								 "-olcxrefsuffix=0:#" "")
			)
		  (if (or skip-check1 (eq xref-pilot-char ?\*))
			  (progn
				(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)
				(message "Checking new references.")
				(push-references 'sync (buffer-file-name) (point) t nil "-olcxsafetycheck2" 0 
								 (list 'eval continuation))
				;; NO CODE HERE !!!
				)
			(eval continuation)
			;; NO CODE HERE !!!
			)
		  )
	  (message "** Changes unsaved, save buffers manually **")
	  )
))

(defun xref-trivial-safety-check ()
  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxsafetycheckinit" 
					 "-olcxrefsuffix=0:#" "")
  (xref-non-interactive-update-references xref-fast-updates-inside-refactorings)
  (push-references 'sync (buffer-file-name) (point) t nil "-olcxsafetycheck2" 0 
				   '(xref-safety-check-finish))
)

(defun xref-undoable-move-file (new-fname)
  (xref-multifile-undo-add-buffer-write)
  (xref-delete-file (buffer-file-name nil))
  (xref-write-file new-fname t)
)

(defun xref-class-rename-finish (old-name new-name)
  (let ((ofname) (fname) (fsname) (new-fname) (dirname) (safetycheckopt)
		(fmove))
	;; should be via markers?
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
	(xref-moveref nil)
	(setq ofname (buffer-file-name))
	(setq fname (xref-file-last-name ofname))
	(setq dirname (xref-file-directory-name ofname))
	(setq fsname (xref-file-name-without-suffix fname))
	(setq safetycheckopt "-olcxsafetycheck2")
	(if (equal fsname old-name)
		(progn
		  (setq new-fname (concat new-name ".java"))
		  (setq fmove t)
;;		  (setq fmove (yes-or-no-p (format "Move to file '%s' ? "
;;                                   new-fname)))
		  (if fmove
			  (progn
				(xref-undoable-move-file new-fname)
				(setq safetycheckopt 
					  (concat
					   (format 
						"%s -olcxsafetycheckmovedfile \"%s\" \"%s/%s.java\""
						safetycheckopt ofname dirname new-name)
					   ;; (format " -olcxsafetycheckmovedblock=0:500000:0")
					   ))
		  ))))
	(if (xref-safety-check-prompt)
		(progn
		  ;; re-create references, it is safer
		  (xref-non-interactive-update-references nil)
		  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxsafetycheckinit" 
						   "-olcxrefsuffix=0:#" "")
		  (message "Checking new references.")
		  (push-references 'sync (buffer-file-name) (point) t nil safetycheckopt 0 '(xref-safety-check-finish))
		  ))
))

(defun xref-rename-class (poin)
"Safe  renaming of a Java class. 

POINT  should   be  positioned  on  the  class   name.   The  function
interactively asks for  the new name, then it  inspects all references
of the class taken from the  tag file checking them to really point to
the class name. If any of  the references is out of date, the renaming
will not occur and function finishes with an error.  If all references
are O.K.   the function  passes through the  references one  more time
replacing the old class name with  the new one. If the class is stored
in a file  of the same name,  then the buffer will be  saved into file
with the new name and old file is deleted.

Optionally, a  safety check  is applied to  detect name  clashes.  The
safety check will push references  of the renamed class and will check
if  there  is a  one  to one  correspondance  with  references of  the
original class.  If  not, it means that the  refactoring is unsafe and
that there is a name clash. Differences between references can be then
browsed.

This  refactoring  is  stored  in  the  undo  list  as  two  different
refactorings.  First  are stored  all necessary text  manipulations in
sources and then (as second refactoring) is stored moving of the file.
If you  are undoing this refactoring you  will need to undo  it as two
successive refactorings.

"
  (interactive "d")
  (let ((tcount) (prompt) (new-name) (old-name))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
		"You should edit a Java file when using this function.")
	(xref-trivial-pre-checks "-tpchrenameclass")
	(setq old-name (xref-get-identifier-on-point))
	(setq new-name (read-from-minibuffer 
					(format "Rename '%s' to : " old-name) old-name))
	(xref-refactoring-init-actions
	 (format "rename class text from %s to %s" old-name new-name))
	(setq tcount (xref-map-on-all-references  (buffer-file-name) poin t nil "-olcxrename"
										xref-detailed-refactoring-confirmations
										nil
										''none-pre-fun ''renaming 
										''none-post-fun new-name nil
										nil
										(cons 'xref-class-rename-finish 
											  (cons old-name (cons new-name nil)))))
))

(defun xref-rename-safety-check-finish ()
  (if (equal xref-pilot-char ?\;)
	  (progn
		;; safety check does not passed, order of following is important
		(listrefs-continuation)
		(xref-current-reference)
		)
	;; passed
	(xref-switch-to-marker xref-refactoring-init-marker)
	(message "All Done. No conflict detected.")
	)
)

(defun xref-rename-finish ()
  (xref-safety-check (list 'xref-rename-safety-check-finish))
)

(defun xref-rename-precheck-type ()
  (let ((tt))
	(xref-call-process 'sync 0 nil 0 0 "" "-olcxtoptype" "" "")
	(setq tt xref-last-message)
	(if (equal tt "class")
		(if (yes-or-no-p "Use 'rename class' for classes. Interrupt this tentative? ")
			(error "Exiting. Please, retry with 'rename class' function."))
	  (if (equal tt "package")
		(if (yes-or-no-p "Use 'rename package' for packages. Interrupt this tentative? ")
			(error "Exiting. Please, retry with 'rename package' function."))
		))
))

(defun xref-rename (poin)
"Safe  renaming  of a  symbol.

This function  interactively asks for  the new name, then  it inspects
all  references of the  symbol checking  them to  really point  to the
symbol. If  any of the  references is out  of date, renaming  will not
occur and function finishes with an error.  If all references are O.K.
the function passes through the references one more time replacing old
symbol name with the new one.

Optionally, a  safety check  is applied to  detect name  clashes.  The
safety check will push references of the renamed symbol and will check
if  there  is a  one  to one  correspondance  with  references of  the
original symbol.  If not, it  means that the refactoring is unsafe and
that there is a name clash. Differences between references can be then
browsed.

This  function  is  intended  to  rename symbols  such  as  variables,
functions/methods, parameters,  etc. However, it can be  used also for
Java package and class renaming, but in those cases safety checks will
not  work correctly  because it  does not  change file  structure. For
class renaming use 'xref-rename-class' function.
"
  (interactive "d")
  (let ((tcount) (prompt) (new-name) (old-name))
	(setq old-name (xref-get-identifier-on-point))
	(setq new-name (read-from-minibuffer 
					(format "Rename '%s' to : " old-name) old-name))
	(xref-refactoring-init-actions
	 (format "renaming of %s to %s" old-name new-name))
	(setq tcount (xref-map-on-all-references  (buffer-file-name) poin t nil "-olcxrename"
										xref-detailed-refactoring-confirmations
										nil
										''none-pre-fun ''renaming 
										''none-post-fun new-name nil
										'(xref-rename-precheck-type)
										'(xref-rename-finish)
))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-toggle-next-push-manual ()
"This function sets the  `xref-next-push-manual' flag to true. 

This  will  cause  that   the  next  reference  pushing  action  (push
references, list,  rename, arg.  manipulation, ...)  will  first pop a
manual symbol resolution menu as if the symbol is ambiguous.  This can
be usefull if you wish  to pass around the automatic symbol resolution
and you wish  to get for example references of all  the symbols of the
given name,  not only those  of the symbol  pointed by point.  This is
also  a way how  to see  inheritance hierarchy  of static  methods and
fields (which are never ambiguous).
"
  (interactive "")
  (setq xref-next-push-manual t)
  (message "Next reference pushing action will used manual resolution.")
)


(defun xref-pushCxrefs-force-ctree (poin)
"Push references with forced symbol resolution.

This function  will force  the class tree  resolution dialog  for next
push and will then call 'Push Refernces' function.
"
  (interactive "d")
  (xref-toggle-next-push-manual)
  (xref-pushCxrefs poin)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                            Multifile UNDO
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar xref-multifile-undo-state nil)

(defun xref-this-buffer-undo-state ()
  (undo-boundary)
  (cdr buffer-undo-list)
)

(defun xref-is-buffer-undoable (bname)
  (let ((bnamestart) (res))
	(if (not bname)
		(setq res nil)
	  (setq bnamestart (substring bname 0 1))
	  (if (or (equal bnamestart " ") (equal bnamestart "*"))
		  (setq res nil)
		(if (xref-buffer-has-one-of-suffixes bname xref-undo-allowed-suffixes)
			(setq res t)
		  (setq res nil)
		  )))
	res
))

(defun xref-editor-undo-state ()
  (let ((state) (buffs) (bst) (buf))
	(save-excursion
	  (setq state nil)
	  (setq buffs (buffer-list))
;;	  (setq buffs (append buffs (cons (current-buffer) nil)))
	  (while buffs
		(setq buf (car buffs))
		(setq buffs (cdr buffs))
		(if (xref-is-buffer-undoable (buffer-file-name buf))
			(progn
			  (set-buffer buf)
			  (setq bst (xref-this-buffer-undo-state))
			  (setq state (cons (cons buf bst) state))
		  ))))
	state
))

(defun get-corresponding-undo (ulist buf)
  (let ((ul) (found) (res))
	(setq ul ulist)
	(setq found nil)
	(setq res nil)
	(while (and ul (not found))
	  (setq found (eq (car (car ul)) buf))
	  (if (not found)
		  (setq ul (cdr ul))
		)
	  )
	(if found
		(setq res (cdr (car ul)))
	  )
	res
))	  

(defun xref-save-modified-files-with-question-and-error (flag message)
  (if (yes-or-no-p (concat message "Save modified buffers? "))
	  (progn
		(xref-save-some-buffers flag)
		(error "Changes saved.")
		)
	(error "Not saved.")
))

(defun xref-save-modified-files-with-question (flag message)
  (if (yes-or-no-p (concat message "Save modified buffers? "))
	  (progn
		(xref-save-some-buffers flag)
		(message "Changes saved.")
		)
	(message "Not saved.")
))

(defun undo-changes-until (bundo interact-flag)
  (let ((cc) (aa) (iinteract) (loop))
	(setq iinteract interact-flag)
	(setq loop t)
	(while (and pending-undo-list (not (eq pending-undo-list bundo)))
	  (undo-more 1)
	  (if iinteract
		  (progn
			(message "Continue undoing [y/n/a]?")
			(setq aa (xref-get-single-yes-no-event))
			)
		(setq aa 'answer-yes)
		)
	  (if (eq aa 'answer-no)
		  (xref-save-modified-files-with-question-and-error nil "** Undoing breaked. ")
		)
	  (if (eq aa 'answer-all)
		  (setq iinteract nil)
		)
	  (message "")
	  (undo-boundary)
	  )
	(if (not (eq pending-undo-list bundo))
		(error "Not enough undo information available (check 'undo-limit' and 'undo-strong-limit' variables)!") 
	  )
	iinteract
))

(defun xref-undo-single-buffer (buf ustate cont-ustate interact-flag)
  (let ((bnamestart) (bufundo) (iinteract))
	(setq iinteract interact-flag)
	(if (not (buffer-name buf))
		(progn
		  (if (not (yes-or-no-p (format "A killed buffer can't be undone, continue? ")))
			  (xref-save-modified-files-with-question-and-error nil "Aborted. ")			
			)
		  )
	  (if (xref-is-buffer-undoable (buffer-file-name buf))
		  (progn
			(switch-to-buffer buf)
			(setq bufundo (get-corresponding-undo ustate buf))
			(if cont-ustate
				(setq pending-undo-list (get-corresponding-undo 
										 cont-ustate buf))
			  (undo-start)
			  (if bufundo (undo-more 1))
			  )
			(setq iinteract (undo-changes-until bufundo iinteract))
			))
	  )
	iinteract
))

(defun xref-undo-until-undo-state (ustate cont-ustate interact)
  (let ((buffs) (buf) (buffs2) (bufundo) (iinteract))
	(setq iinteract interact)
	(setq buffs (buffer-list))
	;; first process bufferes without undo-state
	(while buffs
	  (setq buf (car buffs))
	  (setq buffs (cdr buffs))
	  (setq bufundo (get-corresponding-undo ustate buf))
	  (if (not bufundo)
		  (setq iinteract (xref-undo-single-buffer buf ustate cont-ustate iinteract))
		)
	  )
	;; that those from list
	(setq buffs2 ustate)
	(while buffs2
	  (setq buf (car (car buffs2)))
	  (setq buffs2 (cdr buffs2))
	  (setq bufundo (get-corresponding-undo ustate buf))
	  (if bufundo
		  (setq iinteract (xref-undo-single-buffer buf ustate cont-ustate iinteract))
		)
	  )
))

(defun cut-long-refactoring-undo-list ()
  (let ((ul) (deep))
	(if (> (length xref-multifile-undo-state) xref-multifile-undo-deep)
		(progn
		  (setq ul xref-multifile-undo-state)
		  (setq deep xref-multifile-undo-deep)
		  (while (and ul (> deep 0))
			(setq deep (- deep 1))
			(setq ul (cdr ul))
			)
		  (if ul
			  (setcdr ul nil)
			))
	  )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun xref-undo-last-refactoring () 
"Undo a suite of refactorings.

This function  undoes changes made  in multiple buffers.   Buffers are
switched at beginnings of  refactorings. The undo itself is considered
as a refactoring,  so further invocations can be  used to re-do undone
changes.   It uses  standard  Emacs  undo mecanism  which  may not  be
appropriate for  such a  massive undoing, in  case of  problems verify
your 'undo-strong-limit'  and 'undo-limit' variable  settings.  It has
also some  problems if you have  killed a buffer or  saved buffer with
changed name. 

NOTE!  Do  not kill  any buffer if  you are using  undoing! Xrefactory
memorize  states  of all  opened  buffers  at  the beginning  of  each
refactoring. If  any of opened  buffers is killed in  between, undoing
will not work!
"
  (interactive "")
  (let ((undolist) (cloop) (cont-undo-state) (aa) (cbuff) (interact) 
		(u-type) (next-ask))
	(setq interact xref-detailed-refactoring-confirmations)
	(setq cbuff (current-buffer))
	(if (eq xref-multifile-undo-state nil)
		(error "** Refactorings undo stack is empty **")
	  )
	(set-marker xref-undo-marker (point))
	(setq cont-undo-state nil)
	(setq cloop t)
	(setq undolist xref-multifile-undo-state)
	(xref-delete-windows)
	(setq cloop (yes-or-no-p 
				 (format "Really undo %s? " (car (car undolist))
				  )))
	(while cloop
	  (xref-multifile-undo-set-buffer-switch-point (format "an undone refactoring (redoing %s)" (car (car undolist))))
	  (setq next-ask t)
	  (setq u-type (car (cdr (car undolist))))
	  (if (equal u-type "move-buffer")
		  (progn
			(setq interact (xref-undo-file-moving (cdr (cdr (car undolist))) interact))
;;			(setq next-ask nil)
			)
		(if (equal u-type "move-dir")
		  (progn
			(setq interact (xref-undo-dir-moving (cdr (cdr (car undolist))) interact))
;;			(setq next-ask nil)
			)
		  (setq interact (xref-undo-until-undo-state (cdr (cdr (car undolist))) cont-undo-state interact))
		  (setq cont-undo-state (cdr (cdr (car undolist))))
		))
	  (setq undolist (cdr undolist))
	  (if (eq undolist nil)
		  (setq cloop nil)
		(if (not next-ask)
			(setq cloop t)
		  (sit-for 0.01)   ;; refresh screen
		  (setq cloop (yes-or-no-p 
					   (format 
						"The refactoring is undone, continue by undoing %s? "
						(car (car undolist))
					  )))))
	  )
	(switch-to-buffer cbuff)
	(sit-for 0)
	(xref-switch-to-marker xref-undo-marker)
	(xref-save-modified-files-with-question t "Undone. ")
))


(defun xref-undo-file-moving (data interact)
  (let ((buf) (fname) (confirmed))
	(setq fname (car data))
	(setq buf (car (cdr data)))
	(set-buffer buf)
	(if interact
		(progn
		  (switch-to-buffer buf)
		  (setq confirmed (yes-or-no-p (format "Move to file %s ? " fname)))
		  )
	  (setq confirmed t)
	  )
	(if confirmed
		(progn
		  (xref-undoable-move-file fname)
		  ))
	)
  interact
)

(defun xref-multifile-undo-add-buffer-write ()
  (let ((comment))
;;	(if xref-multifile-undo-state
;;		(setq comment (car (car xref-multifile-undo-state)))
	(setq comment (format "moving file %s" (xref-file-last-name (buffer-file-name nil))))
;;	  )
	(setq xref-multifile-undo-state 
		  (cons 
		   (list
			comment
			"move-buffer"
			(buffer-file-name nil)
			(current-buffer)
			)
		   xref-multifile-undo-state
		   ))
	(cut-long-refactoring-undo-list)
;;(insert (format "multifile-undo-state : %S\n\n" xref-multifile-undo-state))
))


(defun xref-undo-dir-moving (data interact)
  (let ((olddir) (newdir) (confirmed))
	(setq olddir (car data))
	(setq newdir (car (cdr data)))
	(if interact
		(progn
		  (setq confirmed (yes-or-no-p (format "Move directory %s to %s ? " 
											   (xref-file-last-name newdir)
											   (xref-file-last-name olddir))))
		  )
	  (setq confirmed t)
	  )
	(if confirmed
		(progn
		  (xref-undoable-move-directory newdir olddir)
		  ))
	)
  interact
)

(defun xref-multifile-undo-add-move-directory (olddir newdir)
  (let ((comment))
;;	(if xref-multifile-undo-state
;;		(setq comment (car (car xref-multifile-undo-state)))
	(setq comment (format "moving directory %s to %s" 
						  (xref-file-last-name olddir)
						  (xref-file-last-name newdir)))
;;	  )
	(setq xref-multifile-undo-state 
		  (cons 
		   (list
			comment
			"move-dir"
			olddir
			newdir
			)
		   xref-multifile-undo-state
		   ))
	(cut-long-refactoring-undo-list)
;;(insert (format "multifile-undo-state : %S\n\n" xref-multifile-undo-state))
))


(defun xref-multifile-undo-set-buffer-switch-point (comment)
"Set a multifile undo 'change-point'.

This function memorizes  the state of undo informations  of all opened
buffers.  An invocation of `xref-undo-last-refactoring' will then undo
changes under  those points  in all those  buffers.  The COMMENT  is a
string commenting before which action the state is memorized.
"
;;  (interactive "")
  (setq xref-multifile-undo-state 
		(cons 
		 (cons
		  comment
		  (cons
		   "standard"
		   (xref-editor-undo-state)
		  ))
		 xref-multifile-undo-state
		 ))
  (cut-long-refactoring-undo-list)
;;(insert (format "multifile-undo-state : %S\n\n" xref-multifile-undo-state))
)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;        IDE     ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; the following function can be used in order to get an xref
;; environment variable set by "-set" option inside the
;; .xrefrc file (its value depends on active project selection).

(defun xref-get-env (name) 
  (let ((res))
  (xref-call-process 'sync 1 nil 0 0 "" "-get" name "-olcxrefsuffix=0:")
  (if (eq xref-pilot-char ?\*) 
	  (setq res xref-last-message)
	(setq res nil)
	)
  res
))

(defun xref-get-env-or-error (name) 
  (let ((res))
	(setq res (xref-get-env name))
	(if (not res) (error "Can't get value of '%s'" name))
  res
))

;; this function can be used to run an user defined non-interactive
;; command, the value of 'name' argument has to be set in .xrefrc file.

(defun xref-compile-function (action) 
  (let ((comm) (name) (arg) (bfile) (shcomm) (cwin) (sdir))
	(require 'compile)
	(xref-delete-windows)
	(if (not (featurep 'compile))
		(error "Your (X)Emacs do not contain the 'compile.el' package, install it first")
	  )
	(if (eq action 'compilefile)
		(progn
		  (setq name "compilefile")
		  (setq arg (xref-file-last-name (buffer-file-name)))
		  )
	  (if (eq action 'compiledir)
		  (progn
			(setq name "compiledir")
			(setq arg (backslashify-name (xref-file-directory-name (buffer-file-name))))
			)
		(if (eq action 'compileproject)
			(progn
			  (setq name "compileproject")
			  (xref-call-process 'sync 1 nil 0 0 "" "-olcxgetprojectname" 
								 "-olcxrefsuffix=0:" "")
			  (if (eq xref-pilot-char ?\*)
				  (setq arg xref-last-message)
				(error "No active project selected")
				)
			  ))))
	(setq sdir default-directory)
	(setq comm (format (xref-get-env name) (xref-optionify-string arg "\"")))
	(if (and (not xref-always-batch-file) (not (string-match "\n" comm)))
		(setq shcomm comm)
	  (setq bfile (xref-create-batch-file default-directory comm))
	  (if (eq xref-platform 'windows)
		  (setq shcomm bfile)
		(setq shcomm (format "%s %s" xref-shell bfile))
		))
	(setq xref-last-referencing-action 'compilation)
	(setq xref-ide-last-compile-command action)
	(compile shcomm)
	(sleep-for 0.5)
	(setq cwin (get-buffer-window "*compilation*"))
	(if cwin
		(progn
		  (select-window cwin)
		  (end-of-buffer)
		  (setq default-directory sdir)
		  (setq tab-width 8) 	;; as usual on terminals
		  ))
	(bury-buffer (get-buffer "*compilation*"))  ;; hope it has yet a window
))

(defun xref-ide-compile () 
"Repeat last  compilation command.

This is  one of \"Compile  file\", \"Compile Directory\"  or \"Compile
Project\" command.
"
  (interactive "")
  (xref-compile-function xref-ide-last-compile-command)
)

(defun xref-ide-compile-file () 
"Compile file. 

The compile  command must  be specified in  your .xrefrc file  via the
'-set compilefile <command>' option.  If the <command> string contains
the %s character sequence, it will be replaced by the currently edited
file name.   Read xref  manual file for  more info about  -set option.
This function  actually only calls the (X)Emacs  compile function with
the appropriate command string. You need to have the 'compile' package
installed in your editor.

Xref-Tech is  not author of  the compile.el package which  is released
under GPL.  All credits for making the 'compile' package belong to its
(respective) authors.
"
  (interactive "")
  (xref-compile-function 'compilefile)
)

(defun xref-ide-compile-dir () 
"Compile  directory. 

The compile  command must  be specified in  your .xrefrc file  via the
'-set compiledir  <command>' option. If the  <command> string contains
the  %s  character  sequence,  it  will be  replaced  by  the  current
directory  name.  Read  xref  manual  file for  more  info about  -set
option.   This  function  actually  only calls  the  (X)Emacs  compile
function with  the appropriate  command string. You  need to  have the
'compile' package installed in your editor.

Xref-Tech is  not author of  the compile.el package which  is released
under GPL. All credits for making the 'compile' package belong to its
(respective) authors.
"
  (interactive "")
  (xref-compile-function 'compiledir)
)

(defun xref-ide-compile-project () 
"Compile  project. 

The compile  command must  be specified in  your .xrefrc file  via the
'-set  compileproject  <command>'  option.    You  need  to  have  the
'compile' package installed in your editor.

Xref-Tech is  not author of  the compile.el package which  is released
under GPL. All credits for making the 'compile' package belong to its
(respective) authors.
"
  (interactive "")
  (xref-compile-function 'compileproject)
)

(defun xref-ide-previous-error ()
"Move to  the previous  error of the  last compilation.  

This function actually calls `next-error' with argument -1. See also
`xref-previousref'.

"
  (interactive "")
  (setq xref-last-referencing-action 'compilation)
  (next-error -1)
)

(defun xref-ide-next-error ()
"Move  to the  next  error  of the  last  compilation.  

This  function actually calls `next-error' with argument 1. See also
`xref-nextref'.

"
  (interactive "")
  (setq xref-last-referencing-action 'compilation)
  (next-error 1)
)

;; kill the process associated to this buffer (if any)
(defun xref-kill-this-buffers-process ()
  (let ((pl) (pp))
	(setq pl (process-list))
	(while pl
	  (setq pp (car pl))
	  (setq pl (cdr pl))
	  (if (equal (process-buffer pp) (current-buffer))
		  (progn
			(kill-process pp)
			(sleep-for 0.1)
			(discard-input)
			(while (eq (process-status pp) 'run)
			  (message "Waiting until process dies.")
			  (sleep-for 0.1)
			  (discard-input)
			  )
			(message "Done.")
			(setq pl nil)
			)))
))

(defun xref-escape-dq (ss)
  (let ((res) (mp) (lmp))
	(setq res ss)
	(setq lmp 0)
	(setq mp (string-match "\"" res))
	(while mp
	  (setq res (format "%s\\%s" (substring res 0 mp) (substring res mp)))
	  (setq mp (string-match "\"" res (+ mp 2)))
	  )
	res
))

(defun xref-cr-echos-commands (command)
  (let ((res) (mp) (lmp) (line))
	(setq res "")
	(setq lmp 0)
	(setq mp (string-match "\n" command))
	(while mp
	  (setq line (substring command lmp mp))
	  ;; do not echo exit and empty lines to not damage return value
	  (if (or (equal line "exit") (xref-exact-string-match "[ \t]*" line))
		  (setq res (format "%s\n%s" res line))
		(setq res (format "%s\necho \">%s\"\n%s" res (xref-escape-dq line) line))
		)
	  (setq lmp (+ mp 1))
	  (setq mp (string-match "\n" command lmp))
	  )
	;; do not this, command ends with \n (last line is empty)
	;; following lines damages return value
	;;(setq line (substring command lmp))
	;;(setq res (format "%s\necho %s\n%s" res line line))
	res
))


(defun xref-create-batch-file (bdir command)
  (let ((res) (cc))
	(setq cc "")
	(if (and (eq xref-platform 'windows)
			 (> (length bdir) 1)
			 (equal (elt bdir 1) ?:))
		(setq cc (format "%s%s:\n" cc (substring bdir 0 1)))
	  )
	(setq cc (format "%scd %s\n%s\n" cc (backslashify-name bdir) command))
	(if (not (eq xref-platform 'windows))
		(setq cc (xref-cr-echos-commands cc))
	  )
	(get-buffer-create " *xref-run-command*")
	(set-buffer " *xref-run-command*")
	(xref-erase-buffer)
	(insert cc)
	(write-region 1 (+ (buffer-size) 1) xref-run-batch-file)
	(kill-buffer nil)
	(setq res xref-run-batch-file)
	res
))
  

;; this function can be used to run an xref user defined interactive
;; commands, the value of 'name' argument has to be set in .xrefrc file.
(defun xref-run-function (name skip-one-win)
  (let ((bb) (rc) (args) (ww) (bdir) (cc) (obuffername) (obuffer) (bfile) (command))
	(setq bdir default-directory)
	(setq obuffer "xref-run")
	(xref-call-process 'sync 1 nil 0 0 "" "-olcxgetprojectname" 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\*)
		(progn
		  (require 'comint)
		  (if (not (featurep 'comint))
			  (error "Your (X)Emacs does not contain 'comint.el' package, install it first")
			)
		  (setq rc (xref-get-env name))
		  (if rc
			  (progn
				(if (equal name xref-run-this-option)
					(setq command (format rc (xref-get-current-class 
											  "-olcxcurrentclass -noerrors" 
											   nil)))
				  (setq command rc)
				  )
				(setq bfile (xref-create-batch-file bdir command))
				(if (eq xref-platform 'windows)
					(setq args (cons bfile nil))
				  (setq args (cons xref-shell (cons bfile nil)))
				  )
				(if skip-one-win
					(other-window 1)
				  )
				(split-window-vertically)
				(other-window 1)
				(setq obuffername (format "*%s*" obuffer))
				(switch-to-buffer obuffername t)
				(xref-kill-this-buffers-process)
				;;(get-buffer-create obuffername)
				;;(set-buffer obuffername)
				;;(setq ww (xref-display-buffer obuffername))
				;;(select-window ww)
				(goto-char (point-max))
				(xref-erase-buffer)
				(setq default-directory bdir)
				(setq xref-ide-last-run-command name)
				(apply 'make-comint obuffer (car args) nil (cdr args))
				)
			)
		  )
	  (error "No project Selected")
	)
))

(defun xref-ide-run ()
"Repeat last of run commands (i.e.  one of run1, run2, ...). 

If  non of  run  commands  was previously  executed,  run the  command
specified vith the  '-set run <command>' option in  your .xrefrc file.
This  function actually  only  calls the  `make-comint' function  with
command  string taken from  .xrefrc configuration  file.  You  need to
have the  'comint' package installed in  your editor in  order to make
this function work.

Xref-Tech is  not author  of the comint.el  package which  is released
under  GPL.    All  credits  for   'comint'  package  belong   to  its
(respective) authors.
"
  (interactive "")
  (xref-delete-windows)
  (xref-run-function xref-ide-last-run-command nil)
)

(defun xref-ide-run-this ()
"Run the command specified by  '-set runthis <command>' option in your
.xrefrc file.   The <command>  can contain '%s'  string which  will be
replaced by curently edited class  name.  It also sets this command to
be executed by default run function 'xref-ide-run'.
"
  (interactive "")
  (xref-delete-windows)
  (xref-run-function xref-run-this-option nil)
)

(defun xref-ide-run1 ()
"Run the  command specified  by '-set run1  <command>' option  in your
.xrefrc file.  It also sets this command to be executed by default run
function 'xref-ide-run'.
"
  (interactive "")
  (xref-delete-windows)
  (xref-run-function "run1" nil)
)


(defun xref-ide-run2 ()
"Run the  command specified  by '-set run2  <command>' option  in your
.xrefrc file.  It also sets this command to be executed by default run
function 'xref-ide-run'.
"
  (interactive "")
  (xref-delete-windows)
  (xref-run-function "run2" nil)
)


(defun xref-ide-run3 ()
"Run the  command specified  by '-set run3  <command>' option  in your
.xrefrc file.  It also sets this command to be executed by default run
function 'xref-ide-run'.
"
  (interactive "")
  (xref-delete-windows)
  (xref-run-function "run3" nil)
)


(defun xref-ide-run4 ()
"Run the  command specified  by '-set run4  <command>' option  in your
.xrefrc file.  It also sets this command to be executed by default run
function 'xref-ide-run'.
"
  (interactive "")
  (xref-delete-windows)
  (xref-run-function "run4" nil)
)


(defun xref-ide-run5 ()
"Run the  command specified  by '-set run5  <command>' option  in your
.xrefrc file.  It also sets this command to be executed by default run
function 'xref-ide-run'.
"
  (interactive "")
  (xref-delete-windows)
  (xref-run-function "run5" nil)
)

(defun xref-ide-compile-run () 
"Invoke (last)  compile command followed  by (last) run  command.  

The   run   command   is   invoked   only  in   case   of   successful
compilation. Also run is not invoked if the lastly invoked run command
was an  empty string.  See also  `xref-ide-compile' and `xref-ide-run'
functions.
"
  (interactive "")
  (let ((cbu) (cbuf) (cproc) (owin) (cwin) (rc) (opc) (sdir))
	(xref-delete-windows)
	(setq owin (get-buffer-window (current-buffer)))
	(setq sdir default-directory)
	(setq opc process-connection-type)
	(setq  process-connection-type nil)
	(xref-ide-compile)
	(setq cbu "compilation")
	(setq cbuf (format "*%s*" cbu))
	(setq cproc (get-buffer-process cbuf))
	(while (eq (process-status cproc) 'run)
	  (setq cwin (get-buffer-window cbuf))
	  (if cwin
		  (progn
			(select-window (get-buffer-window cbuf))
			(end-of-buffer)
			(setq default-directory sdir)
			(setq tab-width 8) 	;; as usual on terminals
			))
	  (sleep-for 0.1) ;; sleep at least some time
	  (discard-input) ;; discard input, so sit-for will refresh and wait
	  (sit-for 0.1)
	  )
	(sleep-for 0.1)
	(discard-input)
	(sit-for 0.1)
	(select-window owin)
	(setq process-connection-type opc)
	(if (eq (process-exit-status cproc) 0)
		(progn
		  (setq rc (xref-get-env xref-ide-last-run-command))
		  (if (and rc (not (equal rc "")))
			  (progn
				(xref-run-function xref-ide-last-run-command t)
				(setq xref-last-referencing-action 'compilation)
		  ))))
	(bury-buffer (get-buffer cbuf))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;    Projects    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-split-path-on-path-list (pname cut-slash)
  (let ((lplist) (i) (path) (len) (lchar))
	;; cut project path
	(setq lplist nil)
	(setq i (length pname))
	(while (> i 0) 
	  (setq i (- i 1))
	  (if (eq (elt pname i) xref-path-separator)
		  (progn
			(setq path (substring pname (+ i 1)))
			(if cut-slash (setq path (remove-pending-slash path)))
			(setq lplist (append lplist (cons path nil)))
			(setq pname (substring pname 0 i))
			)))
	(setq lplist (append lplist (cons pname nil)))
	lplist
))

	
(defun xref-get-project-list ()
  (let ((new-name) (loop) (mbeg) (mend) (pname "") (project-list) (i) (len) (lplist))
	(save-excursion
	  (set-buffer (get-buffer-create " xref-project-list"))
	  (xref-erase-buffer)
	  (insert-file xref-options-file)
	  (goto-char (point-min))
	  (setq project-list nil)
	  (setq loop t)
	  (while loop
		(setq loop (search-forward-regexp "\\[\\([^\]]*\\)\\]" 
										  (buffer-size) 1))
		(if loop
			(progn
			  (setq mbeg (match-beginning  1))
			  (setq mend (match-end  1))
			  (setq pname (buffer-substring mbeg mend))
			  (setq lplist (xref-split-path-on-path-list pname nil))
			  (setq project-list (append lplist project-list))
			  )))
	  (kill-buffer nil)
	  )
	project-list
))

(defun xref-softly-preset-project (pname)
  ;; do it like this just for now, then we will see
  ;; dummy action, just for soft preactivation of project
  (xref-call-process 'sync 1 nil 0 0 "" 
					 (format "-olcxgetsymboltype -noerrors -p \"%s\"" pname)
					 "-olcxrefsuffix=0:" "")
)

(defun xref-prj-list-get-prj-on-line ()
  (let ((res) (ppp) (bl) (el))
	(setq ppp (point))
	(beginning-of-line)
	(setq bl (point))
	(end-of-line)
	(setq el (point))
	(setq res (buffer-substring (+ bl 2) el))
	(goto-char ppp)
	res
))

(defun interactive-xref-project-select (&optional argp)
  "Goto to the reference corresponding to this line."
  (interactive "P")
  (let ((bl) (el) (prj))
	(setq prj (xref-prj-list-get-prj-on-line))
	(if (string-equal prj xref-directory-dep-prj-name)
		(progn
		  (setq xref-current-project nil)
		  ;; and reseting of softly selected project
		  (xref-softly-preset-project "")
		  )
	  (setq xref-current-project prj)
	  )
	(xref-delete-window  " *xref-project-list*")
	(message "Project '%s' is active." prj)
  )
  t
)

(defun interactive-xref-m-project-select (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-project-select)
  (other-window -1)
  t
)

(defun interactive-project-escape (&optional argp)
  "Escape from the project selection window."
  (interactive "P")
  (xref-delete-window  " *xref-project-list*")
)

(defvar xref-project-list-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\e\e\e" 'interactive-project-escape)
    (define-key map "\C-m" 'interactive-xref-project-select)
    (define-key map " " 'interactive-xref-project-select)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-m-project-select)
		  (define-key map 'button2 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-m-project-select)
	)
    map)
  "Keymap for `xref-project-list-mode'."
)
(xref-add-bindings-to-keymap xref-project-list-mode-map)


(defun xref-project-set-active-new ()
"Set  active  project.   

This function  is meningfull  only if your  '.xrefrc' file  contains a
section defining options for your  project. After setting a project to
be active  all Xrefactory functions will proceed  according to options
corresponding to this project name.
"
  (interactive "")
  (let ((prj) (cpl))
	(setq cpl (append 
			   (cons (cons xref-directory-dep-prj-name nil) nil)
			   (xref-get-project-list-completions) 
			   ))
	(setq prj (completing-read "Activate project: " cpl))
	(if prj
		(progn
		  (if (string-equal prj xref-directory-dep-prj-name)
			  (setq xref-current-project nil)
			(setq xref-current-project prj)
			)))
	(xref-project-active)
))

(defun xref-display-project-list (last-project local-keymap)
  (let ((xref-project-list))
	(xref-delete-windows)
	(setq xref-project-list (xref-get-project-list))
	(get-buffer-create " *xref-project-list*")
	(set-buffer " *xref-project-list*")
	(xref-erase-buffer)
	(insert "  ")
	(insert last-project)
	(put-text-property 3 (point) 'mouse-face 'highlight)
	(newline)
	(while xref-project-list
	  (goto-char (point-min))
	  (if (string-equal xref-current-project (car xref-project-list))
		  (insert "> ")
		(insert "  ")
		)
	  (insert (car xref-project-list))
	  (put-text-property 3 (point) 'mouse-face 'highlight)
	  (newline)
	  (setq xref-project-list (cdr xref-project-list))
	  )
	(xref-display-buffer " *xref-project-list*")
	(use-local-map local-keymap)
	(message "Use button2, <space> or <CR> to select the project")
))

(defun xref-project-set-active ()
"Set  active  project.   

This function  is meningfull  only if your  '.xrefrc' file  contains a
section defining options for your  project. After setting a project to
be active  all Xrefactory functions will proceed  according to options
corresponding to this project name.
"
  (interactive "")
  (xref-display-project-list xref-directory-dep-prj-name 
							 xref-project-list-mode-map)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun interactive-xref-project-delete (&optional argp)
  (interactive "P")
  (let ((prj) (eprj) (ppp) (eppp))
	(setq prj (xref-prj-list-get-prj-on-line))
	(if (string-equal prj xref-abandon-deletion)
		(progn
		  (xref-delete-window  " *xref-project-list*")
		  (message "Deletion canceled.")
		  )
	  (next-line 1)
	  (setq eprj (xref-prj-list-get-prj-on-line))
	  (xref-delete-window  " *xref-project-list*")
	  (find-file xref-options-file)
	  (beginning-of-buffer)
	  (setq ppp (search-forward (format "[%s]" prj) nil t))
	  (if (not ppp)
		  (error "Project section not found, probably is sharing options, delete it manually.")
		(setq ppp (- ppp (length (format "[%s]" prj))))
		)
	  (if (equal eprj xref-abandon-deletion)
		  (setq eppp (point-max))
		(setq eppp (search-forward (format "[%s]" eprj) nil t))
		(if (not eppp)
			(setq eppp (search-forward (format "[%s:" eprj) nil t))
		  )
		(if (not eppp)
			(error "Can't find end of project section, internal error, sorry.")
		  (setq eppp (- eppp (length (format "[%s]" eprj))))
		  ))
	  (if (> ppp eppp)
		  (error "[ppp>eppp] internal check failed, sorry")
		(goto-char ppp)
		(if (yes-or-no-p (format "Really delete %s" prj))
			(progn
			  (delete-char (- eppp ppp))
			  (save-buffer)
			  (kill-buffer (current-buffer))
			  (message "Project %s deleted." prj)
			  )
		  (message "No deletion.")
		  )
		))
  t
))

(defun interactive-xref-m-project-delete (event)
  (interactive "e")
  (mouse-set-point event)
  (interactive-xref-project-delete)
  (other-window -1)
  t
)

(defvar xref-project-delete-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\e\e\e" 'interactive-project-escape)
    (define-key map "\C-m" 'interactive-xref-project-delete)
    (define-key map " " 'interactive-xref-project-delete)
	(if (string-match "XEmacs" emacs-version)
		(progn
		  ;; XEmacs
		  (define-key map 'button2up 'interactive-xref-m-project-delete)
		  (define-key map 'button2 'undefined)
		  )
	  ;; Emacs
	  (define-key map [mouse-2] 'interactive-xref-m-project-delete)
	)
    map)
  "Keymap for `xref-project-delete-mode'."
)
(xref-add-bindings-to-keymap xref-project-delete-mode-map)


(defun xref-project-delete ()
"Delete a project.   

This function asks you to select  the project you wish to delete. Then
the part of .xrefrc file describing this project will be deleted.
"
  (interactive "")
  (xref-display-project-list xref-abandon-deletion
							 xref-project-delete-mode-map)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun remove-pending-slash (pfiles)
  (let ((dlen))
	(setq dlen (- (length pfiles) 1))
	(if (> dlen 0)
		(progn
		  (if (or (eq (elt pfiles dlen) ?/) (eq (elt pfiles dlen) ?\\))
			  (setq pfiles (substring pfiles 0 dlen))
			)))
	pfiles
))

(defun xref-char-replace (ss schars dchar)
  (let ((res) (i) (j) (len) (slen) (cc) (sc) (loop))
	(setq res "")
	(setq len (length ss))
	(setq slen (length schars))
	(setq i 0)
	(while (< i len) 
	  (setq cc (elt ss i))
	  (setq j 0)
	  (setq loop t)
	  (while (and (< j slen) loop)
		(setq sc (elt schars j))
		(if (eq cc sc)
			(setq loop nil)
		  )
		(setq j (+ j 1))
		)
	  (if loop
		  (setq res (format "%s%c" res cc))
		(setq res (format "%s%c" res dchar))
		)
	  (setq i (+ i 1))
	  )
	res
))

(defun backslashify-name (fname) 
  (let ((spos) (len) (res))
	(setq res fname)
	(if (or (eq xref-platform 'unix) (eq (length fname) 0))
		(setq res fname)
	  (setq len (length fname))
	  (setq spos 0)
	  (while (< spos len) 
		(if (or (eq (elt fname spos) ?/) (eq (elt fname spos) ?\\))
			(store-substring fname spos xref-slash)
		  )
		(setq spos (+ spos 1))
		)
	  )
	res
))

(defun xref-optionify-string (ss qt)
  (let ((res))
	(if (string-match " " ss)
		(setq res (format "%s%s%s" qt ss qt))
	  (setq res ss)
	  )
	res
))

(defun xref-append-new-project-section (pname planguage pcomments 
											  pfiles mclass classpath 
											  sourcepath 
											  javadocpath classdir
											  apfiles javahome ifiles
											  rest refs htmlopt exactp)
  (let ((comment) (dlen) (cldopt) (tdq))
	(setq comment (or (equal pcomments "y") (equal pcomments "Y")))
	(goto-char (point-max))
	(insert (concat "\n\n[" pname "]\n"))
	(if (or (equal planguage "j") (equal planguage "J"))
		(insert "  -javafilesonly\t// pure Java project\n")
	  )
	(if comment
		(insert "  //  input files and directories (processed recursively)\n")
	  )
	(insert (concat "  " apfiles))
	(newline)
	(setq pfiles (remove-pending-slash pfiles))
	(if comment
		(insert "  //  directory where Tag files are stored\n")
	  )
	(insert (format "  -refs %s\n" (xref-optionify-string refs "\"")))
	(if (equal htmlopt "")
		(progn
		  (insert "  //  number of Tag files\n")
		  (if (or (equal exactp "y") (equal exactp "Y"))
			  (insert "  -refnum=100\n")
			(insert "  -refnum=10\n")
			)
		  )
	  (insert "  //  split Tag files using first letter\n")
	  (insert "  -refalphahash\n")
	  )

	(if (or (equal planguage "c") (equal planguage "C")
			(equal planguage "b") (equal planguage "B"))
		(progn
		  (if (not (equal ifiles ""))
			  (insert (format "%s\n" ifiles))
			)
		  ))
	(if (or (equal planguage "c") (equal planguage "C"))
		(progn
		  (if (or (equal exactp "y") (equal exactp "Y"))
			  (insert "  // resolve symbols using definition place\n  -exactpositionresolve\n")
			)
		  (if comment (insert "  //  setting for (X)Emacs compile and run\n"))
		  (insert "  -set compilefile \"cc %s\"\n")
		  (insert "  -set compiledir \"cc *.c\"\n")
		  (insert (format "  -set compileproject \"\n\tcd %s\n\tmake\n\t\"\n" pname))
		  (if (eq xref-platform 'windows)
			  (insert "  -set run1 \"a.exe\"\n")
			(insert "  -set run1 \"a.out\"\n")
			)
		  (insert "  -set run5 \"\"  // an empty run; C-F8 will only compile\n")
		  (insert "  //  set default to run1\n")
		  (insert "  -set run ${run1}\n")
		))
	(if (or (equal planguage "j") (equal planguage "J")
			(equal planguage "b") (equal planguage "B"))
		(progn
		  (insert (format "  -set cp %s\n" (xref-optionify-string classpath "\"")))
		  (insert (format "  -set qcp \"${dq}${cp}${dq}\" \t// quoted classpath (if contains spaces)\n"))
		  (if (not (equal javahome ""))
			  (progn
				(insert (format "  -set jhome \"%s\"\n" javahome))
				(insert (format "  -set jbin \"${jhome}%s\"\n" (backslashify-name "bin/")))
				(insert (format "  -jdkclasspath \"${jhome}%s\"" 
									   (backslashify-name "jre/lib/rt.jar")))
				(if comment
					(insert "\t// java run time for xrefactory\n")
				  )
			))
		  (insert (format "  -classpath ${cp}"))
		  (if comment
			  (insert "\t// classpath for xrefactory\n")
			)
		  (insert (format "  -sourcepath %s" (xref-optionify-string sourcepath "\"")))
		  (if comment
			  (insert "\t// sourcepath for xrefactory\n")
			)
		  (if javadocpath
			  (progn
				(insert (format "  -javadocpath %s\n" (xref-optionify-string javadocpath "\"")))
				))

		  (if (or (equal planguage "b") (equal planguage "B"))
			  (progn
				(insert "  //  include directories for JNI\n")
				(insert "  -I ${jhome}include -I ${jhome}include/linux -I ${jhome}include/win32\n")
				))

		  (if comment
			  (insert "  //  setting for (X)Emacs compile and run\n")
			)
		  (if classdir
			  (progn
				(setq cldopt (format "-d %s" (xref-optionify-string classdir "${dq}")))
				)
			(setq cldopt "")
			)
		  (if (string-match " " javahome)
			  (progn
				(setq tdq "${dq}")
				)
			(setq tdq "")
			)
		  (insert (format "  -set compilefile \"%s${jbin}javac%s -classpath ${qcp} %s %%s\"\n" tdq tdq cldopt))
		  (insert (format "  -set compiledir \"%s${jbin}javac%s -classpath ${qcp} %s *.java\"\n" tdq tdq cldopt))
		  (insert (format "  -set compileproject \"\n\tcd %s\n\tant\n\t\"\n" pname))
		  (insert (format "  -set runthis \"%s${jbin}java%s -classpath ${qcp} %%s\"\n" tdq tdq))
		  (if (not (equal mclass ""))
			  (insert (format "  -set run1 \"%s${jbin}java%s -classpath ${qcp} %s\"\n" tdq tdq mclass))
			)
		  (insert "  -set run5 \"\"  // an empty run; C-F8 will only compile\n")
		  (if (not (equal mclass ""))
			  (progn
				(if comment (insert "  //  set default to run1\n"))
				(insert "  -set run ${run1}\n")
				)
			(if comment (insert "  //  set default to runthis\n"))
			(insert "  -set run ${runthis}\n")
		  )))
	(if (not (equal htmlopt ""))
		(insert (format "%s\n" htmlopt))
	  )
	(if (not (equal rest ""))
		(insert (format "%s\n" rest))
	  )
))


(defun xref-java-path-completionfun (cstr filter type)
  (let ((res) (cc) (fname) (dir) (sep) (str) (prefix))
	(setq str cstr)
	(setq sep (string-match (format "\\%c" xref-path-separator) str))
	(setq prefix "")
	(while sep
	  (progn 
		(setq prefix (concat prefix (substring str  0 (+ sep 1))))
		(setq str (substring str (+ sep 1) nil))
		(setq sep (string-match (format "\\%c" xref-path-separator) str))
	  ))
;;	(setq dir (concat (xref-file-directory-name str) "/"))
	(setq dir (file-name-directory str))
	(setq fname (file-name-nondirectory str))
	(if (eq type t)
		(setq res (file-name-all-completions fname dir))
	  (setq res (file-name-completion fname dir))
	  (if (eq xref-platform 'windows)
		  (setq res (backslashify-name res))
		)
	  )
	(if (stringp res)
		(setq res (concat cstr (substring res (length fname))))
	  )
	res
))

(defun xref-read-jpath-from-minibuffer (prompt default)
  (let ((res))
	(setq res (completing-read prompt 'xref-java-path-completionfun nil nil default))
	res
))

(defun xref-read-jpath-from-minibuffer-xx (prompt)
  (let ((res) (ini))
	(if (eq xref-platform 'windows)
		(setq ini "c:\\")
	  (setq ini "/")
	  )
	(setq res (completing-read prompt 'xref-java-path-completionfun nil nil ini))
	res
))

(defvar xref-foo-macros-counter 1)
(defun xref-collect-macros-for-new-project (prefix mess1 mess2)
  (let ((ifloop) (rrr) (rest) (aaa) (deflt))
	(setq rest "")
	(setq ifloop t)
	(while ifloop
	  (if (eq xref-foo-macros-counter 1)
		  (setq deflt "FOO")
		(setq deflt (format "FOO%d" xref-foo-macros-counter))
		)
	  (setq xref-foo-macros-counter
			(+ xref-foo-macros-counter 1))
	  (setq rrr (read-from-minibuffer mess1 deflt))
	  (setq rest (format "%s\n  %s-D%s" rest prefix rrr))
	  (setq aaa (read-from-minibuffer 
				 (format "%s [yn]? " mess2) "n"))
	  (if (not (or (equal aaa "y") (equal aaa "Y"))) (setq ifloop nil))
	  )
	rest
))

(defun xref-remove-dangerous-fname-chars (ss)
  (let ((res) (i) (len))
	(setq res ss)
	(if (eq xref-platform 'windows)
		(progn
		  (setq i 0)
		  (setq len (length ss))
		  (setq res "")
		  (while (< i len)
			(if (and (> i 2) (eq (elt ss i) ?\:))
				(setq res res)
			  (setq res (concat res (char-to-string (elt ss i))))
			  )
			(setq i (+ i 1))
			)
		  )
	  )
	res
))

(defun xref-project-new ()
"Create new project in  the .xrefrc Xrefactory option file. 

This function will pose you  few questions about your project and then
will create  a project description in .xrefrc  configuration file.  It
puts there some common options  for further manual edit. This function
creates a simple  description which can cover most  usual projects. It
is  usefull  rather for  Xrefactory  beginners.   Advanced users  will
probably prefer to  create and edit their .xrefrc  file manually.  See
also  'xrefrc' and  'xref' manual  pages for  more info  on  format of
.xrefrc file and available options.

To  delete a  project,  just edit  .xrefrc  file and  delete the  part
belonging to this project.
"
  (interactive "")
  (let ((pname) (checked) (tname) (breakcheck) (planguage) (pcomments)
		(pact) (pedit) (sfiles) (pfiles) (mclass) (system-class-path) (refs)
		(classpath) (sourcepath) (classdir nil) (spcp) (ljd) (javadocpath) 
		(inidir) (mif) (miff) (mifloop) (apfiles) (crtag) (stat) (jdkcp nil) 
		(javahome) (ptdef) (if) (ifiles "") (ifloop) (iff) (ifmess) (rest "")
		(aaa) (rrr) (pasi) (pasn) (hom) (htmlopt) (exactp))
	(if (eq xref-platform 'windows)
		(setq inidir "c:\\")
	  (setq inidir "/")
	  )
	(setq hom (getenv "HOME"))
	(if hom (setq hom (remove-pending-slash (backslashify-name hom))))
	;; this is a hack, just to create initial file (if not yet done)
	(xref-call-non-interactive-process "dummy-no-project" "-help" 
									   'synchronized 'nowin "Creating .xrefrc")
	(xref-delete-windows)
	;;
	(setq pname (backslashify-name (xref-file-directory-name (buffer-file-name))))
	(setq sfiles pname)
	(setq pname (xref-read-jpath-from-minibuffer "Enter new project name: " pname))
	(find-file xref-options-file)
	(setq checked nil)
	(while (not checked)
	  (setq pname (remove-pending-slash pname))
	  (beginning-of-buffer)
	  (if (search-forward (concat "[" pname "]") nil t)
		  (setq pname (xref-read-jpath-from-minibuffer
					   "** Name already used, new name :" pname))
		(beginning-of-buffer)
		(if (or (search-forward (concat "[" pname "/") nil t)
				(search-forward (concat "[" pname "\\") nil t))
			(setq pname (xref-read-jpath-from-minibuffer
						 "** Name overlaps an existing project, enter new name :" pname))
		  (setq tname pname)
		  (setq breakcheck nil)
		  (while (and tname (not breakcheck))
			(setq tname (xref-file-directory-name tname))
			(beginning-of-buffer)
			(if (and tname 
					 (search-forward (concat "[" tname "]") nil t))
				(progn
				  (setq pname (xref-read-jpath-from-minibuffer
						 "** This name is yet covered, enter new name :" pname))
				  (setq breakcheck t)
				  )))
		  (if (not breakcheck)
			  (setq checked t)
			)
		)))
	;; The name is checked 
	(save-buffer) 
	(kill-buffer nil)
	(if (xref-buffer-has-one-of-suffixes (buffer-file-name) xref-java-allowed-suffixes)
		(setq ptdef "j")
	  (if (xref-buffer-has-one-of-suffixes (buffer-file-name) xref-c-allowed-suffixes)
		  (setq ptdef "c")
		(setq ptdef "b")
		))
	(setq planguage (read-from-minibuffer 
				 "This will be C or Java project (or Both) [cjb]? " ptdef))
;;   	(setq pcomments (read-from-minibuffer 
;;				 "Do you wish that project description contains commentaries [yn]? " "y"))
   	(setq pfiles (xref-read-jpath-from-minibuffer
				 "Directory containing project sources: " sfiles))
	(setq mifloop t)
	(setq apfiles (xref-optionify-string pfiles "\""))
	(while mifloop
	  (progn
		(setq mif (read-from-minibuffer 
				   "Add another source directory, file or jar [yn]? " "n"))
		(if (or (equal mif "y") (equal mif "Y"))
			(progn
			  (setq miff (xref-read-jpath-from-minibuffer
							"Additional source:  " sfiles))
			  (setq apfiles (format "%s\n  %s" apfiles (xref-optionify-string miff "\"")))
			  )
		  (setq mifloop nil)
		  )))
	(if hom
		(progn
		  (setq refs (format "%s%cXrefs%c%s" hom xref-slash xref-slash 
							 (xref-cut-string-prefix 
							  (xref-cut-string-prefix (backslashify-name pname) 
													  (format "%s%c" hom xref-slash)
													  (eq xref-platform 'windows))
							  (format "%c" xref-slash)
							  nil)))
		  )
	  (setq refs (format "%s%cXrefs" pfiles xref-slash))
	  )
	(setq refs (xref-remove-dangerous-fname-chars refs))
	(setq refs (xref-read-jpath-from-minibuffer
				"Directory to store Tag files:  " refs))
	(if (or (equal planguage "j") (equal planguage "J")
			(equal planguage "b") (equal planguage "B"))
		(progn
		  ;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Java project specifics
		  (setq javahome (xref-get-simple-answer-from-process "-olcxgetjavahome -noerrors" nil))
		  (setq javahome (xref-read-jpath-from-minibuffer
						  "Java home directory (containing jre, lib, bin ...): " javahome))
		  (if (not (equal javahome ""))
			  (setq javahome (backslashify-name (concat (remove-pending-slash javahome) "/")))
			)

		  (setq mclass (xref-get-current-class "-olcxcurrentclass -noerrors" 
											   nil))
		  (setq mclass (xref-read-jpath-from-minibuffer
						"Main class for 'run' commands? " mclass))
		  (setq classpath (getenv "CLASSPATH"))
		  (if classpath 
			  (progn
				(setq system-class-path (read-from-minibuffer 
						   "Your system CLASSPATH is set, use it for this project [yn]? " "n"))
				(if (or (equal system-class-path "n") (equal system-class-path "N"))
					(setq classpath (xref-read-jpath-from-minibuffer "Enter classpath for this project: " pfiles))
				  (setq classpath "${CLASSPATH}")
				  ))
			(setq classpath (xref-read-jpath-from-minibuffer "Enter classpath for this project: " pfiles))
		  )
		  (setq classpath (remove-pending-slash classpath))
		  (setq spcp (read-from-minibuffer 
					  "Are your source files stored in same directories as classes [yn]? " "y"))
		  (if (or (equal spcp "n") (equal spcp "N"))
			  (progn
				(setq sourcepath (remove-pending-slash (xref-read-jpath-from-minibuffer "Enter sourcepath for this project: " pfiles)))
				(setq classdir (remove-pending-slash (xref-read-jpath-from-minibuffer "Directory to generate .class files (-d javac option): " classpath)))
				)
			(setq sourcepath "${cp}")
			)
		  (setq ljd (read-from-minibuffer 
					  "Have you a local copy of Javadoc documentation on your computer [yn]? " "n"))
		  (if (or (equal ljd "y") (equal ljd "Y"))
			  (setq javadocpath (xref-read-jpath-from-minibuffer "Enter javadocpath: " (concat javahome (backslashify-name "docs/api"))))
			(setq javadocpath nil)
			)
		  )
	  ;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; C project specifics
	  (setq ifiles "")
	  (setq ifmess "Do you use some ")
	  (if (eq xref-platform 'windows)
		  (progn
			(setq iff (xref-read-jpath-from-minibuffer "Directory containing standard headers (stdio.h, stdlib.h, ...): " inidir))
			(setq ifiles (format "%s\n  -I %s" ifiles 
								 (xref-optionify-string (remove-pending-slash iff) "\"")))
			(setq ifmess "Add any ")
			))
	  (setq ifloop t)
	  (while ifloop
		(progn
		  (setq if (read-from-minibuffer 
					 (concat ifmess "nonstandard include directory (-I option) [yn]? ") "n"))
		  (if (or (equal if "y") (equal if "Y"))
			  (progn
				(setq iff (xref-read-jpath-from-minibuffer
						   "Additional include directory:  " inidir))
				(setq ifiles (format "%s\n  -I %s" ifiles 
									 (xref-optionify-string (remove-pending-slash iff) "\"")))
				)
			(setq ifloop nil)
			)
		  (setq ifmess "Add another ")
		  ))
	  (if (not (equal ifiles ""))
		  (setq ifiles (concat "  //  include directories" ifiles))
		)
	  ;;
	  (setq exactp (read-from-minibuffer 
				 "Is this a huge project with numerous name conflicts [yn]? " "n"))
	  ;;
	  (setq rest "")
	  (setq xref-foo-macros-counter 1)
	  (setq aaa (read-from-minibuffer 
				 "Do you compile sources with command line macro definitions (-D option) [yn]? " "n"))
	  (if (or (equal aaa "y") (equal aaa "Y"))
		  (progn
			(setq aaa (read-from-minibuffer 
					   "Do you compile sources several times with different macro settings [yn]? " "n"))
			(if (or (equal aaa "y") (equal aaa "Y"))
				(progn
				  (setq pasn (string-to-int 
							  (read-from-minibuffer 
							   "How many compilations with different initial macro settings: " "2")))
				  (setq aaa (read-from-minibuffer 
							 "Are there macros common for all compilations [yn]? " "n"))
				  (if (or (equal aaa "y") (equal aaa "Y"))
					  (progn
						(setq rest (concat rest 
										   (xref-collect-macros-for-new-project ""
											"A common macro to be defined in all compilations: "
											"Add another common macro definition")))
						))
				  (setq pasi 1)
				  (while (<= pasi pasn)
					(progn
					  (setq rest (format "%s\n  -pass%d" rest pasi))
					  (setq rest (concat rest 
										 (xref-collect-macros-for-new-project "  "
										  (format "A macro specific to compilation #%d: " pasi)
										  (format "Add another macro definition for compilation #%d"pasi))))
					  (setq pasi (+ pasi 1))
					  ))
				  )
			  ;; a single pass macros
			  (setq rest (concat rest 
								 (xref-collect-macros-for-new-project ""
								  "A macro defined during compilation: "
								  "Add another macro definition")))
			  )
			))
	  (if (not (equal rest ""))
		  (setq rest (concat "  //  pre-processor macros and passes specification" 
							 rest))
		)
	  )

	;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; HTML common questions
	(setq htmlopt "")
	(setq aaa (read-from-minibuffer 
			   "Will you generate HTML documentation for this project [yn]? " "n"))
	(if (or (equal aaa "y") (equal aaa "Y"))
		(progn	
		  (if hom
			  (setq rrr (format "%s%cHTML" hom xref-slash))
			(setq rrr (format "%sHTML" inidir))
			)
		  (setq rrr (xref-read-jpath-from-minibuffer
					 "Output directory to generate HTML documentation: " rrr))
		  (setq rrr (remove-pending-slash (backslashify-name rrr)))
		  (setq htmlopt "  //  HTML configuration")
		  (setq htmlopt (format "%s\n  %s" htmlopt 
								(xref-optionify-string (concat "-htmlroot=" rrr) "\"")))
		  (setq htmlopt (format "%s\n  -htmlgxlist -htmllxlist -htmldirectx -htmllinenums" htmlopt))
		  (setq htmlopt (format "%s\n  -htmltab=%d -htmllinenumcolor=000000" htmlopt tab-width))
		  (if (or (equal planguage "j") (equal planguage "J")
				  (equal planguage "b") (equal planguage "B"))
			  (progn
				(setq htmlopt (format "%s\n  -htmlgenjavadoclinks -htmlcutsuffix -htmllinenumlabel=line" htmlopt))
				))
		  ))

	;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; all questions done
	(find-file xref-options-file)
	(xref-append-new-project-section pname planguage "y" ;; pcomments 
									 pfiles mclass
									 classpath sourcepath javadocpath
									 classdir apfiles javahome ifiles 
									 rest refs htmlopt exactp)
	;;
	(save-buffer)
	(search-backward (concat "[" pname "]"))
	(recenter)

	(setq pedit "n")
;;	(setq pedit (read-from-minibuffer 
;;				 "Do you wish to adjust options manually [yn]? " "n"))
	(if (or (equal pedit "n") (equal pedit "N"))
		(progn
		  (kill-buffer nil)
		  (setq crtag "y")
;;		  (setq crtag (read-from-minibuffer 
;;					   "Create Tag file now [yn]? " "y"))
		  (if (or (equal crtag "y") (equal crtag "Y"))
			  (progn
				(setq stat (xref-call-non-interactive-process
							pname "-create" 'synchronized 'nowin 
							"Creating Tag File. "))
				(xref-project-exit-message-on-nonzero stat)

	  ))))
	(setq xref-current-project nil)
	(xref-softly-preset-project pname)
	(setq pact (read-from-minibuffer 
				"Make the new project active [yn]? " "n"))
	(if (or (equal pact "y") (equal pact "Y"))
		(progn
		  (setq xref-current-project pname)
		  (message "Project '%s' is now active." pname)
		  )
	  (setq xref-current-project nil)
	  (message "Setting project auto-detection.")
	  )
	(sit-for 1)  ;; to read project message
	(message "All Done. New project was created.")
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-nonint-get-active-project-or-error ()
  (let ((res))
	(xref-call-process 'sync 1 nil 0 0 "" "-olcxgetprojectname" 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\*) 
		(setq res xref-last-message)
	  (error "No active project")
	  )	
	res
))

(defun xref-project-active () 
"Show currently  active project name. 

This  function is usefull  mainly if  you are  not sure  which .xrefrc
section applies on the currently edited file.
"
  (interactive "")
  (if xref-current-project
	  (message "Active project (man): %s" xref-current-project)
	(xref-call-process 'sync 1 nil 0 0 "" "-olcxgetprojectname" 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\*) 
		(message "Active project (auto): %s" xref-last-message)
	  )
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-project-edit-options ()
"Edit manually .xrefrc file. 

This  function just  load  the .xrefrc  file  and goes  to the  active
project section.  You need to edit  this text file  manually. For more
info about .xrefrc  file format, read 'xrefrc' manual  page.  For more
info about available options, read 'xref' manual page.
"
  (interactive "")
  (let ((prjstr "") (r))
	;; get project name
	(xref-delete-windows)
	(xref-call-process 'sync 1 nil 0 0 "" "-olcxgetprojectname" 
					 "-olcxrefsuffix=0:" "")
	(if (eq xref-pilot-char ?\*) 
		(setq prjstr xref-last-message)
	  )
	(find-file xref-options-file)
	(beginning-of-buffer)
	(if (not (search-forward (concat "[" prjstr "]") nil t))
		(if (not (search-forward (concat "[" prjstr ":") nil t))
			(if (not (search-forward (concat ":" prjstr "]") nil t))
				(if (not (search-forward (concat ":" prjstr ":") nil t))
					(message "Options for %s found" prjstr)
					))))
	(beginning-of-line)
	(next-line 1)
))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; MOVINGS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-get-current-class (option abort-on-error)
  (let ((res))
	(setq xref-pilot-char ?\ )
	(xref-call-process 'sync 1 nil (point) 0 "" option
					   "" "-olcxrefsuffix=0:#")
	(if (eq xref-pilot-char ?*)
		(setq res xref-last-message)
	  (if abort-on-error
		  (error "Can't determine class on cursor position, sorry.")
		(setq res "")
	  ))
	res
))

(defun xref-get-simple-answer-from-process (option abort-on-error)
  (let ((res))
	(setq xref-pilot-char ?\ )
	(xref-call-process 'sync 0 nil (point) 0 "" option "" "")
	(if (eq xref-pilot-char ?*)
		(setq res xref-last-message)
	  (if abort-on-error
		  (error "Can't get answer on request '%s', sorry." option)
		(setq res "")
	  ))
	res
))

(defun xref-mark-move-target-place (poin)
"Set  the destination  place  for moving  refactorings. 

This function memorizes current cursor position and current class name
for future use by moving refactorings. The next invocation of a field,
method or class moving function will move to this memorized place.
"
  (interactive "d")
  (if (eq (point) (point-max)) 
	  (progn
		(newline)
		(backward-char 1)
		))
  (set-marker xref-move-target-marker (point))
  (setq xref-move-target-class (xref-get-current-class "-olcxtarget" t))
  (if (and 
	   (> (length xref-move-target-class) 0) 
	   (equal (elt xref-move-target-class 0) ?\ ))
	  (progn
		(message "Next moving refactoring will move to %s:%d (package%s)"
				 (xref-file-last-name (xref-move-target-file))
				 (xref-move-target-line) 
				 (xref-char-replace xref-move-target-class "/$" ?.))
		)
	(message "Next moving refactoring will move to %s:%d (class %s)"
			 (xref-file-last-name (xref-move-target-file))
			 (xref-move-target-line) 
			 (xref-char-replace xref-move-target-class "/$" ?.))
	)
)

(defun xref-move-target-is-class ()
  (let ((res))
	(setq res nil)
	(if (and 
		 (> (length xref-move-target-class) 0) 
		 (not (equal (elt xref-move-target-class 0) ?\ )))
		(setq res t)
	  )
	res
))

(defun xref-move-target-is-package ()
  (let ((res))
	(setq res nil)
	(if (and 
		 (> (length xref-move-target-class) 0) 
		 (equal (elt xref-move-target-class 0) ?\ ))
		(setq res t)
	  )
	res
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                      Encapsulate field (with new parsing)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-skip-blank (ends)
  (let ((rs) (loop) (c1) (c2) (c3))
	(setq loop t)
	(while loop
	  (progn
		(setq rs (search-forward-regexp "[^ \t\n]" ends 0))
		(if (not rs)
			(setq loop nil)
		  (backward-char 1)
		  (setq c1 (char-after (point)))
		  (if (< (+ (point) 1) ends)
			  (setq c2 (char-after (+ (point) 1)))
			(setq c2 ?\ )
			)
		  (if (< (+ (point) 2) ends)
			  (setq c3 (char-after (+ (point) 2)))
			(setq c3 ?\ )
			)
		  (cond
		   ((and (eq c1 ?/) (eq c2 ?*) (not (eq c3 ?&)))
			;; /* */ comment skip it
			(forward-char 2)
			(setq rs (search-forward "*/" ends 0))
			)
		   ((and (eq c1 ?/) (eq c2 ?/) (not (eq c3 ?&)))
			;; // comment
			(setq rs (search-forward "\n" ends 0))
			)
		   ((and (eq c1 ?/) (eq c2 ?*) (eq c3 ?&))
			(forward-char 3)
			)
		   ((and (eq c1 ?&) (eq c2 ?*) (eq c3 ?/))
			(forward-char 3)
			)
		   ((and (eq c1 ?/) (eq c2 ?/) (eq c3 ?&))
			(forward-char 3)
			)
		   (t
			(setq loop nil)
			)))
	  ))
))

(defun xref-next-lexem (ends)
  (let ((res) (b) (e) (ccc))
	(xref-skip-blank ends)
	(if (>= (point) ends)
		(setq res nil)
	  (setq b (point))
	  (setq ccc (char-after (point)))
	  (cond
	   ((or (and (>= ccc ?A) (<= ccc ?Z)) (and (>= ccc ?a) (<= ccc ?z))
			(eq ccc ?\_) (eq ccc ?$))
		;; identifier
		(while (or (and (>= ccc ?A) (<= ccc ?Z)) (and (>= ccc ?a) (<= ccc ?z))
				   (eq ccc ?\_) (eq ccc ?$) (and (>= ccc ?0) (<= ccc ?9)))
		  (progn
			(forward-char 1)
			(setq ccc (char-after (point)))
			))
		(setq res 'identifier)
		)
	   ((eq ccc ?\")
		(search-forward-regexp xref-string-passing-regexp)
		(message "checking '%c'" (char-after (- (point) 1)))
		(if (and (> (point) 1) (eq (char-after (- (point) 1)) ?\n))
			(progn
			  (xref-wait-until-problem-confirmed "String through end of line?")
			  ))
		(setq res 'literal)
		)
	   ((eq ccc ?\')
		(search-forward-regexp "\\([^\\]'\\|\n\\)")
		(setq res 'literal)
		)
	   (t
		;; other
		(forward-char 1)
		(setq res 'character)
		))
	  )
	res
))

(defun xref-pass-to-assignement-equality (ends)
  (let ((cl) (cc) (c) (res))
	(setq cl (xref-next-lexem ends))
	(if (not (equal cl 'character))
		(error "Not a character in affectation? Do not know to handle this, sorry.")
	  )
	(setq cc (char-after (- (point) 1)))
	(setq res 'assignement)
	(if (not (equal cc ?=))
		(error "Error while parsing an assignement reference, = expected.")
	  )
	res
))

(defun xref-max-expression-pos ()
  (let ((res))
	(setq res (+ (point) xref-expression-max-context))
	(if (> res (point-max))
		(setq res (point-max))
	  )
	res
))

(defun xref-min-name-pos (poin)
  (let ((res))
	(setq res (- poin xref-qualified-name-max-context))
	(if (< res (point-min))
		(setq res (point-min))
	  )
	res
))

(defun xref-skip-expression (ends)
  (let ((cc) (level) (ll))
	(backward-char 1)
	(setq level 0)
	(while (>= level 0)
	  (setq ll (xref-next-lexem ends))
	  (if (>= (point) ends)
		  (error "Expression too long, sorry.")
		)
	  (if (equal ll 'character)
		  (progn
			(setq cc (char-after (- (point) 1)))
			(if (> (point) ends)
				(error "Expression too long, Sorry.")
			  )
			(if (or (eq cc ?\() (eq cc ?\{) (eq cc ?\[))
				(setq level (+ level 1))
			  (if (or (eq cc ?\)) (eq cc ?\}) (eq cc ?\]))
				  (setq level (- level 1))
				(if (and (equal level 0) (or (eq cc ?\;) (eq cc ?\,)))
					(setq level -1)
				))))))
))

(defun xref-encapsulate-l-value (name)
"This should encapsulate  an l-value reference into a  call of setting
method.  Unfortunately due to  the complexity of possible expressions,
this will not work for base types and references as i++, --i, ...
"
  (let ((pos) (eb) (s-point) (mpos) (op) (uprefix) (up0) (up1) (up2) (up3)
		(lastdot) (prefix) (suffix) (lvar) (aaa) (cl))
	(setq pos (point))
	(setq mpos (xref-max-expression-pos))
	(setq s-point (xref-min-name-pos (point)))
	(setq cl (xref-next-lexem mpos))
	(if (not (equal cl 'identifier))
		(error "Not an identifier on l-val reference? Internal error, sorry.")
	  )
   	(setq op (xref-pass-to-assignement-equality mpos))
	(setq aaa (char-after (- (point) 2)))
	(xref-skip-blank mpos)
	(delete-char (- pos (point)))
	(insert (format "set%s(" name))
	(forward-char 1)
	(xref-skip-expression mpos)
	(backward-char 1)
	(insert (format ")"))
  nil
))

(defun xref-encapsulate-r-value (name)
  (insert (format "get%s()" name))
  (xref-delete-pending-ident)
  nil
)

(defun xref-encapsulate-reference (name dummy)
  (if (not (xref-last-move-was-to-definition-or-declaration))
	  (progn
		(if (or (eq xref-move-usage ?,) (eq xref-move-usage ?.))
			(progn
			  ;; l-val or addr ref
			  (if (eq xref-move-usage ?.)
				  (error "** An address reference, internal error!")
				)
			  (xref-encapsulate-l-value name)
			  )
		  (xref-encapsulate-r-value name)
		  )))
  nil
)

(defun xref-encapsulate-reference-check (name dummy)
  (if (eq xref-move-usage ?.)
	  (progn
		(error "** Sorry, an address usage. Turn it into assignment (i.e. %s = %s op ...) first." name name)
		))
  nil
)

(defun xref-get-matches (decl modifiers-list)
  (let ((ll) (res))
	(setq res "")
	(setq ll modifiers-list)
	(while ll
	  (if (string-match (car ll) decl)
		  (setq res (concat res (car ll) " "))
		)
	  (setq ll (cdr ll))
	  )
	res
))

(defun xref-get-declaration-modifiers (decl)
  (let ((res))
	(setq res (xref-get-matches decl '("public" "protected" "private" "static" "final" "transient" "volatile")))
	res
))

(defun xref-get-indentation ()
  (let ((res) (pos) (bl))
	(setq res "")
	(setq pos (point))
	(beginning-of-line)
	(setq bl (point))
	(search-forward-regexp "[^ \t]")
	(if (> (- (point) 1) bl)
		(setq res (buffer-substring bl (- (point) 1)))
	  )
	(goto-char pos)
	res
))

(defun xref-remove-multiple-spaces (sss)
  (let ((res) (loop) (mi))
	(setq res sss)
	(setq loop t)
	(while loop
	  (setq mi (string-match "  " res))
	  (if mi
		  (setq res (concat (substring res 0 mi) (substring res (+ mi 1))))
		(setq loop nil)
		))
	;; remove leading spaces too
	(while (or (equal (substring res 0 1) " ") 
			   (equal (substring res 0 1) "\t"))
	  (setq res (substring res 1))
	  )
	res
))

(defun xref-encapsulate-field-finish-safety-check3 ()
  (let ((scopt) (file) (line))
	(xref-switch-to-marker xref-encapsulate-getter-marker)
	(setq file (buffer-file-name nil))
	(setq line (count-lines 1 (point)))
	(setq scopt 
		  (concat
		   "-olcxencapsulatesc2"
		   (format " -olcxsafetycheckmovedblock=%d:%d:%d" line 10000000 (+ line 3))
		   (format " -olcxsafetycheckmovedfile \"%s\" \"%s\" " file file)))
	(push-references 'sync (buffer-file-name) (point) t nil scopt 0 '(listrefs-continuation))

))

(defun xref-encapsulate-field-finish-safety-check2 ()
  (xref-switch-to-marker xref-encapsulate-setter-marker)
  (push-references 'sync (buffer-file-name) (point) t nil "-olcxencapsulatesc1" 0 
				   '(xref-encapsulate-field-finish-safety-check3))
)

(defun xref-encapsulate-field-finish (name uname)
  (let ((prefix) (dpos) (ftype) (dprefix) (class-name) (indent) (body)
		(boundpos))
	;; should be via markers?
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
	(xref-moveref nil)
	(setq dpos (point))
	(setq ftype (xref-get-type-of-symbol (buffer-file-name)))
	(beginning-of-line) 
	(setq prefix (buffer-substring (point) dpos))
	(goto-char dpos)
	(setq indent (xref-get-indentation))
	(xref-non-int-change-field-access-attribute "private" nil)
	(end-of-line)
	(insert "\n")
	(setq dprefix (xref-get-declaration-modifiers prefix))
	(setq body (concat 
				indent
				(xref-remove-multiple-spaces 
				(format "%s %s get%s() {return %s;}\n" 
						dprefix ftype uname name))))
	(if (string-match "static" prefix)
		(progn
		  (setq class-name (xref-get-current-class "-olcxcurrentclass" t))
		  (setq body (concat 
					  body indent
					  (xref-remove-multiple-spaces 
					   (format 
						"%s %s set%s(%s %s) {%s.%s=%s; return %s;}\n" 
						dprefix ftype uname ftype name 
						class-name name name name)))))
	  (setq body (concat
				  body indent
				  (xref-remove-multiple-spaces
				   (format 
					"%s %s set%s(%s %s) { this.%s=%s; return %s;}\n" 
					dprefix ftype uname ftype name name name name))))
	  )
	(insert body)
	;; (insert "\n")
	(setq boundpos (point))
	(goto-char dpos)
	(search-forward "set" boundpos) 
	(set-marker xref-encapsulate-setter-marker (point))
	(xref-nonint-reduce-useless-long-names-of-method (point) nil)
	(goto-char dpos)
	(search-forward "get" boundpos) 
	(set-marker xref-encapsulate-getter-marker (point))
	(xref-nonint-reduce-useless-long-names-of-method (point) nil)

	(if (xref-safety-check-prompt)
		(progn
		  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxsafetycheckinit" 
							 "-olcxrefsuffix=0:#" "")
		  (xref-non-interactive-update-references 
		   xref-fast-updates-inside-refactorings)
		  ;; it should be placed on get, but anyway
		  (xref-switch-to-marker xref-encapsulate-getter-marker)
		  (push-references 'sync (buffer-file-name) (point) t nil "-olcxencapsulatesc1" 0 
						   '(xref-encapsulate-field-finish-safety-check2))
		  )
	  (xref-popCxrefs (point))
	  (message "All done.")
	  )
))

(defun xref-self-encapsulate-field ()
"Self encapsulates field  by  creating  a  pair of  get/set methods.

This function also makes the field private and replaces all references
of  the   field  by  appropriate  method  calls.    At  difference  to
'encapsulate field'  it replaces all field  references including those
inside definition  class.  At invocation  of this function  the cursor
has to be positioned on the  field defintion (on its name).  The field
has to be declared on a separate line, i.e.  its definition should not
exceed one line and no other symbol can be declared on the same line.

Also  it is  possible that  due to  differences between  resolution of
method and field names there will be sometimes another setting/getting
method invoked  due to  virtual method calling  mechanism. 
"
  (interactive "")
  (let ((name) (uname))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
		"You should edit a Java file when using this function.")

	(xref-trivial-pre-checks "-tpchmovefield")
	(setq name (xref-get-identifier-on-point))
	(xref-delete-windows)
	(xref-refactoring-init-actions
	 (format "self encapsulate %s " name))
	(setq uname (format "%s%s" 
						(upcase (substring name 0 1)) 
						(substring name 1)))
	(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxrename"
								 xref-detailed-refactoring-confirmations
								 (list 'list ''xref-encapsulate-reference-check name nil)
								 ''none-pre-fun
								 ''xref-encapsulate-reference
								 ''standard-post-fun 
								 uname
								 nil
								 nil
								 (list 'xref-encapsulate-field-finish name uname)
								 )
))

(defun xref-encapsulate-field ()
"Encapsulates field  by  creating  a  pair of  get/set methods.

This  function   also  makes  the  field  private   and  replaces  all
non-private references  of the field  by appropriate method  calls. At
difference  to 'self  encapsulate field'  this function  replaces only
those field references  which have no longer access  to the field.  At
invocation of  this function  the cursor has  to be positioned  on the
field defintion  (on its  name).  The  field has to  be declared  on a
separate line, i.e.  its definition  should not exceed one line and no
other symbol can be declared on the same line.

Also  it is  possible that  due to  differences between  resolution of
method and field names there will be sometimes another setting/getting
method invoked  due to  virtual method calling  mechanism. 
"
  (interactive "")
  (let ((name) (uname))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
		"You should edit a Java file when using this function.")

	(xref-trivial-pre-checks "-tpchmovefield")
	(setq name (xref-get-identifier-on-point))
	(xref-delete-windows)
	(xref-refactoring-init-actions
	 (format "encapsulate %s " name))
	(setq uname (format "%s%s" 
						(upcase (substring name 0 1)) 
						(substring name 1)))
	(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxencapsulate"
								 xref-detailed-refactoring-confirmations
								 (list 'list ''xref-encapsulate-reference-check name nil)
								 ''none-pre-fun
								 ''xref-encapsulate-reference
								 ''standard-post-fun 
								 uname
								 nil
								 nil
								 (list 'xref-encapsulate-field-finish name uname)
								 )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;; FQT parsing

(defun ident-search-state (state nchar)
  (let ((res))
	(if (eq nchar 'letter) (setq res 'in-ident)
	  (if (eq nchar 'blank)
		  (progn
			(if (eq state 'base) (setq res 'base))
			(if (eq state 'in-ident) (setq res 'ident-blank))
			(if (eq state 'ident-blank) (setq res 'ident-blank))
			(if (eq state 'ident-dot) (setq res 'ident-dot))
			)
		(if (eq nchar 'dot)
			(progn
			  (if (eq state 'base) (setq res 'base))
			  (if (eq state 'in-ident) (setq res 'ident-dot))
			  (if (eq state 'ident-blank) (setq res 'ident-dot))
			  (if (eq state 'ident-dot) (setq res 'base))
			  )
		  (error "Internal error, unexpect char for state change")
		  )))
	res
))
  

(defun xref-detect-qualified-name (s-point o-point parse-primary)
  (let ((res) (id-start -1) (last-expr-fin -1) (last-sub-id-start -1)
		(last-dot -1) (state 'base) (stack nil) 
		(top) (ccc) (last-id))
	(goto-char s-point)
	;; following should avoid cases when I am starting inside a string
	;; (beginning-of-line)     
	(while (<= (point) o-point)
	  (setq ccc (char-after (point)))
	  (forward-char 1) 
	  (cond
	   ;; character constant
	   ((eq ccc ?\')
		(setq state 'base)
		(backward-char)
		(search-forward-regexp "\\([^\\]'\\|\n\\)"))
	   ;; string
	   ((eq ccc ?\")
		(setq state 'base)
		(backward-char)
		(search-forward-regexp xref-string-passing-regexp))
	   ((and (eq ccc ?/) (eq (char-after (point)) ?*)
			 (eq (char-after (+ (point) 1)) ?\&))
		;; ignore /*& programm-comment
		(forward-char 2)
		)
	   ((and (eq ccc ?\&) (eq (char-after (point)) ?*)
			 (eq (char-after (+ (point) 1)) ?/))
		;; ignore &*/ programm-comment
		(forward-char 2)
		)
	   ((and (eq ccc ?/) (eq (char-after (point)) ?/)
			 (eq (char-after (+ (point) 1)) ?\&))
		;; ignore //& programm-comment
		(forward-char 2)
		)
	   ((and (eq ccc ?/) (eq (char-after (point)) ?*) 
			 (not (eq (char-after (+ (point) 1)) ?\&)))
		(setq state (ident-search-state state 'blank))
		;; /* */ coment
		(search-forward "*/"))
	   ;; // comment
	   ((and (eq ccc ?/) (eq (char-after (point)) ?/)
			 (not (eq (char-after (+ (point) 1)) ?\&)))
		(setq state (ident-search-state state 'blank))
		(search-forward "\n"))
	   ;; 
	   ((eq ccc ?\( )
		(if (or (eq state 'in-ident) (eq state 'ident-blank))
			(progn
			  ;; method invocation
			  (setq top (cons state (cons id-start nil)))
			  (setq stack (cons top stack))
			  (setq state 'base)
			  (setq last-expr-fin (point))
			  (setq id-start -1)
			  (setq last-sub-id-start (point))
			  )
		  ;; parenthesed expression (try to pass expression as a letter, space?
		  ;; orig not work for (Cast)toto, but work for (Cast) toto
		  ;; (setq top (cons 'in-ident (cons (- (point) 1) nil)))
		  (setq top (cons 'ident-blank (cons (- (point) 1) nil)))
		  (setq stack (cons top stack))
		  (setq state 'base)
		  (setq last-expr-fin (point))
		  (setq id-start -1)
		  (setq last-sub-id-start (point))
		  )		  
		)
	   ((or (eq ccc ?\)) (eq ccc ?\]))
		(if stack
			(progn
			  (setq top (car stack))
			  (setq state (car top))
			  (setq id-start (car (cdr top)))
			  (setq stack (cdr stack))
			  )
		  (setq state 'base)
		  (setq id-start -1)
		  )
		(setq last-expr-fin (point))
		(setq last-sub-id-start (point))
		)
	   ;; 
	   ((eq ccc ?\[ )
		;; array access expression
		;; do the same as method invocation
		(setq top (cons state (cons id-start nil)))
		(setq stack (cons top stack))
		(setq state 'base)
		(setq last-expr-fin (point))
		(setq id-start -1)
		(setq last-sub-id-start (point))
		)
	   ;;
	   ((or (eq ccc ?\ ) (eq ccc ?\t) (eq ccc ?\r) (eq ccc ?\n))
		;; blank, ignore them
		(setq state (ident-search-state state 'blank))
		)
	   ((eq ccc ?\.)
		;; dot ignored
		(setq last-dot (point))
		(setq state (ident-search-state state 'dot))
		)
	   ((or (and (>= ccc ?A) (<= ccc ?Z)) (and (>= ccc ?a) (<= ccc ?z))
			(eq ccc ?\_) (eq ccc ?$)
			(and (>= ccc ?0) (<= ccc ?9) (eq state 'in-ident))
			)
		;; identifier
		(if (eq state 'base)
			(progn 
			  (setq id-start (- (point) 1))
			  (setq last-expr-fin id-start)
			  (setq last-sub-id-start id-start)
			  )
		  (if (eq state 'ident-blank)
			  (progn 
				;; check for 'new xxx()' which has larger precedence that dot
				;; this has to be handled as special case, because of
				;; two different situations reduce red long name -> 'new sk.xref.Toto()'
				;; and turn dyn. to st. 'new Toto().method()'
				;; parameter 'parse-primary' is used to distinguish those cases
				(setq last-id (xref-get-identifier-on last-sub-id-start))
				(if (and parse-primary (equal last-id "new"))
					(progn
					  (setq last-expr-fin (- (point) 1))
					  (setq last-sub-id-start (- (point) 1))
					  )
				  (setq id-start (- (point) 1))
				  (setq last-expr-fin id-start)
				  (setq last-sub-id-start id-start)
				  )
				)
			(if (eq state 'ident-dot)
				(setq last-sub-id-start (- (point) 1))
			  )
			)
		  )
		(setq state (ident-search-state state 'letter))
		)
	   (t
		;; something unexpected, reset state
		(setq state 'base)
		)
	   )
	  )
;;(message "ID START == %d" id-start)
	(if (< last-dot id-start)
		(setq last-dot -1)
	  )
	;; Why is there the second check on 'base ???
	(if (and (not (eq id-start -1)) (not (eq state 'base)))
		(progn
		 (goto-char id-start)
		 (setq res (list id-start last-expr-fin last-sub-id-start last-dot nil))
		 )
	  (setq res nil)
	  (backward-char 1)
	  )
	res
))

;; return list of position of   id-beg,id-end,dot,  id-beg,id-end,dot  end....
(defun xref-parse-static-qualified-name ()
  (let ((res) (par-level) (state) (inloop) (start-pos) (ccc))
	(setq start-pos (point))
	(setq state 'base)
	(setq par-level 0)
	(setq inloop t)
	(setq res nil)
	(while inloop
	  (setq ccc (char-after (point)))
	  (forward-char 1) 
	  (cond
	   ;; character constant
	   ((eq ccc ?\')
		(setq state 'base)
		(backward-char)
		(search-forward-regexp "\\([^\\]'\\|\n\\)"))
	   ;; string
	   ((eq ccc ?\")
		(setq state 'base)
		(backward-char)
		(search-forward-regexp xref-string-passing-regexp))
	   ((and (eq ccc ?/) (eq (char-after (point)) ?*)
			 (eq (char-after (+ (point) 1)) ?\&))
		;; ignore /*& programm-comment
		(forward-char 2)
		)
	   ((and (eq ccc ?\&) (eq (char-after (point)) ?*)
			 (eq (char-after (+ (point) 1)) ?/))
		;; ignore &*/ programm-comment
		(forward-char 2)
		)
	   ((and (eq ccc ?/) (eq (char-after (point)) ?/)
			 (eq (char-after (+ (point) 1)) ?\&))
		;; ignore //& programm-comment
		(forward-char 2)
		)
	   ((and (eq ccc ?/) (eq (char-after (point)) ?*) 
			 (not (eq (char-after (+ (point) 1)) ?\&)))
		(setq state (ident-search-state state 'blank))
		;; /* */ coment
		(search-forward "*/"))
	   ;; // comment
	   ((and (eq ccc ?/) (eq (char-after (point)) ?/)
			 (not (eq (char-after (+ (point) 1)) ?\&)))
		(setq state (ident-search-state state 'blank))
		(search-forward "\n"))
	   ;; 
	   ((or (eq ccc ?\( ) (eq ccc ?\[ ))
		(setq par-level (+ par-level 1))
		(setq state 'base)
		)
	   ((or (eq ccc ?\) ) (eq ccc ?\] ))
		(if (> par-level 0)
			(setq par-level (- par-level 1))
		  (setq inloop nil)
		  )
		(setq state 'base)
		)
	   ((or (eq ccc ?\ ) (eq ccc ?\t) (eq ccc ?\r) (eq ccc ?\n))
		;; blank, ignore them
		(setq state (ident-search-state state 'blank))
		)
	   ((eq ccc ?\.)
		;; dot ignored
		(setq state (ident-search-state state 'dot))
		(if (eq par-level 0)
			(setq res (append res (list (- (point) 1))))
		  )
		)
	   ((or (and (>= ccc ?A) (<= ccc ?Z)) (and (>= ccc ?a) (<= ccc ?z))
			(eq ccc ?\_) (eq ccc ?$)
			(and (>= ccc ?0) (<= ccc ?9) (eq state 'in-ident))
			)
		;; identifier
		(if (or (eq state 'base) (eq state 'ident-blank) (eq state 'ident-dot))
			(progn 
			  (if (eq par-level 0)
				  (setq res (append res (list (- (point) 1))))
			  ))
		  )
		(setq state (ident-search-state state 'letter))
		)
	   (t
		;; something unexpected, reset state
		(setq inloop nil)
		)
	   )
	  )
	(setq res (append res (list (- (point) 2))))
	(goto-char start-pos)
;;	(message "%S" res)
	res
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; delete prefix, but not comments and newlines
(defun xref-delete-useless-prefix (s-point o-point)
  (let ((count) (ppp) (ccc))
	(goto-char s-point)
	(setq count (- o-point s-point))
;;(message "deleting useless prefix from %d to %d (%d)" s-point o-point count)
	(while (> count 0)
	  (setq ccc (char-after (point)))
	  (cond
	   ((and (eq ccc ?/) (eq (char-after (+ (point) 1)) ?*)
			 (eq (char-after (+ (point) 1)) ?\&))
		;; ignore /*& programm-comment
		(forward-char 3)
		(setq count (- count 3))
		)
	   ((and (eq ccc ?\&) (eq (char-after (+ (point) 1)) ?*)
			 (eq (char-after (+ (point) 2)) ?/))
		;; ignore &*/ programm-comment
		(forward-char 3)
		(setq count (- count 3))
		)
	   ((and (eq ccc ?/) (eq (char-after (+ (point) 1)) ?/)
			 (eq (char-after (+ (point) 2)) ?\&))
		;; ignore //& programm-comment
		(forward-char 3)
		(setq count (- count 3))
		)
	   ((and (eq ccc ?/) (eq (char-after (+ (point) 1)) ?*))
		;; /* */ coment
		(setq ppp (point))
		(search-forward "*/")
		(setq count (- count (- (point) ppp)))
		)
	   ;; // comment
	   ((and (eq ccc ?/) (eq (char-after (+ (point) 1)) ?/))
		(setq ppp (point))
		(search-forward "\n")
		(setq count (- count (- (point) ppp)))
        )
	   ;; 
	   ((or (eq ccc ?\r) (eq ccc ?\n) (eq ccc ?\ ) (eq ccc ?\t))
		(setq count (- count 1))
		(forward-char 1)
		)
	   ((or (eq ccc ?\.) (eq ccc ?\_) (eq ccc ?$)
			(and (>= ccc ?A) (<= ccc ?Z)) (and (>= ccc ?a) (<= ccc ?z))
			(and (>= ccc ?0) (<= ccc ?9))
			)
		(setq count (- count 1))
		(delete-char 1)
		)
	   (t
		;; something unexpected, reset state
		(error "xref internal error: wrong character inside name prefix")
		)
	   )
	  )
))

(defun xref-fqt-name-modify-static-prefix-check ()
  (let ((poin) (s-point) (uprefix) (up0) (up1) (lastdot) (up2))
	(if (and (not (xref-last-move-was-to-definition-or-declaration))
			 (not (xref-last-move-was-to-constructor-definition)))
		(progn
		  (setq poin (point))
		  (if (> poin xref-qualified-name-max-context)
			  (setq s-point (- poin xref-qualified-name-max-context))
			(setq s-point 0)
			)
		  (setq uprefix (xref-detect-qualified-name s-point poin nil))
		  (if uprefix
			  (progn
				(setq up0 (car uprefix))
				(setq up1 (car (cdr uprefix)))
				(setq up2 (car (cdr (cdr uprefix))))
				(setq lastdot (car (cdr (cdr (cdr uprefix)))))
				(if (not (equal up0 up1))
					(error "** Expression reference, don't know to handle this, please turn it to simple qualified name first!")
				  )
				))
		  ))
	nil
))

(defun xref-fqt-name-modify-static-prefix (new-prefix delete-name-flag)
	(let ((poin) (s-point) (uprefix) (up0) (up1) (lastdot) (up2))
	  (if (and (not (xref-last-move-was-to-definition-or-declaration))
			   (not (xref-last-move-was-to-constructor-definition)))
		  (progn
			(setq poin (point))
			(if (> poin xref-qualified-name-max-context)
				(setq s-point (- poin xref-qualified-name-max-context))
			  (setq s-point 0)
			  )
			(setq uprefix (xref-detect-qualified-name s-point poin nil))
			(if uprefix
				(progn
				  (setq up0 (car uprefix))
				  (setq up1 (car (cdr uprefix)))
				  (setq up2 (car (cdr (cdr uprefix))))
				  (setq lastdot (car (cdr (cdr (cdr uprefix)))))
				  (if (not (equal up0 up1))
					  (message "** Not a static reference at %s:%d, do not know to do this, sorry **" (xref-file-last-name (buffer-file-name)) (count-lines 1 (point)))
					;; else make action
					(xref-delete-useless-prefix up1 up2)
					(insert new-prefix)
					(if delete-name-flag (xref-delete-pending-ident))
					)
				  ))
			))
	  nil
))

(defun xref-move-field-make-move ()
  (let ((pp) (bp) (ep) (nbp) (def-string) (def-offset) (ofile) (oline))
	(setq pp (point))
	(beginning-of-line)
	(setq ofile (buffer-file-name))
	(setq oline (count-lines 1 (+ (point) 1)))
	(setq bp (point))
	(setq def-offset (- pp bp))
	(end-of-line)
	(setq ep (point))
	(setq def-string (buffer-substring bp ep))
	(goto-char bp)
	(delete-char (- ep bp))
	(xref-switch-to-marker xref-move-target-marker)
	(beginning-of-line)
	(insert def-string)
	(beginning-of-line)
	(setq nbp (point))
	(goto-char (+ nbp def-offset))
))

(defun xref-move-static-field-finish()
  (let ((tmp-mark) (ofile) (oline) (nfile) (nline) (safetycheckopt) (moved-pos-mark))
	;; should be via markers?
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
	(xref-moveref t)

	(setq ofile (buffer-file-name))
	(setq oline (count-lines 1 (+ (point) 1)))
	(xref-move-method-move-method-body xref-move-target-marker t xref-move-method-end-marker)
	;; (xref-move-field-make-move)
	(setq nfile (buffer-file-name))
	(setq nline (count-lines 1 (+ (point) 1)))

	(xref-refactoring-though-point "STEP 3. Making field public (temporarily) and reducing (possible) long names.")
	(xref-non-int-change-field-access-attribute "public" nil)
	(setq moved-pos-mark (make-marker))
	(set-marker moved-pos-mark (point))

	(setq tmp-mark (make-marker))
	(set-marker tmp-mark (point))
	(xref-nonint-reduce-useless-long-names-of-method (point) nil)
	(xref-switch-to-marker tmp-mark)

	(xref-refactoring-though-point "STEP 4. Updating Tag File")
	(message "Updating Tag file, please wait.")
	(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)

	(xref-refactoring-though-point 
	 "STEP 5. Reducing redundant long references")

	;; repush references (for reduce long names) as they are now shifted
	(xref-poponly)	
	(xref-switch-to-marker moved-pos-mark)
	(push-references 'sync (buffer-file-name) (point) t nil "-olcxrename" 0 nil)
	(xref-moveref nil) ???
	(xref-switch-to-marker moved-pos-mark)
	(xref-nonint-remove-useless-long-names (point) 
		  xref-detailed-refactoring-confirmations nil 
		  (list 'remove-useless-long-names-yet-pushed))

	(xref-poponly)	
	(xref-switch-to-marker tmp-mark)

	;; (xref-save-modified-files-with-question t "All done. ")

;;	(if (xref-safety-check-prompt)
;;		(progn
;;		  ;; update should be sufficient (?for most cases)
;;		  (xref-non-interactive-update-references xref-fast-updates-inside-refactorings)
;;		  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxsafetycheckinit" 
;;						   "-olcxrefsuffix=0:#" "")
;;		  (message "Checking new references.")
;;		  (setq 
;;		   safetycheckopt
;;		   (concat
;;			"-olcxsafetycheck2 "
;;			(format "-olcxsafetycheckmovedfile \"%s\" \"%s\" " ofile nfile)
;;			(format "-olcxsafetycheckmovedblock=%d:%d:%d" oline 1 nline)
;;			))
;;		  (push-references 'sync (buffer-file-name) (point) t nil safetycheckopt 0 '(xref-safety-check-finish))
;;		  ))
	(xref-refactoring-final-update-actions)
))

(defun xref-trivial-pre-check-post-action ()
  (if (equal xref-pilot-char ?\:)
	  (progn
		(beep t)
		(if (not (xref-yes-or-no-in-minibuffer 
				  "Continue this refactoring without fixing the problem? " "no"))
			(error "Refactoring aborted.")
		  ))
	(if (equal xref-pilot-char ?\;)
		(progn
		  (listrefs-continuation)
		  (xref-current-reference)
		  (beep t)
		  (if (not (xref-yes-or-no-in-minibuffer
					"Continue this refactoring without fixing the problem? " "no"))
			  (error "Refactoring aborted. Enlarge accessibilities first.")
			))
	  (if (not (equal xref-pilot-char ?*))
		  (error "Refactoring aborted.")
		)))
)

(defun xref-trivial-pre-checks (option)
  (let ((res))
	;; symrequire argument is nil because of getimportline, ...
	(setq res (push-references 'sync (buffer-file-name) (point) nil nil 
							   (concat "-olcxtrivialprecheck " option) 0 
							   '(xref-trivial-pre-check-post-action)))
	(if res
		(error res)
		)
))

(defun xref-trivial-post-check-post-action (origmarker )
  (if (equal xref-pilot-char ?\:)
	  (progn
		(xref-switch-to-marker origmarker)
		(beep t)
		(if (not (xref-yes-or-no-in-minibuffer
				  "Continue this refactoring without fixing the problem? " "no"))
			(progn
			  (kill-buffer xref-move-method-pre-check-tmp-buffer)
			  (error "Refactoring aborted.")
			  )
		  ))
	  (if (not (equal xref-pilot-char ?*))
		  (progn
			(kill-buffer xref-move-method-pre-check-tmp-buffer)
			(xref-switch-to-marker origmarker)
			(error "Refactoring aborted.")
	  )))
)

(defun xref-trivial-post-checks (option origmarker asfile)
  (let ((res))
	;; symrequire argument is nil because of getimportline, ...
	(setq res (push-references 'sync asfile (point) nil nil 
							   (concat "-olcxtrivialprecheck " option) 0 
							   (list 'xref-trivial-post-check-post-action origmarker)))
	(if res
		(error res)
		)
))

(defun xref-check-if-target-is-set (check-class)
  (if (equal xref-move-target-class "")
	  (error "No target class, use 'Set Target Position' to set target class.")
	)
  (if check-class
	  (progn
		(if (not (xref-move-target-is-class))
			(error "Target position is not pointing inside a class, reset it.")
		  )))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-move-static-field (poin)
"Move  a static  field  into  another class.   

At invocation of this function the  cursor has to be positioned on the
field defintion  (on its  name).  The field  has to be  declared alone
i.e.   no   other  field  declaration  should  share   its  type  name
(declarator).   The target  place and  class where  the field  will be
moved has  to be specified by  the 'Set Target  Place' function before
invocation of field moving.

  The refactoring  will make  the field public,  you can  restrict the
modifier after.
"
(interactive "d")
(let ((name) (target-class))
  (xref-initial-file-suffix-check xref-java-allowed-suffixes
	  "You should edit a Java file when using this function.")
  (xref-trivial-pre-checks "-tpchmovestaticfield")
  (xref-delete-windows)
  (setq name (xref-get-identifier-on-point))
  (xref-check-if-target-is-set t)
  (setq target-class (xref-char-replace xref-move-target-class "/$" ?.))
  (if (yes-or-no-p 
	   (format 
		"Move '%s' to class %s (%s:%d)? " 
		name target-class (xref-file-last-name (xref-move-target-file)) 
		(xref-move-target-line)))
	  (progn
		
		(xref-refactoring-init-actions (format "moving of %s" name))

		(set-marker xref-tmp-mark (point))
   		(xref-refactoring-though-point 
		 "STEP 1. Expanding context dependent names, if any. (Hit any key to continue)")
   		(xref-expand-short-method-refs (point) nil)
		(xref-switch-to-marker xref-tmp-mark)
		
		(save-excursion
		  (goto-line (xref-get-method-borne-line))
		  (set-marker xref-move-method-end-marker (point))
		  )

		(xref-refactoring-though-point 
		 "STEP 2. Changing references and moving field (hit any key to continue)")
		
		(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxrename"
							   xref-detailed-refactoring-confirmations
							   ''(xref-fqt-name-modify-static-prefix-check)
							   ''none-pre-fun
							   ''xref-fqt-name-modify-static-prefix
							   ''standard-post-fun 
							   (concat target-class ".")
							   nil
							   nil
							   '(xref-move-static-field-finish)
							   )
))))


(defun xref-move-method-move-method-body (target-marker delete-flag method-end-marker)
  (let ((pp) (bp) (ep) (nbp) (def-string) (def-offset))

	(setq def-offset (xref-current-column))
	(beginning-of-line)
	(setq bp (point))

	;; goto end of the method
	(xref-switch-to-marker method-end-marker)
	(beginning-of-line)

	(setq ep (point))
	(setq def-string (buffer-substring bp ep))

	(if delete-flag
		(progn
		  (goto-char bp)
		  (delete-char (- ep bp))
		  ))

	(set-buffer (marker-buffer target-marker))
	(goto-char (marker-position target-marker))
	(insert def-string)
	(goto-char (+ (marker-position target-marker) def-offset))
))


(defun xref-move-static-method-finish()
  (let ((moved-pos-mark))

	(xref-switch-to-marker xref-move-target-marker)
	(beginning-of-line)
	(set-marker xref-tmp-mark (point))

	;; should be via markers?
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
	(xref-moveref t)
	(xref-move-method-move-method-body xref-tmp-mark t xref-move-method-end-marker)
	(xref-non-int-change-field-access-attribute "public" nil)
	(setq moved-pos-mark (make-marker))
	(set-marker moved-pos-mark (point))

	;; following will refresh the screen
	(switch-to-buffer (current-buffer))
	(xref-refactoring-though-point "STEP 4. Updating Tag File")

	(message "Updating Tag file, please wait.")
	(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)

	;; repush references (for reduce long names) as they are now shifted
	(xref-poponly)	
	(xref-switch-to-marker moved-pos-mark)
	(xref-refactoring-though-point "STEP 5. Reduce Redundant Long Invocations")
	(push-references 'sync (buffer-file-name) (point) t nil "-olcxrename" 0 nil)
	(xref-moveref nil) ???
	(xref-switch-to-marker moved-pos-mark)
	(xref-nonint-remove-useless-long-names (point) xref-detailed-refactoring-confirmations nil 
	   (list 'remove-useless-long-names-yet-pushed))

	(xref-poponly)	
	(xref-switch-to-marker moved-pos-mark)

	(xref-refactoring-though-point 
	 "STEP 6. Reduce Redundant Long Names in the Method")

	(xref-nonint-reduce-useless-long-names-of-method (point) nil)
	;;	(xref-save-modified-files-with-question t "All done. ")
	(xref-refactoring-final-update-actions)
))

(defun xref-yes-or-no-in-minibuffer (prompt default)
  (let ((aa) (res))
	(setq res 0)
	(while (eq res 0)
	  (setq aa (read-from-minibuffer (concat prompt "(yes/no) ") default))
	  (if (equal aa "yes")
		  (setq res t)
		(if (equal aa "no")
			(setq res nil)
		  ))
	  (if (eq res 0)
		  (progn
			(message "Please answer yes or no.")
			(sit-for 0.5)
			))
	  )
	res
))

(defvar xref-move-method-pre-check-tmp-buffer " *xref-temporary-buffer*")

(defun xref-move-method-context-pre-check (pre-check-option 
										   target-modify-action
										   post-check-option
										   abort-message)
  (interactive "")
  (let ((target-buffer) (spos) (tpos))
	(set-marker xref-move-tmp-mark (point))
	;; first pushing of original parse-tree
	(push-references 'sync (buffer-file-name) (point) nil nil 
					 "-olcxpushallinmethod -olallchecks" 0 nil)
	(xref-switch-to-marker xref-move-target-marker)
	(beginning-of-line)
	(setq tpos (point))
	(setq target-buffer (current-buffer))
	(get-buffer-create xref-move-method-pre-check-tmp-buffer)
	(set-buffer xref-move-method-pre-check-tmp-buffer)
	(xref-erase-buffer)
	(insert-buffer target-buffer)
	(goto-char tpos)
	(set-marker xref-tmp-mark (point))
	
	;; goto first reference (this should be method definition)
	(xref-call-process 'sync 0 nil 0 0 "" "-olcxgoto1" "-olcxrefsuffix=0:" "")
	(xref-moveref t)

	;; handle special case when source buffer equals target buffer
	(if (equal (buffer-file-name) (xref-move-target-file))
		(progn
		  (setq spos (point))
		  (set-buffer (marker-buffer xref-tmp-mark)) ;; set-buffer
		  (goto-char spos)
		  ))
	(xref-move-method-move-method-body xref-tmp-mark nil xref-move-method-end-marker)
	;; TRIVIAL_POST_CHECKS
	(if post-check-option 
		(progn
		  (xref-trivial-post-checks post-check-option xref-move-tmp-mark
									(xref-move-target-file))
		  ))
	;; now pushing parse-tree in new location
	(xref-call-process 'sync 1 (xref-move-target-file) (point) 0 "" 
					   "-olcxpushallinmethod -olallchecks" 
					   "-olcxrefsuffix=0:#" "")
	;; following was commented, Why ??? Because of debugging !
	(kill-buffer xref-move-method-pre-check-tmp-buffer)

	;; following is for case of refresh display
	(xref-switch-to-marker xref-move-tmp-mark)

	(xref-call-process 'sync 0 nil 0 0 "" pre-check-option "-olcxrefsuffix=0:#" "")
	(if (equal xref-pilot-char ?\!)
		(error "Something is going wrong, aborting.")
	  )
	(if (equal xref-pilot-char ?\;)
		(progn
		  (xref-after-pushing-continuation t nil)
		  (beep t)
		  (if (xref-yes-or-no-in-minibuffer
			   "References misinterpreted in new location. Will you fix them? "
			   "yes"
			   )
			  (progn
				(listrefs-continuation)
				(error abort-message)
				)
			(xref-delete-windows)
			(xref-switch-to-marker xref-move-tmp-mark)
			)))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-get-method-borne-line ()
  (let ((res))
	(xref-call-process 'sync 1 nil (point) 0 "" "-olcxmethodlines" 
					   "" "-olcxrefsuffix=0:#")
	(if (not (eq xref-pilot-char ?\*))
		(error "Can't get method bornes, position the cursor on a method definition.")
	  (message "")
	  )
	(setq res (string-to-int xref-last-message))
	res
))

(defun xref-get-class-borne-line ()
  (let ((res))
	(xref-call-process 'sync 1 nil (point) 0 "" "-olcxclasslines" 
					   "" "-olcxrefsuffix=0:#")
	(if (not (eq xref-pilot-char ?\*))
		(error "Can't get class bornes, position the cursor on a class definition.")
	  (message "")
	  )
	(setq res (string-to-int xref-last-message))
	res
))

(defun xref-move-static-method (poin)
"Move  a static  method into  another  class.  

At  invocation of this  function the  cursor has  to be  positioned on
method's  definition (on  its name).   The method  declaration  has to
start on this line and no other  field or method has to be declared on
the same  line.  The target place  and class where the  method will be
moved has  to be  specified by the  'Set Target Place'  function.  The
method will be moved in  four steps.  1.)  all context dependent names
inside the method will be expanded to fully qualified names.  2.)  all
references inside method will be checked to remain correct in method's
new place.  3.)  the method  will be moved and all references updated.
4.)  redundant fully qualififed names will be reduced.

  The refactoring  will make the  method public, you can  restrict the
modifier after.
"
(interactive "d")
(let ((name) (target-class))
  (xref-initial-file-suffix-check xref-java-allowed-suffixes
	  "You should edit a Java file when using this function.")
  (xref-trivial-pre-checks "-tpchmovestaticmethod")
  (xref-delete-windows)
  (setq name (xref-get-identifier-on-point))
  (xref-check-if-target-is-set t)
  (setq target-class (xref-char-replace xref-move-target-class "/$" ?.))
  (save-excursion
	(goto-line (xref-get-method-borne-line))
	(set-marker xref-move-method-end-marker (point))
	)
  (if (yes-or-no-p 
	   (format 
		"Move '%s' to class %s (%s:%d)? " 
		name target-class (xref-file-last-name (xref-move-target-file)) 
		(xref-move-target-line)))
	  (progn
		(xref-refactoring-init-actions (format "moving of %s" name))

		(set-marker xref-tmp-mark (point))
   		(xref-refactoring-though-point 
		 "STEP 1. Expanding all context dependent names. (Hit any key to continue)")
   		(xref-expand-short-method-refs (point) nil)
		(xref-switch-to-marker xref-tmp-mark)
		
		(xref-refactoring-though-point "STEP 2. Context check.")
		(xref-move-method-context-pre-check "-olcxmmprecheck" nil nil
		  "Aborting. To resolve the problem move critical symbols or make them public.")

		(xref-refactoring-though-point 
		 "STEP 3. Moving the method and updating all references")
		
		(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxrename"
							   xref-detailed-refactoring-confirmations
							   ''(xref-fqt-name-modify-static-prefix-check)
							   ''none-pre-fun
							   ''xref-fqt-name-modify-static-prefix
							   ''standard-post-fun 
							   (concat target-class ".")
							   nil
							   nil
							   '(xref-move-static-method-finish)
							   )
))))

;;;;;;;;;;;      Turn Dynamic Method into Static    ;;;;;;;;;;

(defun xref-fqt-name-dynamic-to-static (target-class super-class)
	(let ((poin) (s-point) (uprefix) (up0) (up1) (up2) (prefix-str)
		  (lastdot) (res nil) (fqtname))
	  (setq poin (point))
	  (if (> poin xref-qualified-name-max-context)
		  (setq s-point (- poin xref-qualified-name-max-context))
		(setq s-point 0)
		)
	  (setq uprefix (xref-detect-qualified-name s-point poin t))
	  (if uprefix
		  (progn
			(setq up0 (car uprefix))
			(setq up1 (car (cdr uprefix)))
			(setq up2 (car (cdr (cdr uprefix))))
			(setq lastdot (car (cdr (cdr (cdr uprefix)))))
			(if (equal lastdot -1)
				(progn
				  (setq fqtname (xref-get-current-ref-appl-class))
				  (if (string-match "\\.[0-9]" fqtname)
					  ;; Anonymous nested class, no expansion
					  (setq prefix-str "this")
					(setq prefix-str (format "%s.this" fqtname))
					)
				  ;;(setq prefix-str (format "%s.this" target-class)) 
				  ;; has to be vApplClass not target-class
				  )
			  (setq prefix-str (buffer-substring up0 (- lastdot 1)))
			  )
			(goto-char poin)
			(pre-position-to-argument-insert 1 nil)
			(setq res (insert-argument
					   1 (list 
						  (format "%s %s" target-class
								  xref-dynamic-to-static-new-arg-name)
						  prefix-str
						  )))
			(if xref-insert-arg-renaming-new-name
				(setq xref-dynamic-to-static-new-arg-name xref-insert-arg-renaming-new-name)
			  )
			(if (not (xref-last-move-was-to-definition-or-declaration))
				(progn
				  (goto-char up0)
				  (delete-char (- up2 up0))
				  (insert target-class ".")
			))))
	  res
))

(defun xref-dyn-to-st-optionally-delete-qualified-this-prefix (this-class)
  (let ((qualif) (tcl))
  (if (eq (char-after (- (point) 1)) ?.)
	  (progn
		;; all this is very textual, very bad, but I affraid just to
		;; delete qualified prefix without any checks, TODO do it better
		(setq qualif "-")
		(setq tcl (length this-class))
		(if (> (point) tcl)
			(setq qualif (buffer-substring (- (point) tcl 1) (- (point) 1)))
		  )
		(if (equal qualif this-class)
			(delete-char (- 0 tcl 1))
		  (xref-wait-until-problem-confirmed "Hmm. Qualified 'this/supper'. Need to be adjust manually.")
		  )
		)
	)
))

(defun xref-dynamic-to-static-insert-new-arg (this-class super-class)
  (let ((id) (qualif) (tcl))
	(setq id (xref-get-identifier-on-point))
	;; 'this' and 'super' can be qualified only by this type name (this-class)
	;; otherwise outer scope reference would be reported by pre-checks,
	;; at least I hope!
	;; NO! There can be 'this' references in inner classes nested inside the
	;; method
	(if (equal id "this")
		(progn
		  (xref-dyn-to-st-optionally-delete-qualified-this-prefix this-class)
		  (delete-char 4)
		  (insert xref-dynamic-to-static-new-arg-name)
		  )
	  (if (equal id "super")
		  (progn
			;; 'super' probably can not be qualified in this context, but anyway
			(xref-dyn-to-st-optionally-delete-qualified-this-prefix this-class)
			(delete-char 5)
			(insert 
			 (format "((%s)%s)" super-class 
					 xref-dynamic-to-static-new-arg-name))
			)
		;; neither this or super
		(insert xref-dynamic-to-static-new-arg-name ".")
		(if (equal id "new")
			(progn
			  ;; TODO!!! delete fully qualified name after NEW, 
			  ;; keep only single
			  nil
			  )
		  )))
))

(defun xref-turn-dynamic-to-static-finish2 ()
	;; finally insert the static keyword
  (let ((defm))
	(xref-popCxrefs (point))
	;; should be via markers?
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
	(xref-moveref t)
	(setq defm (point-marker))
	(beginning-of-line)
	(search-forward-regexp "[^ \t]")
	(forward-char -1)
	(insert "static ")
	(goto-char (marker-position defm))

	(xref-refactoring-though-point "STEP 3. Updating Tag File")

	(message "Updating Tag file, please wait.")
	(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)

	;; do not need to repush references here? Do it to be sure
	(xref-poponly)	
	(xref-switch-to-marker defm)
	(push-references 'sync (buffer-file-name) (point) t nil "-olcxrename" 0 nil)
	(xref-moveref nil) ???
	(xref-switch-to-marker defm)

	(xref-refactoring-though-point 
	 "STEP 4. Reduce introduced redundant long names")
	(xref-nonint-remove-useless-long-names 
	 (point) 
	 xref-detailed-refactoring-confirmations nil 
	 (list 'remove-useless-long-names-yet-pushed))

	;; I have put this here, but it seems it was not the right thing?
	(xref-popCxrefs (point))
	(xref-switch-to-marker defm)
	(xref-nonint-reduce-useless-long-names-of-method (point) nil)

	(xref-switch-to-marker defm)
	(xref-refactoring-final-update-actions)
))

(defun xref-turn-dynamic-to-static-finish (this-class super-class)
  (let ((poin))
	;; update all maybethis references		  ;; should be via markers?
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
	(xref-moveref t)
	(setq poin (point))

	(xref-refactoring-though-point 
	 "STEP 2. Redirect method's nonstatic references to new argument")

	(xref-map-on-all-references  (buffer-file-name) poin nil nil 
								 "-olcxmaybethis"
								 nil   ;; non-interactive for local changes 
								 nil
								 ''none-pre-fun
								 ''xref-dynamic-to-static-insert-new-arg
								 ''standard-post-fun 
								 this-class
								 super-class
								 nil
								 '(xref-turn-dynamic-to-static-finish2)
						   )
))

(defun xref-dts-par-name (class)
  (let ((res) (p) (i))
	(setq i (- (length class) 1))
	(setq p 0)
	(while (and (>= i 0) (not (eq (elt class i) ?\.)))
	  (setq i (- i 1))
	  )
	(setq res (substring class (+ i 1)))
	(if (equal (length res) 1)
		(setq res (downcase res))
	  (if (> (length res) 1)
		  (setq res (concat (downcase (substring res 0 1)) (substring res 1)))
		))
))

(defun xref-turn-dynamic-to-static (poin)
"Turn virtual method  into static.

This function inserts method's  object as new parameter. When invoking
this  function  the  cursor  has  to be  positioned  on  the  method's
definition (on its name).  REQUIRED PRECONDITIONS:
 (1) Method is never invoked  via virtual mechanism. I.e.  all method's
invocations can be recognized syntactically at compile time.
 (2) If the method is defined  in an inner class it can't use nonstatic
fields from its lexically enclosing (outer) classes.
"
  (interactive "d")
  (let ((class-name) (super-class) (name))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
	   "You should edit a Java file when using this function.")
	(xref-trivial-pre-checks "-tpchturndyntostatic")
	(setq name (xref-get-identifier-on-point))
	(setq class-name (xref-get-current-class "-olcxcurrentclass" t))
	(setq super-class (xref-get-current-class "-olcxcurrentsuperclass" nil))
	(setq xref-dynamic-to-static-new-arg-name 
		  (read-from-minibuffer "Name of the new paramater: " 
								(xref-dts-par-name class-name)))
	(xref-refactoring-init-actions (format "turning %s static" name))
	(xref-refactoring-though-point 
	 "STEP 1. Method's references to static; adding new argument (hit any key to continue)")
	(xref-map-on-all-references  (buffer-file-name) (point) 
								 t nil "-olcxdynamictostatic1"
								 xref-detailed-refactoring-confirmations
								 nil
								 ''none-pre-fun
								 ''xref-fqt-name-dynamic-to-static
								 ''standard-post-fun 
								 class-name
								 super-class
								 nil
								 (list 'xref-turn-dynamic-to-static-finish
									   class-name super-class 
									   )
						   )
))


;;;;;;;;;;;      Turn Static Method into Dynamic    ;;;;;;;;;;

(defun xref-exact-string-match (regexp str)
  (let ((res) (si) (ei) (rr))
	(setq rr (string-match (concat ">>>>>" regexp "<<<<<")
						   (concat ">>>>>" str    "<<<<<")))
	(setq res (eq rr 0))
	res
))

;; detect whether two field names are equal
(defun xref-same-field-name (name1 name2)
  (let ((res))
;; this may cause problems if using wildcard characters!
	(setq res (xref-exact-string-match (format "[ \t\n]*%s[ \t\n]*" name1) name2))
	(setq res (or res (xref-exact-string-match (format "[ \t\n]*%s[ \t\n]*" name2) name1)))
	res
))

(defun xref-st-to-dyn-parenthesize-if-needed (ss aprefixs)
  (let ((res))
	(setq res ss)
	(if (not (xref-exact-string-match "(.*)" res))
		(setq res (format "(%s)" res))
	  )
	res
))

(defun xref-remove-newlines-from-string (ss)
  (let ((i) (len) (res))
	(setq len (length ss))
	(setq i 0)
	(setq res "")
	(while (< i len)
	  (if (eq (elt ss i) ?\n)
		  (setq res (format "%s " res))
		(setq res (format "%s%c" res (elt ss i)))
		)
	  (setq i (1+ i))
	  )
	res
))

(defun xref-fqt-name-static-to-dynamic (argn field)
	(let ((poin) (s-point) (up0) (up1) (up2) (lastdot) (result nil) (apos)
		  (args) (sr) (uprefix) (aprefixs) (eee) (bbb))
	  (setq poin (point))
	  (if (not (xref-last-move-was-to-definition-or-declaration))
		  (progn
			(if (> poin xref-qualified-name-max-context)
				(setq s-point (- poin xref-qualified-name-max-context))
			  (setq s-point 0)
			  )
			(setq uprefix (xref-detect-qualified-name s-point poin nil))
			(if uprefix
				(progn
				  (setq up0 (car uprefix))
				  (setq up1 (car (cdr uprefix)))
				  (setq up2 (car (cdr (cdr uprefix))))
				  (setq lastdot (car (cdr (cdr (cdr uprefix)))))
				  (goto-char poin)
				  (setq apos (pre-position-to-argument-delete argn nil))
				  (if apos
					  (setq result apos)
					(setq bbb (point))
					(setq eee (mark t))
					(setq args (buffer-substring bbb eee))
					;; here can be a problem if the argument contains newline,
					;; so after inserting it in prefix, it puts new line 
					;; and confuses pop (goto definition)
					;; handle this as a special case !
					(if (string-match "\n" args)
						(setq args (xref-remove-newlines-from-string args))
					  )
					;; (setq aprefixs (xref-detect-qualified-name bbb eee))
					(xref-delete-useless-prefix up1 up2)
					(if (not (xref-exact-string-match "[ \t\n]*this[ \t\n]*" args))
						(progn
;;						  (if (equal bbb (car aprefixs))
							  (insert args ".")
;;							(insert "(" args ")" ".")
;;							)
						  )
					  )
					(if (not (equal field ""))
						(insert field ".")
					  )
					))))
)))

(defun xref-insert-this-dot-if-needed (file-name)
  (let ((res))
	(insert "this.")
;; just for now, put there "-olcxmmprecheck", then implement special check ?
	(if (xref-is-useless-part-of-method file-name
										(- (point) 5)
										(- (point) 0)
										"-olcxmmprecheck"  
										)
		(delete-backward-char 5)
	  )
))

(defun xref-static-to-dynamic-fun-mapped-on-argument (argname field)
  (let ((result nil) (crn) (npl) (flen) (cfield))
	(if (not (xref-last-move-was-to-definition-or-declaration))
		(progn
		  (setq crn (xref-identifier-after-point))
		  (if (not (equal crn argname))
			  (error "[Xref] Reference not pointing to argument, internal error")
			)
		  (setq npl (xref-parse-static-qualified-name))
		  (if (> (length npl) 2)
			  (progn
				(if (equal field "")
					(progn
					  ;; but following may generate useless this references
					  (kill-region (nth 0 npl) (nth 2 npl))
					  (xref-insert-this-dot-if-needed (buffer-file-name))
					  )
				  (setq cfield (buffer-substring (nth 2 npl) (nth 3 npl)))
				  (if (xref-same-field-name field cfield)
					  (if (eq (length npl) 4)
						  (progn
							(kill-region (nth 0 npl) (nth 3 npl))
							(insert "this")
							)
						(kill-region (nth 0 npl) (nth 4 npl))
						(xref-insert-this-dot-if-needed (buffer-file-name))
					))))
			(if (equal field "")
				(progn
				  (kill-region (point) (mark t))
				  (insert "this")
			  ))
		  )))
	result
))

;; symbol must be on the definition position!
(defun xref-is-this-an-unused-symbol ()
  (let ((res nil) (defpos) (defbuff))
	(setq defbuff (current-buffer))
	;; use -rename instead of push, rename is always pushing, so pop recovers
	;; reference stack
	(push-references 'sync (buffer-file-name) (point) t nil "-olcxrename" 0 nil)
	(xref-moveref nil)
	(if (not (equal defbuff (current-buffer)))
		(error "[xref-is-symbol-unused] reference moved from current file, internal error!")
	  )
	(setq defpos (point))
	(xref-this-symbol-previous-ref (point))
	(if (and (equal defpos (point)) (equal defbuff (current-buffer)))
		(setq res t)
	  )
	(xref-popCxrefs (point))
	res
))

(defun xref-turn-static-to-dynamic-finish3 ()
  (xref-popCxrefs (point))
  (xref-save-some-buffers t)
  (message "All Done.")
)

(defun xref-turn-static-to-dynamic-finish2 (argn defmarker)
  (let ((pos) (defpos))
	(xref-popCxrefs (point))
	(if (xref-is-this-an-unused-symbol)
		(progn
		  ;; the argument has no usage, remove it

		  (xref-refactoring-though-point "STEP 3. Updating Tag File")
		  (message "Updating Tag file, please wait.")
		  (xref-non-interactive-update-references xref-fast-updates-inside-refactorings)

		  (xref-refactoring-though-point "STEP 4. Removing useless argument")
		  (goto-char (marker-position defmarker))
		  (xref-non-interact-arg-delete (point) argn nil 
										(list 'xref-turn-static-to-dynamic-finish3))
		  )
	  )
))

(defun xref-turn-static-to-dynamic-finish (argn field)
  (let ((defmarker) (poin) (sr) (argname))
	;; update all maybethis references		  ;; should be via markers?
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
	(xref-moveref t)
	(setq defmarker (make-marker))
	(set-marker defmarker (point))
	(xref-popCxrefs (point))
	(xref-switch-to-marker defmarker)
	(setq poin (point))
	;; definition, remove static keyword
	(beginning-of-line)
	(setq sr (search-forward "static" poin t))
	(if sr
		(progn
		  (delete-char -6)
		  (delete-char 1)         ;; also the space after static
		  ))
	(xref-switch-to-marker defmarker)
	(xref-move-to-paramater-name argn)
	(setq argname (xref-get-identifier-on-point))

	(xref-refactoring-though-point "STEP 2. Remove useless argument's usages")

	(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxrename"
						   nil ;; non-interactive local action
						   nil
						   ''none-pre-fun
						   ''xref-static-to-dynamic-fun-mapped-on-argument
						   ''standard-post-fun 
						   argname
						   field
						   nil
						   (list 'xref-turn-static-to-dynamic-finish2 argn defmarker)
						   )
))

(defun xref-get-type-of-symbol (target-fname)
  (let ((ftype))
  (setq xref-pilot-char ?\ )
  (xref-call-process 'sync 1 target-fname (point) 0 "" 
					 "-olcxgetsymboltype -noerrors"
					 "-olcxrefsuffix=0:" "")
  (if (eq xref-pilot-char ?\*)
	  (setq ftype xref-last-message)
	(setq ftype nil)
	)
  ftype
))

(defun xref-turn-static-to-dynamic-class-prechecks (argn field)
  (let ((tpos) (target-buffer) (argname) (this-class) 
		(bb) (errm) (accstr) (cont) (target-fname) (ftype))
	(setq errm nil)
	(setq tpos (point))
	(xref-move-to-paramater-name argn)
	(setq argname (xref-identifier-after-point))
	(goto-char tpos)
	(setq this-class (xref-get-current-class "-olcxcurrentclass" t))
	;; O.K. I have everything I need, now try if it fits
	;; do any changes on a copy in temporary buffer
	(setq target-buffer (current-buffer))
	(setq target-fname (buffer-file-name target-buffer))
	(get-buffer-create xref-move-method-pre-check-tmp-buffer)
	(set-buffer xref-move-method-pre-check-tmp-buffer)
	(xref-erase-buffer)
	(insert-buffer target-buffer)
	(goto-char tpos)
	;; following has to be reworked on semantic level
	(setq bb (search-forward "{" (point-max) t))
	(if (not bb)
		(setq errm "Sorry, can't find beginning of method, continue anyway? ")
   	  (if (equal field "")
		  (setq accstr argname)
		(setq accstr (format "%s.%s" argname field))
		)
	  (insert (format "xxx(%s);" accstr))
	  (backward-char 3)
	  (setq ftype (xref-get-type-of-symbol target-fname))
	  (if (not ftype)
		  (setq errm (format "Can't get type of %s, continue anyway? " accstr))		
		(if (not (equal ftype this-class))
			(error (format "'%s' is not of type %s" accstr this-class))
		  )
		)
	  )
	(kill-buffer xref-move-method-pre-check-tmp-buffer)
	(set-buffer target-buffer)
	(if (not (eq errm nil))
		(progn
		  (setq cont (yes-or-no-p errm))
		  (if (not cont) (error "Aborting."))
		  ))
))


(defun xref-turn-static-to-dynamic (poin)
"Turn static method into virtual.

The  cursor has  to  positioned  on the  method's  definition (on  its
name).  The target  object  will be  obtained  from method's  argument
either directly or  via an additional field.  For  example a method of
class 'C'  can be turned  into virtual only  if it has an  argument of
class 'C', or if it has an  argument of a class 'D' and this class 'D'
has a  field of type 'C'.  In  both cases we call  the obtained object
(of  class C)  the 'target'  object.   The xref-turn-static-to-dynamic
function will  prompt for the  order of argument  and for the  name of
field which will be used to determine target object.  In case that the
target  object is obtained  directly from  method's argument  (in this
case insert  the empty string as  field name) then this  function is a
reverse function to 'xref-turn-dynamic-to-static'.
"
  (interactive "d")
  (let ((argns) (arg) (name) (field))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
	    "You should edit a Java file when using this function")
	(xref-trivial-pre-checks "-tpchturnstatictodyn")
	(setq name (xref-get-identifier-on-point))
	(setq argns (read-from-minibuffer 
				 (format 
				  "Argument to use to determine method's object [ 1 - arity('%s') ]? " name)
				 "1"
				 ))
	(setq field (read-from-minibuffer
				 "Optionally a field name getting method's object from the argument [\"\"] "))
	(setq arg (string-to-int argns))
	(xref-turn-static-to-dynamic-class-prechecks arg field)
	(xref-refactoring-init-actions (format "turning %s dynamic" name))
	(xref-refactoring-though-point 
	 "STEP 1. Turn method's references to dynamic (hit any key to continue)")
	(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxrename"
						   xref-detailed-refactoring-confirmations
						   nil
						   ''none-pre-fun
						   ''xref-fqt-name-static-to-dynamic
						   ''standard-post-fun 
						   arg
						   field
						   nil
						   (list 'xref-turn-static-to-dynamic-finish arg field)
						   )
))


(defun xref-non-int-change-field-access-attribute (new-attribute with-safety-checks)
  (let ((field) (oldattr) (newattr) (dpos) (attrpos) (hist))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
	    "You should edit a Java file when using this function.")
	(setq field (xref-get-identifier-on-point))
	(setq dpos (point))
	(set-marker xref-tmp-mark (point))
	(beginning-of-line)
	(if (search-forward "public" dpos t)
		(progn
		  (setq attrpos (- (point) 6))
		  (setq oldattr "public")
		  )
	  (beginning-of-line)
	  (if (search-forward "private" dpos t)
		  (progn
			(setq attrpos (- (point) 7))
			(setq oldattr "private")
			)
		(beginning-of-line)
		(if (search-forward "protected" dpos t)
			(progn
			  (setq attrpos (- (point) 9))
			  (setq oldattr "protected")
			  )
		  (beginning-of-line)
		  (search-forward-regexp "[^ \t]" dpos t)
		  (setq attrpos (- (point) 1))
		  (setq oldattr "")
		  )))
	(setq xref-last-change-attr-old-attribute oldattr)
	(if (not new-attribute)
		(progn
		  (setq hist '("public" "private" "protected" ""))
		  (setq newattr (read-from-minibuffer 
						 (format "Change '%s' attribute to :" oldattr) 
						 oldattr nil nil 'hist))
		  )
	  (setq newattr new-attribute)
	  )	 
	(goto-char (marker-position xref-tmp-mark))
	(if with-safety-checks
		(progn
		  (xref-refactoring-init-actions (format "%s %s" field newattr))
		  (push-references 'sync (buffer-file-name) (point) t nil 
						   "-olcxpushonly -olnodialog" 0 nil)
		  ))
	(goto-char attrpos)
	(if (equal oldattr "")
		(insert " ")
	  )
	(goto-char attrpos)
	(insert newattr)
	(xref-delete-pending-ident)
	(goto-char (marker-position xref-tmp-mark))
	(if with-safety-checks
		(xref-trivial-safety-check)
	  )
))


(defun xref-change-field-access-attribute (poin)
"Change access attribute of a field.

This function changes  an access attribute of a Java  class field in a
secure way.  The cursor has  to be positioned on the field definition,
more exactly on its name. The field has to be defined on a single line
and no other field  can be defined on the same line  and with the same
declarator (through  comma separator).  You  will be prompted  for new
access  attribute  and the  attribute  will  be  replaced.  All  field
references will be checked and apparition of new or a lost of old will
be reported. Usually  if you change attribute from  public to private,
then this  function shows you  those references of the  original field
which  will be  lost.  If  you change  the attribute  from  private to
public you will receive those references which now accidentally refers
to this field, but originally have refered to a field of the same name
from a  nested scope (in such  case you should prefix  them by 'this.'
keyword in order to fix the problem)."
  (interactive "d")
  (xref-non-int-change-field-access-attribute nil t)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;; PULL UP / PUSH DOWN ;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun xref-expand-this-this-occurence (casts dummy)
  (let ((id) (this-cast) (super-cast) (spoint) (ctc "") (csc ""))
	(setq this-cast (car casts))
	(setq super-cast (car (cdr casts)))
	(setq spoint (point))
	(search-forward-regexp xref-forward-pass-identifier-regexp)
	(setq id (buffer-substring spoint (- (point) 1)))
	(if (equal id "this")
		(progn
		  (if (> (- (point) (length this-cast)) 0)
			  (setq ctc (buffer-substring (- (point) (length this-cast)) (point)))
			)
		  (if (> (- (point) (length super-cast)) 0)
			  (setq csc (buffer-substring (- (point) (length super-cast)) (point)))
			)
		  (if (and (not (equal ctc this-cast)) (not (equal csc super-cast)))
			  (progn
				;; only if the cast is not yet there
				(backward-char 1)
				(backward-delete-char 4)
				(insert this-cast)
				)))
	  (if (equal id "super")
		  (progn
			(backward-char 1)
			(backward-delete-char 5)
			(insert super-cast)
			)))
	nil
))

(defun xref-expand-this-references-in-method (file-name
											  this-class
											  super-class)
  (let ((this-cast) (super-cast))
	(setq this-cast (format "((%s)this)" this-class))
	(setq super-cast (format "((%s)this)" super-class))
	(xref-map-on-all-references file-name (point) nil nil
								"-olcxmaybethis"
								nil
								nil
								''none-pre-fun
								''xref-expand-this-this-occurence
								''none-post-fun
								(list 'quote (list this-cast super-cast));;arg1
								nil  ;; arg2?
								nil
								(list 'xref-popCxrefs (point))
								)
))

;; an undocumented function (used inside pull-up/push-down
(defun xref-add-casts-to-this-references-of-method ()
  "Expand occurences of 'this' keyword by adding casts to it."
  (interactive "")
  (let ((this-class)(super-class))
	(setq this-class (xref-get-current-class "-olcxcurrentclass" t))
	(setq super-class (xref-get-current-class "-olcxcurrentsuperclass" nil))
	(xref-expand-this-references-in-method (buffer-file-name) this-class super-class)
))

;;;;;;;;;;;;;;;;;;;;;;

(defun xref-reduce-this-this-occurence (casts dummy)
  (let ((id) (this-cast) (super-cast) (spoint) (ctc "") (csc "") (tcl) (scl))
	(setq this-cast (car casts))
	(setq super-cast (car (cdr casts)))
	(setq spoint (point))
	(search-forward-regexp xref-forward-pass-identifier-regexp)
	(setq id (buffer-substring spoint (- (point) 1)))
	(if (equal id "this")
		(progn
		  (setq tcl (length this-cast))
		  (if (> (- (point) tcl) 0)
			  (setq ctc (buffer-substring (- (point) tcl) (point)))
			)
		  (setq scl (length super-cast))
		  (if (> (- (point) scl) 0)
			  (setq csc (buffer-substring (- (point) scl) (point)))
			)
		  (if (equal ctc this-cast)
			  (progn
				(backward-delete-char tcl)
				(insert "this")
				)
			(if (and (equal csc super-cast) (equal (char-after (point)) ?.))
				(progn
				  (backward-delete-char scl)
				  (insert "super")
				  )))))
	nil
))

(defun xref-reduce-this-references-in-method (file-name
											  this-class
											  super-class)
  (let ((this-cast) (this-non-cast) (super-cast) (super-non-cast))
	(setq this-cast (format "((%s)this)" this-class))
	(setq super-cast (format "((%s)this)" super-class))
	(xref-map-on-all-references file-name (point) nil nil
								"-olcxmaybethis"
								nil
								nil
								''none-pre-fun
								''xref-reduce-this-this-occurence
								''none-post-fun
								(list 'quote (list this-cast super-cast));;arg1
								nil  ;; arg2?
								nil
								(list 'xref-popCxrefs (point))
								)
))

;; an undocumented function (used inside pull-up/push-down
(defun xref-remove-casts-to-this-references-in-method ()
  "Reduce occurences of some casts before 'this' keyword."
  (interactive "")
  (let ((this-class)(super-class))
	(setq this-class (xref-get-current-class "-olcxcurrentclass" t))
	(setq super-class (xref-get-current-class "-olcxcurrentsuperclass" nil))
	(xref-reduce-this-references-in-method (buffer-file-name) this-class super-class)
))


;;;;;;;;;;;;;;;;;;;;;;

(defun xref-offset-to-line-col-string (offset)
  (let ((res) (poin) (line) (col))
	(setq poin (point))
	(goto-char offset)
	(beginning-of-line)
	(setq line (count-lines 1 (+ (point) 1)))
	(setq col (- offset (point)))
	(goto-char poin)
	(setq res (format "%d:%d" line col))
	res
))

(defun xref-is-useless-part-of-method (file-name from-offset to-offset checkopt)
  (let ((obuf) (tpos) (res) (opt) (len) (code) (ii))
	(setq obuf (current-buffer))
	(setq tpos (point))
	(get-buffer-create " *xref-up-temporary-buffer*")
	(set-buffer " *xref-up-temporary-buffer*")
	(xref-erase-buffer)
	(insert-buffer obuf)
	(goto-char tpos)
	(setq len (- to-offset from-offset))
	(setq code (buffer-substring from-offset to-offset))
	(goto-char from-offset)
	;; push all refs
	(push-references 'sync file-name (point) nil nil 
					 "-olcxpushallinmethod" 0 nil)
	;; remove from them the tested window
	(setq opt (format "-olcxwindel -olcxwindelfile \"%s\"" file-name))
	(setq opt (format "%s -olcxwindelwin=%sx%s" opt
					  (xref-offset-to-line-col-string from-offset)
					  (xref-offset-to-line-col-string to-offset)
					  ))
	(xref-call-process 'sync 0 file-name 0 0 "" opt "-olcxrefsuffix=0:#" "")
	;; delete window in source
	(setq ii 0)
	(while (< ii len)
	  (delete-char 1)
	  (insert-char ?\  1)
	  (setq ii (+ ii 1))
	  )
	(push-references 'sync file-name (point) nil nil 
					 "-olcxpushallinmethod" 0 nil)
	(goto-char from-offset)
	(delete-char len)
	(insert code)
	;; check that no (other than in window) reference will be lost
	(xref-call-process 'sync 0 file-name 0 0 "" checkopt
					   "-olcxrefsuffix=0:#" "")
	(if (equal xref-pilot-char ?!)
		(error "Something is going wrong, aborting.")
	  )
	(if (equal xref-pilot-char ?\;)
		(progn
		  (setq res nil)
		  (xref-poponly)
		  )
	  (setq res t)
	  )
	(kill-buffer " *xref-up-temporary-buffer*")
	(set-buffer obuf)
	(goto-char tpos)
  res
))


(defun xref-reduce-some-pp-useless-this-cast-occurence (casts file-name)
  (let ((id) (cast) (castlist) (spoint) (ctc "") (tcl))
	(setq spoint (point))
	(search-forward-regexp xref-forward-pass-identifier-regexp)
	(setq id (buffer-substring spoint (- (point) 1)))
	(if (equal id "this")
		(progn
		  (setq castlist casts)
		  (while castlist
			(setq cast (car castlist))
			(setq tcl (length cast))
			(if (> (- (point) tcl) 0)
				(setq ctc (buffer-substring (- (point) tcl) (point)))
			  (setq ctc "")
			  )
			(if (equal ctc cast)
				(progn
				  ;; check on uselessness here !!!
				  (if (xref-is-useless-part-of-method file-name 
													  (+ (- (point) tcl) 1)
													  (- (point) 5)
													  "-olcxppprecheck"
													  )
					  (progn
						(backward-delete-char tcl)
						(insert "this")
						(setq castlist '(""))  ;; this will break the loop
				  ))))
			(setq castlist (cdr castlist))
			)))
	nil
))

(defun xref-reduce-useless-this-casts-in-pp-method (file-name
												 cast1
												 cast2)
  (let ((this-cast) (this-non-cast) (super-cast) (super-non-cast))
	(xref-map-on-all-references file-name (point) nil nil
								"-olcxmaybethis"
								nil
								nil
								''none-pre-fun
								''xref-reduce-some-pp-useless-this-cast-occurence
								''none-post-fun
								(list 'quote (list cast1 cast2))  ;;arg1
								file-name  ;; arg2?
								nil
								(list 'xref-popCxrefs (point))
								)
))

;; an undocumented function (used inside pull-up/push-down
(defun xref-remove-useless-casts-to-this-references-of-method ()
  "Reduce occurences of some useless casts before 'this' keyword."
  (interactive "")
  (let ((this-class) (super-class) (this-cast) (super-cast))
	(setq this-class (xref-get-current-class "-olcxcurrentclass" t))
	(setq super-class (xref-get-current-class "-olcxcurrentsuperclass" nil))
	(setq this-cast (format "((%s)this)" this-class))
	(setq super-cast (format "((%s)this)" super-class))
	(xref-reduce-useless-this-casts-in-pp-method (buffer-file-name) 
											  this-cast super-cast)
))


;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;



(defun xref-pull-up-push-down-reduce-expanded-syms (class1 class2)
  (let ((this-cast) (super-cast) (this-class)(super-class))
	(save-excursion
	  (setq this-cast (format "((%s)this)" class1))
	  (setq super-cast (format "((%s)this)" class2))
	  (xref-reduce-useless-this-casts-in-pp-method (buffer-file-name) 
												this-cast super-cast)
	  )
	(save-excursion
	  (setq this-class (xref-get-current-class "-olcxcurrentclass" t))
	  (setq super-class (xref-get-current-class "-olcxcurrentsuperclass" nil))
	  (xref-reduce-this-references-in-method (buffer-file-name) this-class super-class)
	  )
	(save-excursion
	  (xref-nonint-reduce-useless-long-names-of-method (point) nil)
	)
))

(defun xref-pull-up-push-down-field-method (pull-up pull-method
											ref-name tpch-opt  tpostch-opt
											keep-copy)
  (let ((name) (target-class) (source-class) (super-class) (confirmed)
		(tmp-mark) (tpch-options) (keep-source) (tpch-pilot) (tpc-message))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
		"You should edit a Java file when using this function.")
	(setq name (xref-get-identifier-on-point))
	(xref-check-if-target-is-set t)
	(setq tpch-options (format "%s \"-movetargetclass=%s\"" tpch-opt 
							   xref-move-target-class))
	(setq target-class (xref-char-replace xref-move-target-class "/$" ?.))
	(xref-trivial-pre-checks tpch-options)
	(setq tpc-message "")  ;; xref-last-message) ;; does not work if warning mesage
	(setq tpch-pilot xref-pilot-char)        ;; BIG HACK will use this!
	(xref-delete-windows)
	(save-excursion
	  (goto-line (xref-get-method-borne-line))
	  (set-marker xref-move-method-end-marker (point))
	  )
	(if (equal tpch-pilot ?\:)
		(setq confirmed t)
	  (setq confirmed (yes-or-no-p 
					   (format 
						"%s '%s' to '%s:%d' into class '%s'? " 
						ref-name
						name (xref-file-last-name (xref-move-target-file)) 
						(xref-move-target-line) target-class))))

	(if confirmed
		(progn
		  (xref-refactoring-init-actions (format "%s %s" ref-name name))
		  (setq source-class (xref-get-current-class "-olcxcurrentclass" t))
		  (setq super-class (xref-get-current-class "-olcxcurrentsuperclass" nil))
		  
		  (set-marker xref-tmp-mark (point))
		  (xref-refactoring-though-point 
		   "STEP 1. Expanding all context dependent names. (Hit any key to continue)")
		  (xref-expand-short-method-refs (point) nil)
		  (xref-switch-to-marker xref-tmp-mark)

		  (xref-expand-this-references-in-method
		   (buffer-file-name) source-class super-class)
		  (xref-switch-to-marker xref-tmp-mark)

		  (xref-refactoring-though-point "STEP 2. Context check.")
		  (xref-move-method-context-pre-check "-olcxppprecheck" nil tpostch-opt
		   "Aborting. To resolve the problem move critical symbols, make them public or adjust 'this-super' prefixes.")
		  
		  (xref-refactoring-though-point 
		   (format "STEP 3. %s the definition." ref-name))
		  
		  (setq tmp-mark (make-marker))
		  (set-marker tmp-mark (point))

		  (if (and pull-up (not pull-method) (equal tpch-pilot ?\:))
			  (progn
				;; pulling up an existing field, do nothing, just delete it, HACK!!!
				(setq keep-source nil)
				)
			;; otherwise, really move
			(xref-switch-to-marker xref-move-target-marker)
			(beginning-of-line)

			(set-marker xref-tmp-mark (point))
			(xref-switch-to-marker tmp-mark)		  
			(xref-move-method-move-method-body xref-tmp-mark nil xref-move-method-end-marker)

			(xref-non-int-change-field-access-attribute "public" nil)
			
			(xref-refactoring-though-point 
			 "STEP 4. Reduce Redundant Long Names")

			(xref-pull-up-push-down-reduce-expanded-syms source-class super-class)
			
			(xref-switch-to-marker tmp-mark)
			(if keep-copy
				(if (equal keep-copy 'keep-copy-if-used)
					(progn
					  (xref-refactoring-though-point 
					   "STEP 5. Test if original field can be removed.")
					  (xref-non-interactive-update-references 
					   xref-fast-updates-inside-refactorings)
					  (setq keep-source (not (xref-is-this-an-unused-symbol)))
					  )
				  (if (equal keep-copy 'keep-copy-if-deleting-precheck)
					  (progn
						(setq keep-source t)
						(if (string-match "singleDeletingPushDown" tpc-message)
							(progn
							  (if (yes-or-no-p "Original method can be safely deleted, can I do it? ")
								  (setq keep-source nil)
							  ))
						  ))
					(setq keep-source t)
					))
			  (setq keep-source nil)
			  )
			)
		  ;; common finish
		  (if keep-source
			  (progn
				(xref-pull-up-push-down-reduce-expanded-syms  source-class super-class)
				(xref-switch-to-marker xref-move-target-marker)
				(xref-refactoring-final-update-actions)
				(message "All done. Original field/method is kept in superclass for further push downs.")
				)
			(beginning-of-line)
			(delete-char (- (marker-position xref-move-method-end-marker)
							(point)))
			(xref-switch-to-marker xref-move-target-marker)
			(xref-refactoring-final-update-actions)
			)
))))

(defun xref-pull-up-method ()
"Move a  method into its  superclass.  

At  invocation of this  function the  cursor has  to be  positioned on
method's  definition (on  its name).   The method  declaration  has to
start on this line and no other  field or method has to be declared on
the same  line.  The target place  and class where the  method will be
moved has  to be  specified by the  'Set Target Place'  function.  The
method will be  moved in following steps.  1.)   All context dependent
names inside  the method are  expended to fully qualified  names.  2.)
All references inside method are checked to remain correct in method's
new place.   3.) Method  is moved.  4.)   'this' and  'super' keywords
occuring  inside the  method  are adjusted  to  new environment.   5.)
Redundant fully qualififed names are reduced.

  The refactoring  will make the  method public, you can  restrict the
modifier after.
"
(interactive "")
(xref-pull-up-push-down-field-method t t "Pull up" "-tpchpullupmethod" nil nil)
)


(defun xref-push-down-method ()
"Copy a method into one a subclass.

Unless standard implementation of push down method, this function will
create a copy  of the method in its  target position (subclass).  This
corresponds to  the intuition  that you may  want to  push-down method
into  several subclasses.   With  this  function you  can  do this  in
several successive pushing down.  Then  you can delete the method from
superclass.   When  invoking  this  function  the  cursor  has  to  be
positioned  on   method's  definition  (on  its   name).   The  method
declaration has to start on this line and no other field or method has
to be declared on the same line.  The target place and class where the
method will  be moved has  to be specified  by the 'Set  Target Place'
function.   The method  will be  moved in  following steps.   1.)  All
context  dependent  names inside  the  method  are  expended to  fully
qualified  names.  2.)  All  references inside  method are  checked to
remain correct in method's new place.  3.)  Method is COPIED to target
position.  4.)  'this' and 'super' keywords occuring inside the method
are  adjusted to  new  environment.  5.)   Redundant fully  qualififed
names are reduced.

  The refactoring  will make the  method public, you can  restrict the
modifier after.
"
(interactive "")
(xref-pull-up-push-down-field-method nil t "Push down" 
									 "-tpchpushdownmethod" 
									 "-tpchpushdownmethodpostcheck"
									 'keep-copy-if-deleting-precheck
									 )
;; before
;;(xref-pull-up-push-down-field-method nil t "Push down" 
;;									 "-tpchpushdownmethod" 
;;									 "-tpchpushdownmethodpostcheck"
;;									 'keep-copy)
)


(defun xref-pull-up-field ()
"Move field into  its superclass.  

At  invocation of this  function the  cursor has  to be  positioned on
field's definition (on  its name).  The field definition  has to start
on this line  and no other field  or method has to be  declared on the
same line.  The  target place and class where the  field will be moved
has to  be specified  by the 'Set  Target Place' function.   The field
will be  moved in following  steps.  1.)  All context  dependent names
inside its definition are expended to fully qualified names.  2.)  All
references  inside the  definition are  checked to  remain  correct in
method's new place.  3.) Field  is moved.  4.)  Optionally, 'this' and
'super' keywords  occuring inside an  initializer are adjusted  to new
environment.  5.)  Redundant fully qualififed names are reduced.

  The refactoring  will make  the field public,  you can  restrict the
modifier after.
"
(interactive "")
(xref-pull-up-push-down-field-method t nil "Pull up" "-tpchpullupfield" nil nil)
)


(defun xref-push-down-field ()
"Copy a field into a subclass.

At  invocation of this  function the  cursor has  to be  positioned on
field's definition (on  its name).  The field definition  has to start
on this line  and no other field  or method has to be  declared on the
same line.  The  target place and class where the  field will be moved
has to  be specified  by the 'Set  Target Place' function.   The field
will be  moved in following  steps.  1.)  All context  dependent names
inside its definition are expended to fully qualified names.  2.)  All
references  inside the  definition are  checked to  remain  correct in
method's new  place.  3.) Field is  copied into the  subclass.  If the
field  is  no  longer  used  in  superclass it  is  removed  from  the
superclass.   4.)  Optionally,  'this' and  'super'  keywords occuring
inside an initializer are adjusted to new environment.  5.)  Redundant
fully qualififed names are reduced.

  The refactoring  will make  the field public,  you can  restrict the
modifier after.
"
(interactive "")
(xref-pull-up-push-down-field-method nil nil "Push down" "-tpchpushdownfield" nil
									 'keep-copy-if-used)
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                          MOVE FIELD                               
;;

(defun move-field-mapfun (new-name dummy)
  (let ((pos (point)))
	(if (not (xref-last-move-was-to-definition-or-declaration))
		(progn
		  (kill-region (point) (mark t))
		  (insert new-name)
		  (goto-char pos)
		  ))
	nil
))

(defun xref-move-field-finish()
  (let ((tmp-mark) (ofile) (oline) (nfile) (nline) (safetycheckopt))
	;; should be via markers?
	(xref-call-process 'sync 0 nil (point) 0 "" "-olcxgotodef" "" "")
	(xref-moveref t)

	(setq ofile (buffer-file-name))
	(setq oline (count-lines 1 (+ (point) 1)))
	(xref-move-field-make-move)
	(setq nfile (buffer-file-name))
	(setq nline (count-lines 1 (+ (point) 1)))

	(xref-refactoring-though-point "STEP 3. Making field public and reducing long names (if any).")
	(xref-non-int-change-field-access-attribute "public" nil)
	(setq tmp-mark (make-marker))
	(set-marker tmp-mark (point))
	(xref-nonint-reduce-useless-long-names-of-method (point) nil)
	(xref-switch-to-marker tmp-mark)

	(xref-refactoring-though-point "STEP 4. Updating Tag File")
	(message "Updating Tag file, please wait.")
	(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)

	(if (xref-safety-check-prompt)
		(progn
		  ;; update should be sufficient (?for most cases)
		  (xref-non-interactive-update-references xref-fast-updates-inside-refactorings)
		  (xref-call-process 'sync 0 nil (point) 0 "" "-olcxsafetycheckinit" 
						   "-olcxrefsuffix=0:#" "")
		  (message "Checking new references.")
		  (setq 
		   safetycheckopt
		   (concat
			"-olcxsafetycheck2 "
			(format "-olcxsafetycheckmovedfile \"%s\" \"%s\" " ofile nfile)
			(format "-olcxsafetycheckmovedblock=%d:%d:%d" oline 1 nline)
			))
		  (push-references 'sync (buffer-file-name) (point) t nil safetycheckopt 0 '(xref-safety-check-finish))
		  ))
))


(defun xref-move-field-class-prechecks (prefix target-class)
  (let ((tpos) (target-buffer) (argname) (this-class) 
		(bb) (errm) (accstr) (cont) (target-fname) (ftype))
	(setq errm nil)
	(setq tpos (point))
	(if (equal prefix "")
		(progn
		  ;; this is probably movig to outer scope, do nothing, I hope
		  ;; that safety check will report problem.
		  )
	  ;; do any changes on a copy in temporary buffer
	  (setq target-buffer (current-buffer))
	  (setq target-fname (buffer-file-name target-buffer))
	  (get-buffer-create xref-move-method-pre-check-tmp-buffer)
	  (set-buffer xref-move-method-pre-check-tmp-buffer)
	  (xref-erase-buffer)
	  (insert-buffer target-buffer)
	  (goto-char tpos)
	  ;; following has to be reworked on semantic level
	  (search-backward-regexp xref-backward-pass-identifier-regexp)
	  (forward-char 1)
	  (beginning-of-line)
	  (insert-string (format " _moveFieldClassPreChecks$7634() {%s=null;}"
							 prefix))
	  (search-backward "=");
	  (backward-char 1)
	  (setq ftype (xref-get-type-of-symbol target-fname))
	  (if (not ftype)
		  (setq errm (format "Can't get type of %s, continue anyway? " prefix))
		(if (not (equal ftype target-class))
			(error (format "'%s' is not of type %s" prefix target-class))
		  )
		)
	  )
	(kill-buffer xref-move-method-pre-check-tmp-buffer)
	(set-buffer target-buffer)
	(if (not (eq errm nil))
		(progn
		  (setq cont (yes-or-no-p errm))
		  (if (not cont) (error "Aborting."))
		  ))
))



(defun xref-move-field (poin)
"Move field from its 'source' class into a 'target' class.

At invocation of this function the  cursor has to be positioned on the
field defintion  (on its  name).  The  field has to  be declared  on a
separate line, i.e.  its definition  should not exceed one line and no
other symbol  can be declared on  the same line. The  target place and
class where  the field will be moved  has to be specified  by the 'Set
Target Place' function before invocation  of field moving. You will be
prompted for the  way how to obtain target  object from source object.
This is usually a source class field linking to the target class.

  The refactoring  will not update  initializers of the field  if they
depend  on the  current class.   The refactoring  will make  the field
public, you can restrict the modifier after.
"
(interactive "d")
(let ((name) (target-class) (prefix) (new-name))
  (xref-initial-file-suffix-check xref-java-allowed-suffixes
	  "You should edit a Java file when using this function.")
  (xref-trivial-pre-checks "-tpchmovefield")
  (xref-delete-windows)
  (setq name (xref-get-identifier-on-point))
  (xref-check-if-target-is-set t)
  (setq target-class (xref-char-replace xref-move-target-class "/$" ?.))
  (if (yes-or-no-p 
	   (format 
		"Move '%s' to class %s (%s:%d)? " 
		name target-class (xref-file-last-name (xref-move-target-file)) 
		(xref-move-target-line)))
	  (progn
		(sit-for 0)
  		(setq prefix (read-from-minibuffer 
					  (format "Field (string) to get target object from source object: ") ""))
		(if (equal prefix "")
			(setq new-name name)
		  (setq new-name (format "%s.%s" prefix name))
		  )
		(xref-move-field-class-prechecks prefix target-class)
		(xref-refactoring-init-actions (format "moving of %s" name))

		(set-marker xref-tmp-mark (point))
   		(xref-refactoring-though-point 
		 "STEP 1. Expanding context dependent names, if any. (Hit any key to continue)")
   		(xref-expand-short-method-refs (point) nil)
		(xref-switch-to-marker xref-tmp-mark)
		
		(xref-refactoring-though-point 
		 "STEP 2. Changing references and moving field (hit any key to continue)")

		(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxrename"
										xref-detailed-refactoring-confirmations
										nil
										''none-pre-fun 
										''move-field-mapfun
										''none-post-fun 
										new-name
										nil
										nil
										'(xref-move-field-finish)
							   )
))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                          MOVE (RENAME) PACKAGE
;;


(defun xref-resave-buffers-from-directory (olddir newdir)
  (let ((bb) (cb) (fn) (nfn) (res))
	(setq bb (buffer-list))
	(setq res nil)
	(setq olddir (backslashify-name olddir))
	(while bb
	  (setq cb (car bb))
	  (setq fn (backslashify-name (buffer-file-name cb)))
	  (if (xref-string-has-prefix fn olddir (eq xref-platform 'windows))
		  (progn
			(set-buffer cb)
			(setq nfn (concat newdir 
							  (xref-cut-string-prefix fn olddir (eq xref-platform 'windows))))
			(set-visited-file-name nil)
			(xref-write-file nfn nil)
			))
	  (setq bb (cdr bb))
	  )
	res
))

;; TODO, delete this unused function
(defun xref-unlink-buffers-from-directory (olddir)
  (let ((bb) (cb) (fn) (nfn) (res))
	(setq bb (buffer-list))
	(setq res nil)
	(while bb
	  (setq cb (car bb))
	  (setq fn (buffer-file-name cb))
	  (if (xref-string-has-prefix fn olddir (eq xref-platform 'windows))
		  (progn
			(set-buffer cb)
			(set-visited-file-name nil)
			(setq res (cons (cons cb (cons fn nil)) res))
			))
	  (setq bb (cdr bb))
	  )
	res
))

;; TODO, delete this unused function
(defun xref-resave-unlinked-buffers (blist olddir newdir)
  (let ((bb) (cb) (fn) (nfn))
	(setq bb blist)
	(while bb
	  (setq cb (car (car bb)))
	  (setq fn (car (cdr (car bb))))
	  (setq nfn (concat newdir 
						(xref-cut-string-prefix fn olddir (eq xref-platform 'windows))))
	  (set-buffer cb)
	  (xref-write-file nfn nil)
	  (setq bb (cdr bb))
	  )
))

(defun xref-undoable-move-directory (olddir newdir)
  (let ((pdir) (blinks))
	;; create parent dir first (if not exists)
	(setq pdir (xref-file-directory-name newdir))
	(if (not (file-attributes pdir)) (make-directory pdir))
;;	(setq blinks (xref-unlink-buffers-from-directory olddir))
	(xref-save-some-buffers t)
	(xref-move-directory olddir newdir)
	(xref-multifile-undo-add-move-directory olddir newdir)
	(xref-resave-buffers-from-directory olddir newdir)
	(if (file-attributes olddir) (delete-directory olddir))
;;	(xref-resave-unlinked-buffers blinks olddir newdir)
))

(defun xref-rename-package-finish (olddir newdir move-files)
  (xref-popCxrefs (point))
  (if (not move-files)
	  (progn
		(xref-save-some-buffers t)
		(message "Text renaming done. You need to move/merge directories manually !!")
		)
	(xref-refactoring-though-point "Text renaming done. Going to move directory.")
	(xref-undoable-move-directory (backslashify-name olddir)
								  (backslashify-name newdir))
	(xref-refactoring-final-update-actions)
	)
)

(defun xref-get-dir-from-path-and-package (dir package)
  (let ((res) (i) (len))
	(setq res (format "%s/" dir))
	(setq len (length package))
	(setq i 0)
	(while (< i len)
	  (if (eq (elt package i) ?.)
		  (setq res (format "%s/" res))
		(setq res (format "%s%c" res (elt package i)))
		)
	  (setq i (+ i 1))
	  )
	res
))

(defun xref-infer-source-path (fname package)
  (let ((res) (dn) (sp) (spl) (bcp))
	;; try to infer sourcepath from curent file name
	(setq res nil)
	(setq bcp (xref-get-dir-from-path-and-package "" package))
	(setq dn (xref-file-directory-name fname))
	(setq res (xref-cut-string-suffix dn bcp (eq xref-platform 'windows)))
	(if (equal res dn)
		(progn
		  ;; not editing current package, find something on sourcepath
		  (setq res nil)
		  (setq sp (xref-get-env "-sourcepath"))
		  (if sp
			  (setq spl (xref-split-path-on-path-list sp t))
			(setq spl nil)
			)
		  (while (and (eq res nil) spl)
			(if (file-attributes (format "%s%c%s" (car spl) xref-slash bcp))
				(setq res (car spl))
			  )
			(setq spl (cdr spl))
			)
		  ))
	(if (not res)
		(error "Sorry, can't infer sourcepath, retry while editing a file inside moved package.")
	  )
	(setq res (remove-pending-slash res))
	res
))

(defun xref-rename-package ()
"Rename Java package.

This refactoring renames  a package replacing old name  by new name on
all its  occurences in program  text.  In difference to  simple symbol
renaming this  function will change  WHOLE package name, not  only its
last part and it also  moves corresponding source directory.  It means
that it  can rename for  example package 'org.funny'  to 'com.project'
which  will  also  move  .../org/funny to  .../com/project  directory.
However, it  does not discover package names  inside string constants,
if you  are using  Java reflexivity features  you will need  to change
those references manually.

This  refactoring  is  stored  in  the  undo  list  as  two  different
refactorings.  First  are stored  all necessary text  manipulations in
sources  and then  (as second  refactoring)  is stored  moving of  the
directory.  If you are undoing  this refactoring you will need to undo
it as two successive refactorings.

"
  (interactive "")
  (let ((old-name) (new-name) (spath) (rrr) (olddir nil) (newdir nil) (fat) 
		(move-files))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
		"You should edit a Java file when using this function.")
	(xref-trivial-pre-checks "-tpchrenamepackage")
	(setq old-name xref-last-message)
	(setq move-files t)
	(xref-delete-windows)
	(setq new-name (read-from-minibuffer 
					(format "Rename (move) package to : ") old-name))
	;; Include when realy movig dir
	(setq spath (xref-infer-source-path (buffer-file-name) old-name))
	(if (not spath)
		(progn
		  (setq rrr (yes-or-no-p (format "Can't infer sourcepath. Continue anyway ?" new-name)))
		  (if (not rrr) (error "Exiting."))
		  (setq move-files nil)
		  (setq spath "")
		  ))
	(setq olddir (xref-get-dir-from-path-and-package spath old-name))
	(if (and move-files (not (file-attributes olddir)))
		(error "[error] %s does not exists, something is wrong, exiting" olddir)
	  )
	(setq newdir (xref-get-dir-from-path-and-package spath new-name))
	(setq fat (file-attributes newdir))
	(if (and move-files fat)
		(progn
		  (setq rrr (yes-or-no-p (format "Directory of %s exists. Continue anyway ?" new-name)))
		  (if (not rrr) (error "Exiting."))
		  (setq move-files nil)
		  ))
	;; end of Include when realy movig dir
	(xref-refactoring-init-actions
	 (format "textual renaming of package %s " old-name))
	(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxrename"
						   xref-detailed-refactoring-confirmations
						   nil
						   ''none-pre-fun
						   ''xref-fqt-name-modify-static-prefix
						   ''standard-post-fun 
						   new-name
						   t           ;; delete flag for fqt-modif
						   nil
						   (list 'xref-rename-package-finish olddir newdir move-files)
						   )
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                          MOVE CLASS

(defun xref-nonint-remove-useless-long-class-names (poin interact threewin preaction)
	(xref-map-on-all-references  (buffer-file-name) poin nil threewin
						   "-olcxuselesslongnamesinclass"
						   interact 
						   nil
						   ''none-pre-fun
						   ''xref-fqt-name-modify-static-prefix
						   ''standard-post-fun 
						   ""
						   nil
						   (list 'progn 
								 preaction
								 (list 'xref-call-process ''sync 0 nil 
									   0 0 "" "-olcxgoto0" "" ""
									   ))
						   (list 'xref-popCxrefs poin)
						   )
)


(defun xref-nonint-reduce-useless-long-names-of-class (poin interact)
  (save-excursion
	(search-backward-regexp "[ \t\n]" 0 t)
	(xref-nonint-remove-useless-long-class-names poin interact xref-three-windows (list 'xref-set-olcxfilter 2))
	)
)

(defun xref-reduce-useless-long-names-of-class (poin)
"Reduce all useless Java fully qualified type names within class.

This function is meaningful only for Java programmers.  It causes that
the  currently edited  buffer  is parsed,  references  of symbols  are
recorded. All redundant fully  qualified references within the body of
the class where the point is are reduced to unambiguous short names.
"
  (interactive "d")
  (xref-multifile-undo-set-buffer-switch-point "reduction of long names")
  (xref-nonint-reduce-useless-long-names-of-class poin xref-detailed-refactoring-confirmations)
)

(defun xref-move-class-finish (move-type target-class-type)
  (let ((moved-pos-mark) (dp) (ff))
	;; move the class
	(xref-popCxrefs (point))
	(xref-move-method-move-method-body xref-move-target-marker t 
									   xref-move-method-end-marker)
	(setq moved-pos-mark (make-marker))
	(set-marker moved-pos-mark (point))
	(if (string-match "transpackage" move-type)
		(xref-non-int-change-field-access-attribute "public" nil)
	  )
	;;(message "moving %s to %s" move-type target-class-type)
	(if (and (string-match "outer" move-type)
			 (equal target-class-type "nested"))
		(progn
		  ;; outer->inner, insert static
		  (setq dp (point))
		  (beginning-of-line)
		  (setq ff (search-forward-regexp "[^ \\t]" dp 0))
		  (backward-char 1)
		  (insert "static ")
		  (goto-char dp)
		  ))
	(if (and (string-match "nested" move-type)
			 (equal target-class-type "outer"))
		(progn
		  ;; inner->outer, remove static
		  (setq dp (point))
		  (beginning-of-line)
		  (setq ff (search-forward-regexp "static " dp t))
		  (if ff
			  (progn
				(delete-char -7)
				))
		  (goto-char dp)
		  ))

	;; following will refresh the screen
	(switch-to-buffer (current-buffer))
	(xref-refactoring-though-point "STEP 3. Updating Tag File")

	(message "Updating Tag file, please wait.")
	(xref-non-interactive-update-references xref-fast-updates-inside-refactorings)

	;; repush references (for reduce long names) as they are now shifted
	(xref-switch-to-marker moved-pos-mark)
	(xref-refactoring-though-point "STEP 4. Reduce Redundant Long References")
	(push-references 'sync (buffer-file-name) (point) t nil "-olcxrename" 0 nil)
	(xref-moveref nil) ???
	(xref-switch-to-marker moved-pos-mark)
	(xref-nonint-remove-useless-long-names (point) xref-detailed-refactoring-confirmations nil 
	   (list 'remove-useless-long-names-yet-pushed))

	(xref-poponly)	
	(xref-switch-to-marker moved-pos-mark)

	(xref-refactoring-though-point 
	 "STEP 5. Reduce Redundant Long Names inside Class")

	(xref-nonint-reduce-useless-long-names-of-class (point) nil)
	;;	(xref-save-modified-files-with-question t "All done. ")
	(xref-refactoring-final-update-actions)
))


(defun xref-move-class ()
"Move Java class.

This  function is available  only for  Java programming  language.  At
invocation of  this function  the cursor has  to be positioned  on the
class defintion (on  its name).  The class modifiers  should be on the
same line  as class name.   The target place  where the class  will be
moved has  to be specified by  the 'Set Target  Place' function before
invocation of class moving.

  This  function   can  move  top-level  classes   and  static  nested
classes. It is not able to move inner classes (even if they do not use
symbols from enclosing instances).

  If you  move a  top-level class outside  from its  original package,
then the class is made public and should be moved into a file with the
same name as moved class.  Xrefactory will not create the file nor the
'package'  clause or  import  clauses  in the  new  file.  This  'file
header' has to be written by user before moving the body of the class.
"
  (interactive "")
  (let ((name) (target-class) (move-type) (target-class-type))
	(xref-initial-file-suffix-check xref-java-allowed-suffixes
		"You should edit a Java file when using this function.")
	(xref-check-if-target-is-set nil)
	(xref-trivial-pre-checks 
	 (format "-tpchmoveclass \"-movetargetclass=%s\" \"-movetargetfile=%s\"" 
			 xref-move-target-class (xref-move-target-file)))
	(setq move-type xref-last-message)
	(setq target-class (xref-char-replace xref-move-target-class "/$" ?.))
	(xref-delete-windows)
	(setq name (xref-get-identifier-on-point))
	(save-excursion
	  (goto-line (xref-get-class-borne-line))
	  (set-marker xref-move-method-end-marker (point))
	  )
	(if (yes-or-no-p 
		 (format 
		  "Move '%s' to %s:%d? " 
		  name (xref-file-last-name (xref-move-target-file)) 
		  (xref-move-target-line)))
	  (progn
		(if (and (> (length target-class) 0) (eq (elt target-class 0) ?\ ))
			(progn
			  (setq target-class (substring target-class 1))
			  (setq target-class-type "outer")
			  )
		  (setq target-class-type "nested")
		  )
		(xref-refactoring-init-actions (format "moving of %s" name))
		(xref-refactoring-though-point "STEP 1. Expanding short names.")
		(xref-expand-short-class-refs (point) nil)
		(xref-refactoring-though-point "STEP 2. Updating class references and moving it.")
		(xref-map-on-all-references  (buffer-file-name) (point) t nil "-olcxrename"
									 xref-detailed-refactoring-confirmations
									 nil
									 ''none-pre-fun
									 ''xref-fqt-name-modify-static-prefix
									 ''standard-post-fun 
									 (format "%s.%s" target-class name)
									 t           ;; delete flag for fqt-modif
									 nil
									 (list 'xref-move-class-finish move-type target-class-type)
						   )
		
		))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


